/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.backend;

import static org.eclipse.core.runtime.Assert.isTrue;
import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createDefaultDesignSpaceExploration;

import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.exploration.alg.dse.Explorer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3SystemModelAdapter;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformExploration;
import org.fortiss.af3.exploration.alg.feature.safetyarch.SafetyArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.scheduling.TTScheduleExploration;
import org.fortiss.af3.exploration.alg.feature.taskmapping.TaskMappingExploration;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.service.IDSEBackend;
import org.fortiss.af3.partition.model.allocation.TaskToPartitionAllocationTable;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.GatewayUnit;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.allocation.ComponentToTaskAllocationEntry;
import org.fortiss.af3.task.model.allocation.TaskToExecutionUnitAllocationTable;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Contributor class to allow launching this DSE backend from the DSE perspective.
 * 
 * @author diewald
 */
public class Opt4JDseBackend implements IDSEBackend {

	/** {@inheritDoc} */
	@Override
	public Set<Class<?>> getSynthesizedArtifacts() {
		Set<Class<?>> synthArtifacts = new HashSet<>();
		synthArtifacts.add(PlatformArchitecture.class);
		synthArtifacts.add(TaskToExecutionUnitAllocationTable.class);
		synthArtifacts.add(TaskToPartitionAllocationTable.class);
		return synthArtifacts;
	}

	/** {@inheritDoc} */
	@Override
	public void validateExplorationSpecification(ExplorationSpecification spec) throws Exception {
		SuperSetMap ssMap = spec.getSearchSpace();
		isTrue(ssMap.containsKey(Task.class), "The task set is missing from the search space.");
		isTrue(ssMap.containsKey(Signal.class), "The signal set is missing from the search space.");
		isTrue(ssMap.containsKey(ExecutionUnit.class),
				"The execution unit set is missing from the search space.");
		isTrue(ssMap.containsKey(TransmissionUnit.class),
				"The transmission unit set is missing from the search space.");
		isTrue(ssMap.containsKey(GatewayUnit.class),
				"The gateway unit set is missing from the search space.");
	}

	/** {@inheritDoc} */
	@Override
	public Optional<ExplorationSolution> executeDSE(ExplorationSpecification expSpec,
			Set<Class<? extends IModelElement>> solutionModelTypes, IProgressMonitor monitor,
			int timeoutMS) throws Exception {
		SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter;
		SuperSetMap superSets = expSpec.getSearchSpace();

		// Input to DSE: {@link SystemModelAdapter} and {@link DesignSpaceExploration}.
		// SystemModelAdapter systemModelAdapter;
		systemModelAdapter = new AF3SystemModelAdapter(superSets.get(Task.class),
				superSets.get(ComponentToTaskAllocationEntry.class),
				superSets.get(ExecutionUnit.class), superSets.get(Signal.class));
		DseSpecification dseSpec = createDefaultDesignSpaceExploration(expSpec);
		addDefaultExplorationFeatures(dseSpec);

		if(solutionModelTypes.contains(PlatformArchitecture.class)) {
			dseSpec.getFeatures().add(PlatformExploration.class);
		}

		try {
			Explorer explorer = new Explorer();
			ExplorationSolution explorationSolution =
					explorer.explore(dseSpec, systemModelAdapter, monitor, true);
			return Optional.of(explorationSolution);
		} catch(Exception ex) {
			// TODO(AD): readd code to present the schedule which failed... Use the LoggingService
			// from the kernel. The schedule export has to be done before.
			throw ex;
		}
	}

	/**
	 * Adds the default set of {@link IExplorationFeature}s to the given {@link DseSpecification}.
	 */
	private void addDefaultExplorationFeatures(DseSpecification dseSpec) {
		dseSpec.getFeatures().add(SafetyArchitectureExploration.class);
		// dseSpec.getFeatures().add(PartitionArchitectureExploration.class);
		dseSpec.getFeatures().add(TaskMappingExploration.class);
		// dseSpec.getFeatures().add(DesignDiversityExploration.class);
		dseSpec.getFeatures().add(TTScheduleExploration.class);
	}

	/** {@inheritDoc} */
	@Override
	public String getName() {
		return "MOEA-based Exploration";
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return getName();
	}
}
