/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static com.google.inject.Guice.createInjector;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.copyExplorationSpecification;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.getInputProvider;
import static org.fortiss.af3.exploration.alg.util.ExplorationFeatureUtils.getEnabledApplicableFeatures;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;

import java.time.Duration;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Predicate;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.swt.widgets.Display;
import org.fortiss.af3.exploration.alg.dse.backend.IExplorerBackend;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JExplorerBackend;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.DseInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.create.CreateInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.create.CreateInterceptModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.create.CreateInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.DecodeInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.DecodeInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.DecodingInterceptModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.EvaluatorInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.thread.MultiThreadingModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.scheduling.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.plot.XYPlotter;
import org.fortiss.af3.exploration.alg.service.IExpSpecTransformationService;
import org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.solutions.ExplorationResult;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolution;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.fortiss.af3.exploration.moea.model.parameters.IOptimizer;
import org.fortiss.af3.exploration.util.ExplorationUtils;
import org.jfree.ui.RefineryUtilities;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.Sets;
import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Module;
import com.google.inject.TypeLiteral;

/**
 * Explorer - entry point to architectural exploration
 * 
 * @author huang
 * @author diewald
 */
public class Explorer {

	/** Last {@link XYPlotter} (in order to close it before creating a new one. */
	protected XYPlotter lastPlotter;

	/**
	 * Associates the original {@link ExplorationTarget}s with the internal ones passed to the
	 * solver.
	 */
	private Map<ExplorationTarget<?>, ExplorationTarget<?>> origToInternalTarget = new HashMap<>();

	/**
	 * Performs the DSE defined a given {@link DseSpecification} specification on a given
	 * system model (represented by a {@link SystemModelAdapter}).
	 */
	@SuppressWarnings("unchecked")
	public <T> ExplorationSolution explore(DseSpecification dseSpec,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			IProgressMonitor progressMonitor, boolean generateImplicitConstraints)
			throws Exception {
		long dseStartTime = System.currentTimeMillis();

		ExplorationSpecification origExpSpec = dseSpec.getTargetSpecification();

		Set<TypeLiteral<? extends Phenotype>> internalSolutionEncodings =
				Sets.newHashSet(TypeLiteral.get(StrictTTSchedule.class));

		EncodingProviderModule<?> inputProvider =
				getInputProvider(dseSpec, dseSpec.getTargetSpecification());
		Injector inputInjector = Guice.createInjector(inputProvider);
		Collection<IExplorationFeature> enabledFeatures =
				getEnabledApplicableFeatures(dseSpec, inputInjector, false);
		dseSpec.getApplicableFeatures().addAll(enabledFeatures);

		ExplorationSpecification expSpec = perprocessExpSpec(dseSpec, systemModelAdapter,
				inputProvider, generateImplicitConstraints, origExpSpec);
		InternalSpecificationProvider specProvider =
				new InternalSpecificationProvider(dseSpec, internalSolutionEncodings);

		// Invoke the backend in order to perform the DSE
		IExplorerBackend<? extends IOptimizer> explorerBackend = createExplorerBackend(dseSpec,
				expSpec, inputProvider, specProvider, progressMonitor);
		ExplorationSolution internalExplorationResult = explorerBackend.explore();
		conditionallyShowDebugVisualization(internalExplorationResult);

		// We use the original ExplorationSpecification for the returned result.
		ExplorationSolution explorationResult = copy(internalExplorationResult);
		explorationResult.setExplorationSpec(origExpSpec);
		// Also, replace every internal target->result representation with a original target->result
		// one, wherever possible.
		for(SingleExplorationSolution singleSol : explorationResult.getSolutions()) {
			for(Map.Entry<ExplorationTarget<?>, ExplorationTarget<?>> entry : origToInternalTarget
					.entrySet()) {
				ExplorationTarget<?> internalTarget = origToInternalTarget.get(entry.getKey());
				if(!(internalTarget instanceof InternalConstraint)) {
					ExplorationResult<?> result = singleSol.removeKey(internalTarget);
					singleSol.put((ExplorationTarget<T>)entry.getKey(),
							(ExplorationResult<T>)result);
				}
			}
		}

		// Remove the implicit ExplorationTargets from the metrics of each single solution.
		removeInternalTargets(explorationResult);

		if(ExplorationAlgUtils.DSE_DEBUG_PRINT_TIME) {
			Duration readableDuration =
					Duration.ofMillis(System.currentTimeMillis() - dseStartTime);
			System.out.println("Overall duration:");
			System.out.println("Days: " + readableDuration.toDaysPart());
			System.out.println("Hours: " + readableDuration.toHoursPart());
			System.out.println("Minutes: " + readableDuration.toMinutesPart());
			System.out.println("Seconds: " + readableDuration.toSecondsPart());
		}

		return explorationResult;
	}

	/**
	 * Shows a visualization of the {@link ExplorationSolution} that includes the evaluations of
	 * internal constraints and objectives.
	 */
	private void conditionallyShowDebugVisualization(ExplorationSolution explorationSolution) {
		if(ExplorationUtils.isDebugVerboseEnabled()) {
			Display.getDefault().asyncExec(() -> plotResult(explorationSolution));
		}
	}

	/**
	 * Copies the input {@link ExplorationException} and applies the D3SE input transformation
	 * service to the copy.
	 * 
	 * @param dseSpec
	 *            input {@link DseSpecification} defining the parameters of the MOEA backend.
	 * @param systemModelAdapter
	 *            Representation of the input models.
	 * @param inputProvider
	 *            {@link EncodingProviderModule} to provide the internal representations of
	 *            elements extracted from the input models.
	 * @param generateImplicitConstraints
	 *            flag whether or not to generate some of the implicit constraints (debug).
	 * @param origExpSpec
	 *            original input {@link ExplorationSpecification}.
	 * @return copied and modified input {@link ExplorationSpecification}.
	 * @throws ExplorationException
	 *             if some of the applied transformation fails.
	 */
	protected ExplorationSpecification perprocessExpSpec(DseSpecification dseSpec,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			EncodingProviderModule<?> inputProvider, boolean generateImplicitConstraints,
			ExplorationSpecification origExpSpec) throws ExplorationException {
		// Copy the input specification and prepare it for further processing.
		BiMap<EObject, EObject> copyRefMap = HashBiMap.create();
		ExplorationSpecification expSpec = copyExplorationSpecification(origExpSpec, copyRefMap);

		Collection<Module> modules = new ArrayList<>();
		modules.add(inputProvider);
		Injector inputInjector = createInjector(modules);

		// Execute the Constraint Transformation.
		List<ExplorationTarget<?>> targets = expSpec.getTargets();
		IExpSpecTransformationService expSpecSrv = getService(IExpSpecTransformationService.class);
		List<ExplorationTarget<?>> transformedTargets =
				expSpecSrv.transform(inputInjector, targets, origExpSpec.getTargets(), copyRefMap);
		// @CodeFormatterOff
		transformedTargets.stream().
			forEach(tgt -> {
				// We have to track the original to transformed target assocs such that we can
				// return a solution that references only the original targets.
				ExplorationTarget<?> origTarget = (ExplorationTarget<?>)copyRefMap.inverse().get(tgt);
				origToInternalTarget.put(origTarget,tgt);
		});
		// @CodeFormatterOn

		// Add implicit Constraints.
		if(generateImplicitConstraints) {
			expSpec.getTargets().addAll(ImplicitExplorationTargetFactory
					.createImplicitExplorationTargets(dseSpec, inputInjector));
		}
		dseSpec.setTargetSpecification(expSpec);

		return expSpec;
	}

	/** Helper method to issue an exception for invalid DSE configurations */
	private void createDesignSpaceExplorationConfigurationException(final String msg)
			throws Exception {
		throw new Exception("DSE configuration: " + msg);
	}

	/**
	 * Removes the implicit {@link ExplorationTarget}s that were added internally. They shall not be
	 * visible to the user.
	 */
	private void removeInternalTargets(ExplorationSolution explorationSolution) {
		Predicate<? super ExplorationTarget<?>> isOrigTarget =
				t -> !origToInternalTarget.keySet().contains(t);
		Consumer<SingleExplorationSolution> removeNonOrigTargets =
				s -> s.keySet().stream().filter(isOrigTarget).forEach(t -> s.removeKey(t));
		explorationSolution.getSolutions().forEach(removeNonOrigTargets);
	}

	/** Explorer backend */
	private IExplorerBackend<? extends IOptimizer> createExplorerBackend(DseSpecification dse,
			ExplorationSpecification expSpec, EncodingProviderModule<?> inputProvider,
			InternalSpecificationProvider specProvider, IProgressMonitor progressMonitor)
			throws Exception {
		DseInjectorProvider provider = new DseInjectorProvider();

		Collection<Module> dseModules = new ArrayList<>();
		dseModules.add(inputProvider);
		dseModules.add(specProvider);
		dseModules.add(new CreateInterceptModule(new CreateInterceptor()));
		dseModules.add(new CreateInjectorProvider());
		dseModules.add(new DecodingInterceptModule(new DecodeInterceptor()));
		dseModules.add(new DecodeInjectorProvider());
		dseModules.add(new EvaluatorInjectorProvider());
		dseModules.add(new MultiThreadingModule());
		dseModules.add(provider);

		Injector dseInjector = Guice.createInjector(dseModules);
		provider.setInjector(dseInjector);

		if(dse.getOptimizer() instanceof EAOptimizer) {
			return new Opt4JExplorerBackend(dse, expSpec, dseInjector, inputProvider,
					progressMonitor);
		}

		createDesignSpaceExplorationConfigurationException("Unsupported exploration backend.");

		// Never here
		return null;
	}

	/**
	 * Creates and launches the Dialog that displays the results from the exploration.
	 * Pareto-optimal points are displayed in a 2D-graph, while schedules can be inspected for
	 * each solution.
	 */
	public void plotResult(ExplorationSolution solutionSet) {
		if(solutionSet != null) {
			// Keep only one DSE result window
			if(lastPlotter != null) {
				lastPlotter.dispose();
			}

			// draw figure
			XYPlotter plotter = new XYPlotter(solutionSet,
					"Design Space Exploration Result (generations: N/A)");
			plotter.createPlot();
			plotter.pack();
			RefineryUtilities.centerFrameOnScreen(plotter);
			plotter.setVisible(true);

			lastPlotter = plotter;
		}
	}
}
