/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createLatencyMinObjective;
import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createNoEmptyPartitionConstraint;
import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createPartitionMinObjective;
import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createPeriodConstraint;
import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createSameSilInPartitionConstraint;
import static org.fortiss.tooling.common.util.LambdaUtils.filterByType;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.exception.ConstraintGenerationException;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.feature.partition.PartitionArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.scheduling.TTScheduleExploration;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.project.DSE;
import org.fortiss.af3.exploration.model.project.ModelSnapshot;
import org.fortiss.af3.exploration.model.project.ProcessStep;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.PeriodConstraint;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.timing.model.PeriodicConstraint;
import org.fortiss.af3.timing.model.TimingSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.TypeLiteral;

/**
 * Factory for creating implicit Exploration Targets based on the input system model.
 * 
 * @author diewald
 */
public class ImplicitExplorationTargetFactory {

	/** Private constructor to prevent instantiation from external (singleton). */
	private ImplicitExplorationTargetFactory() {
	}

	/**
	 * Creates "implicit" {@link ExplorationTarget}s from the {@link SystemModelAdapter} and adds
	 * them to the given {@link DseSpecification} model.
	 * "Implicit" {@link ExplorationTarget}s are defined by the system model itself and cannot be
	 * <i>directly</i> modified by the user.
	 * 
	 * @param dse
	 *            {@link DseSpecification} to which the "implicit" {@link ExplorationTarget}s
	 *            shall be added.
	 * @param inputInjector
	 *            Injector that holds the {@link InputModule}s.
	 * @return Collection of constraints generated from the model.
	 * @throws ExplorationException
	 *             if any of the constraint generators fails.
	 */
	public static Collection<ExplorationTarget<?>> createImplicitExplorationTargets(
			DseSpecification dse, Injector inputInjector) throws ExplorationException {
		List<ExplorationTarget<?>> targets = new ArrayList<>();
		targets.addAll(createPeriodConstraints(inputInjector));

		if(dse.getFeatures().contains(PartitionArchitectureExploration.class)) {
			targets.add(createNoEmptyPartitionConstraint());
			targets.add(createSameSilInPartitionConstraint());
			targets.add(createPartitionMinObjective());
		}

		// Move to the Opt4JDseBackend for visible results, but broken models.
		if(dse.getFeatures().contains(TTScheduleExploration.class)) {
			targets.add(createLatencyMinObjective());
		}

		// Reassign the ids of generated constraints. By default, they are set to 0. Each constraint
		// needs a "unique" id for its constraint type.
		int constrId = 0;
		for(ExplorationConstraint<?> curConstraint : pickInstanceOf(ExplorationConstraint.class,
				targets)) {
			curConstraint.setId(constrId);
			constrId++;
		}

		return targets;
	}

	/**
	 * Creates a {@link PeriodConstraint} for each {@link ITaskGraph} present in the
	 * {@link SystemModelAdapter}.
	 * 
	 * @throws ConstraintGenerationException
	 *             if the generation of any {@link PeriodConstraint} fails.
	 */
	private static Collection<InternalConstraint<?>> createPeriodConstraints(Injector inputInjector)
			throws ConstraintGenerationException {
		Collection<InternalConstraint<?>> constraints = new ArrayList<>();

		TaskMapper taskMapper = inputInjector.getInstance(Key.get(new TypeLiteral<TaskMapper>() { // ...
		}));

		// HACK: We are directly accessing the FileProject, which is actually forbidden.
		// TODO(AD): Remove as soon as temporal constraints are correctly passed via the
		// ExplorationSpecification.
		TimingSpecification timingSpec = extractTimingSpec(inputInjector);
		for(PeriodicConstraint pC : filterByType(timingSpec.getConstraints(),
				PeriodicConstraint.class)) {
			IModelElement taskME = pC.getEvent().getEntity();
			if(!(taskME instanceof Task)) {
				throw new ConstraintGenerationException(PeriodConstraint.class,
						"One of the periodic constraints is invalid." +
								" It does not point to a Task. ID: " + pC.getId() + ".");
			}
			ITaskAdapter task = taskMapper.getAdapter((Task)taskME);

			constraints.add(createPeriodConstraint(task, pC.getPeriod().doubleValue()));
		}

		return constraints;
	}

	/** Extracts the {@link TimingSpecification} from the input set of the {@link DSE}. */
	private static TimingSpecification extractTimingSpec(Injector inputInjector) {
		Set<IExecutionUnitAdapter> execUnitSet =
				inputInjector.getInstance(Key.get(new TypeLiteral<Set<IExecutionUnitAdapter>>() { // ...
				}));
		ExecutionUnit execUnit = getFirst(execUnitSet).get().getObject(ExecutionUnit.class);
		DSE dse = getParentElement(execUnit, DSE.class, false);
		ProcessStep initialStep = dse.getInitialStep();
		List<ModelSnapshot> snaps = initialStep.getModelSnapshots();
		return (TimingSpecification)snaps.stream()
				.filter(s -> s.getRootElement() instanceof TimingSpecification)
				.map(s -> s.getRootElement()).findAny().get();
	}
}
