/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext;

import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.Map;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.constraint.IConstrainedArtifactMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.module.CreatorModule;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.IConstrainedModule;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

import com.google.inject.AbstractModule;
import com.google.inject.Injector;
import com.google.inject.Module;
import com.google.inject.TypeLiteral;

/**
 * This {@link MethodInterceptor} is intended to be used with {@link CreatorModule}s or
 * {@link DecoderModule}s where the same instance of a {@link Genotype} or {@link Phenotype} may be
 * demanded multiple times in the same phase.
 * multiple ones "produce" the same {@link Decoder}.
 * It intercepts the decode method call (defined by an annotation) and launches all sub-
 * {@link DecoderModule}s before executing its own decoding. Sub-{@link DecoderModule}s are not
 * allowed to interfere with each other.
 * 
 * @author diewald
 */
public abstract class ProvisionInterceptor<T> extends AbstractModule implements MethodInterceptor {

	/**
	 * Reference to the {@link Injector} used to execute the create/decoding. Required to find all
	 * {@link Module}s that produce the same type.
	 */
	protected Injector injector;

	/** Holds the constraints of the current individual. */
	protected IConstrainedArtifactMap cMap;

	/** Artefact type: Returned by the invoked methods. E.g., Genotype, or Phenotype. */
	protected Class<T> artefactType;

	/** The type of the cache used to collect the results of the invoked method. */
	protected Class<? extends ArtefactCache<T>> cacheType;

	/** Sets the {@link Injector} used in the create or decoding phase. */
	public void setInterceptParameters(Injector injector, IConstrainedArtifactMap cMap,
			Class<T> encodingType, Class<? extends ArtefactCache<T>> cacheType) {
		this.injector = injector;
		// TODO(AD): Evaluate whether it is possible to get this map in the interceptor by the
		// injector.
		this.cMap = cMap;
		this.artefactType = encodingType;
		this.cacheType = cacheType;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Object invoke(MethodInvocation invocation) throws Throwable {
		Method invokedMethod = invocation.getMethod();
		Type retType = invokedMethod.getGenericReturnType();
		TypeLiteral<T> providedType = (TypeLiteral<T>)TypeLiteral.get(retType);
		if(!artefactType.isAssignableFrom(providedType.getRawType())) {
			throw new Exception("Could not provide: The return type from an operation is not a " +
					artefactType.getClass() + " or the method interceptor was bound to an" +
					" invalid method.");
		}

		// Check if the requested phenotype was already decoded. If so, return
		// this one instead.
		Map<TypeLiteral<? extends T>, T> aggregateMap = injector.getInstance(cacheType);
		T existingEncoding = aggregateMap.get(providedType);
		if(existingEncoding != null && !invokedMethod.isAnnotationPresent(UpdatesEncoding.class)) {
			return existingEncoding;
		}

		// Execute suboperations before the "primary operation to update the encoding before
		// actually using it.
		// This feature is non-composable: We do not support updates for update methods to avoid
		// endless recursion here.
		T encoding = null;
		if(!invokedMethod.isAnnotationPresent(UpdatesEncoding.class)) {
			encoding = executeSubOperations(invocation, providedType);
		}

		// Pass constraints to the encoding.
		// TODO(AD): Evaluate if this code can be placed in a second interceptor for better
		// encapsulation.
		Object calledObject = invocation.getThis();
		if(calledObject instanceof IConstrainedModule) {
			IConstrainedModule cModule = (IConstrainedModule)calledObject;
			Collection<InternalConstraint<?>> relevantConstaints =
					cMap.getConstraintsOf(cModule.getSourceType(), cModule.getTargetType());
			cModule.setConstraints(relevantConstaints);
		}

		// The return type of a decode method must be a Phenotype. It is the same method whose
		// execution is prevented by the overrides check above (return type 'Phenotype').
		encoding = (T)invocation.proceed();
		if(encoding instanceof IExplorationEncoding &&
				((IExplorationEncoding)encoding).isInfeasible()) {
			throw new InfeasibleEncodingException();
		}

		aggregateMap.put(providedType, encoding);

		// Collect all generated constraint for subsequent calls.
		if(calledObject instanceof IConstrainedModule) {
			Collection<InternalConstraint<?>> genConstr =
					((IConstrainedModule)calledObject).getGeneratedConstraints();
			cMap.addConstraints(genConstr);
		}

		return encoding;
	}

	/**
	 * To be implemented for interceptors that support updating of an encoding. These are launched
	 * before the actual method is invoked (preprocessing). If this feature is not supported, just
	 * return {@code null}.
	 * Update methods must be annotated with @UpdatesEncoding.
	 * 
	 * @param invocation
	 *            Standard decoding method.
	 * @param returnType
	 *            expected return type: return type of the "standard" decoding method.
	 * @return Updated encoding.
	 * @throws Exception
	 *             to be throw if return types do not match.
	 */
	protected abstract T executeSubOperations(MethodInvocation invocation,
			TypeLiteral<T> returnType) throws Exception;
}
