/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.ArtefactCache;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.PreservablePhenotypeHolder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.thread.ThreadScoped;
import org.opt4j.core.Genotype;

import com.google.inject.Provides;
import com.google.inject.TypeLiteral;

/**
 * Provides an {@link ArtefactCache} for {@link Phenotype}s that collects all sub {@link Phenotype}s
 * of a single Individual in the decode phase. This cache is thread-safe as it provides each
 * requesting thread with a different instance.
 * 
 * @author diewald
 */
public class PhenotypeCache extends ArtefactCache<Phenotype> {

	/** Holder singleton of Phenotypes that can be used in the next iteration. */
	public PreservablePhenotypeHolder preservedPhenos;

	/** Prototype Constructor. */
	private PhenotypeCache() {
	}

	/** Constructor. */
	private PhenotypeCache(PreservablePhenotypeHolder preservedPhenos) {
		this.preservedPhenos = preservedPhenos;
	}

	/** Returns a new Prototype of the Phenotype Cache to install the Provider in Guice. */
	public static PhenotypeCache getPrototype() {
		return new PhenotypeCache();
	}

	/**
	 * Returns an instance of the {@link PhenotypeCache} per thread. If the cache is requested from
	 * the same thread more than once, the existing instance will be provided, which is guaranteed
	 * by the {@code @ThreadScoped} annotation.
	 */
	@Provides
	@ThreadScoped
	public PhenotypeCache getEncodingCache(PreservablePhenotypeHolder holder) {
		return new PhenotypeCache(holder);
	}

	/**
	 * Adds all preserved {@link Phenotype}s that are valid for the given {@link Genotype} to
	 * {@code this} {@link PhenotypeCache} instance.
	 * 
	 * @param genotype
	 *            key for which to the fill the {@link PhenotypeCache} with its preserved
	 *            {@link Phenotype}s.
	 */
	@SuppressWarnings("unchecked")
	public <T extends Phenotype> void addAllPreservedPhenos(Genotype genotype) {
		Collection<Phenotype> phenos = preservedPhenos.getClearPreservedPhenosFor(genotype);
		phenos.forEach(p -> put(TypeLiteral.get((Class<Phenotype>)p.getClass()), p));
	}
}
