/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.genotype;

import java.util.Map.Entry;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.constraint.IConstrainedArtifactMap;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;

import com.google.inject.TypeLiteral;

/**
 * {@link CompositeGenotype} that additionally holds an {@link EncodingProviderModule} which
 * provides access to the {@link Genotype}s by Guice @Provides methods.
 * 
 * @author diewald
 */
public class ProvidingGenotype extends CompositeGenotype<TypeLiteral<? extends Genotype>, Genotype>
		implements IExplorationEncoding {

	/**
	 * Holds the reference to the {@link EncodingProviderModule} provided along this
	 * {@link CompositeGenotype}.
	 */
	private EncodingProviderModule<Genotype> providerModule;

	/**
	 * References the generator "singleton" that is used to construct
	 * {@link EncodingProviderModule}s.
	 */
	private GenotypeProviderGenerator providerGenerator;

	/** Map holding constraints from the create phase. */
	private IConstrainedArtifactMap cMap;

	/** Marks an encoding infeasible based on the sub-{@link Genotype}s. */
	private boolean isInfeasible = false;

	/** Constructor. */
	public ProvidingGenotype(
			CompositeGenotype<TypeLiteral<? extends Genotype>, Genotype> compositeGenotype,
			GenotypeProviderGenerator providerGenerator, IConstrainedArtifactMap cMap) {
		for(Entry<TypeLiteral<? extends Genotype>, Genotype> entry : compositeGenotype) {
			map.put(entry.getKey(), entry.getValue());
		}
		this.providerGenerator = providerGenerator;
		this.cMap = cMap;
	}

	/**
	 * Returns the {@link EncodingProviderModule} that is passed along this
	 * {@link CompositeGenotype}.
	 */
	public <T extends Genotype> EncodingProviderModule<Genotype> getProviderModule() {
		if(providerModule == null) {
			this.providerModule = providerGenerator.generateInstanceProviders(this);
		}
		return providerModule;
	}

	/** {@inheritDoc} */
	@Override
	public void setInfeasible() {
		isInfeasible = true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isInfeasible() {
		return isInfeasible;
	}

	/** {@inheritDoc} */
	@Override
	public <T extends InternalConstraint<?>> T getConstraint(Class<T> cType,
			Object constrainedElement) {
		// Not needed for the providing Genotype.
		return null;
	}

	/** Returns the constraint map attached to a {@link ProvidingGenotype}. */
	public IConstrainedArtifactMap getConstraintMap() {
		return cMap;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public ProvidingGenotype newInstance() {
		return new ProvidingGenotype(this, providerGenerator, cMap);
	}
}
