/*
 * Copyright (C) 2008 Google Inc.
 * Copyright (C) 2018 fortiss GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.graph;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;

import com.google.common.collect.Sets;
import com.google.inject.Binder;
import com.google.inject.Binding;
import com.google.inject.Key;
import com.google.inject.TypeLiteral;
import com.google.inject.spi.Dependency;
import com.google.inject.spi.ElementSource;
import com.google.inject.spi.HasDependencies;
import com.google.inject.spi.ProviderInstanceBinding;

/**
 * Walks the dependency graph from a given type that must have a Guice binding to obtain a list of
 * relevant dependent types. These types have to be given at construction time. All dependent types
 * bound in the graph that are a subtype of the given list is returned by
 * {@link #getFoundSubTypes()}.
 * This visitor allows to scan for artifacts provided by some provides method and for the classes
 * that provide artifacts (@Provides Methods).
 * This class is a modified version of the TransitiveDependencyVisitor class of the Grapher
 * extension of google Guice.
 *
 * @author phopkins@gmail.com (Pete Hopkins)
 * @author diewald@fortiss.org (Alexander Diewald)
 */
public class TypeMemorizingDependencyVisitor<T> extends DependencyVisitorBase {

	/**
	 * Set of found types to which a dependency exists and which is compatible with any of the
	 * initially given type-set.
	 */
	private Set<TypeLiteral<? extends T>> foundSubTypes = Sets.newHashSet();

	/** Type set to filter while iterating over the graph. */
	private Collection<Class<?>> seekTypes = new HashSet<>();

	/** Constructor. */
	public TypeMemorizingDependencyVisitor(Collection<Class<?>> seekTypes) {
		this.seekTypes = seekTypes;
	}

	/**
	 * Iterates over the dependency tree spanned by the {@link Binder} and remembers every
	 * encountered type binding if it is a sub-type of the initially given set of types.
	 */
	@Override
	protected Collection<Key<?>> visitHasDependencies(HasDependencies hasDependencies,
			Binding<?> binding) {
		Set<Key<?>> dependencies = Sets.newHashSet();

		for(Dependency<?> dependency : hasDependencies.getDependencies()) {
			Key<?> key = dependency.getKey();
			addIfArtefactSubType(key);
			dependencies.add(key);
		}

		return dependencies;
	}

	/**
	 * Adds the type wrapped by the given {@link Key} to the set of found types if it is a subtype
	 * of any of the initially given type set.
	 */
	@SuppressWarnings("unchecked")
	private void addIfProviderSubType(ProviderInstanceBinding<?> binding) {
		if(!(binding.getSource() instanceof ElementSource)) {
			return;
		}
		ElementSource source = (ElementSource)binding.getSource();
		Object declSource = source.getDeclaringSource();
		if(declSource instanceof Method) {
			Method declMethod = (Method)declSource;
			Class<?> providerClass = declMethod.getDeclaringClass();
			if(seekTypes.stream().anyMatch(c -> c.isAssignableFrom(providerClass))) {
				foundSubTypes
						.add((TypeLiteral<? extends T>)Key.get(providerClass).getTypeLiteral());
			}
		} else {
			// @CodeFormatterOff
			throw new ExplorationRuntimeException("Found a provider that in the decoder chain"
					+ " that does not use the Provides Methods mechanism. Only Provides Methods are"
					+ " supported for Decoders.");
			// @CodeFormatterOn
		}
	}

	/**
	 * Adds the type wrapped by the given {@link Key} to the set of found types if it is a subtype
	 * of any of the initially given type set.
	 */
	@SuppressWarnings("unchecked")
	private void addIfArtefactSubType(Key<?> key) {
		if(seekTypes.stream()
				.anyMatch(c -> c.isAssignableFrom(key.getTypeLiteral().getRawType()))) {
			foundSubTypes.add((TypeLiteral<? extends T>)key.getTypeLiteral());
		}
	}

	/** Visits a {@link ProviderInstanceBinding}. */
	@Override
	public Collection<Key<?>> visit(ProviderInstanceBinding<?> binding) {
		addIfProviderSubType(binding);
		return visitHasDependencies(binding, binding);
	}

	/**
	 * Returns the set of types that are subtypes of the initial set of given types and to which a
	 * dependency exits.
	 */
	public Set<TypeLiteral<? extends T>> getFoundSubTypes() {
		return foundSubTypes;
	}
}
