/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.af3.allocation.utils.AllocationUtils.getAnnotationEntry;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.exploration.moea.model.allocation.InterferenceLatencyEntry;
import org.fortiss.af3.exploration.moea.model.allocation.InterferenceLatencyTable;
import org.fortiss.af3.exploration.moea.model.annotation.InterferenceLatency;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.tooling.base.annotation.AnnotationEntry;

/**
 * Adapter between AF3 {@link ExecutionUnit} (=deployment target of
 * {@link ITaskAdapter}s) and the DSE.
 * 
 * @author barner
 */
public class AF3ExecutionUnitAdapter extends AF3ResourceAdapter<ExecutionUnit>
		implements IExecutionUnitAdapter {

	/** Additional interference latency introduced by this {@link IExecutionUnitAdapter}. */
	Map<ExecutionUnit, Double> interferenceLatency = new HashMap<>();

	/**
	 * Creates a {@link IExecutionUnitAdapter} for an {@link ExecutionUnit}.
	 * 
	 * @throws InvalidModelException
	 *             if a required annotation cannot be found.
	 */
	public AF3ExecutionUnitAdapter(ExecutionUnit executionUnit,
			Collection<ExecutionUnit> allExecUnits) {
		super(executionUnit);

		extractInterferenceLatencies(executionUnit, allExecUnits);
	}

	/**
	 * Extracts inter-core interference latencies for a table attached to the platform architecture.
	 */
	private void extractInterferenceLatencies(ExecutionUnit executionUnit,
			Collection<ExecutionUnit> allExecUnits) {
		// TODO(AD): HACK!!!
		PlatformArchitecture pArch =
				getFirstParentWithType(executionUnit, PlatformArchitecture.class);
		InterferenceLatencyTable ilTable =
				pickFirstInstanceOf(InterferenceLatencyTable.class, pArch.getSpecifications());
		if(ilTable == null) {
			return;
		}

		for(ExecutionUnit curExecUnit : allExecUnits) {
			if(curExecUnit != executionUnit) {
				AnnotationEntry aEntry = getAnnotationEntry(ilTable, executionUnit, curExecUnit,
						InterferenceLatencyEntry.class, InterferenceLatency.class);
				if(aEntry != null) {
					InterferenceLatency iLatency =
							aEntry.getSpecification(InterferenceLatency.class);
					interferenceLatency.put(curExecUnit, iLatency.getLatency().doubleValue());
				} else {
					interferenceLatency.put(curExecUnit, 0.0);
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public double getInterferenceLatency(IExecutionUnitAdapter execUnit) {
		Double latency = interferenceLatency.get(execUnit.getObject(ExecutionUnit.class));
		return (latency != null) ? latency : 0.0;
	}
}
