/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.tooling.base.utils.AnnotationUtils.getAnnotationValue;

import java.util.Collection;
import java.util.Iterator;

import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.platform.model.annotation.MessageSize;
import org.fortiss.af3.task.model.TaskInputPort;
import org.fortiss.af3.task.model.TaskOutputPort;

/**
 * Adapter between the AF3 {@link Channel} and the DSE.
 * 
 * @author barner
 */
public class AF3SignalAdapter extends AF3RequestAdapter<TaskOutputPort> implements ISignalAdapter {
	/** Conversion constant for the message size. */
	private static final double MBITS_TO_BITS = 1e6;

	/** Message's source component */
	private ITaskAdapter source;

	/** Message's target component adapter */
	private Collection<ITaskAdapter> targets;

	/** Message's target AF3 ports. */
	private Collection<TaskInputPort> targetPorts;

	/** Message size annotation */
	private int messageSize;

	/**
	 * Contains the {#link SignalType} of this {@link ISignalAdapter}. Defaults to {#link
	 * SignalType.PERIODIC}.
	 */
	SignalType signalType;

	/**
	 * Creates an {@link ISignalAdapter} for an {@link Channel}.
	 * 
	 * @throws InvalidModelException
	 *             if a required annotation cannot be found.
	 */
	public AF3SignalAdapter(TaskOutputPort senderPort, ITaskAdapter source,
			Collection<TaskInputPort> targetPorts, Collection<ITaskAdapter> targets)
			throws InvalidModelException {
		super(senderPort);
		this.source = source;
		this.targets = targets;
		this.targetPorts = targetPorts;

		// Max message size are actually bits...
		// TODO(AD) fix in the base model of AF3.
		messageSize = getAnnotationValue(modelElement, MessageSize.class, Integer.class);

		signalType = SignalType.PERIODIC;
	}

	/** {@inheritDoc} */
	@Override
	public ITaskAdapter getSource() {
		return source;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<ITaskAdapter> getTargets() {
		return targets;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<TaskInputPort> getTargetPorts() {
		return targetPorts;
	}

	/** {@inheritDoc} */
	@Override
	public long getSignalSize() {
		return messageSize;
	}

	/** {@inheritDoc} */
	@Override
	public double getTransmissionTime(ITransmissionUnitAdapter transmissionUnit) {
		return transmissionUnit.getTransmissionDuration(messageSize / MBITS_TO_BITS);
	}

	/** {@inheritDoc} */
	@Override
	public String getName() {
		String targetNameList = "";
		for(Iterator<ITaskAdapter> targetIterator = targets.iterator(); targetIterator.hasNext();) {
			targetNameList += targetIterator.next().getName();
			if(targetIterator.hasNext()) {
				targetNameList += ", ";
			}
		}
		return source.getName() + "->{" + targetNameList + "}";
	}

	/**
	 * Sets the {#link SignalType} of this {@link AF3SignalAdapter}.
	 */
	public void setSignalType(SignalType signalType) {
		this.signalType = signalType;
	}

	/** {@inheritDoc} */
	@Override
	public SignalType getSignalType() {
		return signalType;
	}
}
