/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.af3.exploration.alg.util.AF3Utils.getAnnotationChecked;
import static org.fortiss.af3.project.utils.ProjectUtils.getFileProject;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstChildWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.math.BigDecimal;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.Enumerator;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.exploration.model.project.DSE;
import org.fortiss.af3.exploration.moea.model.annotation.ReplicationBounds;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.af3.safety.model.annotation.SafetyIntegrityLevel;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.timing.TaskStartEvent;
import org.fortiss.af3.timing.model.TimingSpecification;
import org.fortiss.af3.timing.model.annotation.Wcet;
import org.fortiss.af3.timing.utils.TimingUtils;

/**
 * Adapter between the AF3 deployable {@link Component} (=tasks) and the DSE.
 * 
 * @author barner
 */
public class AF3TaskAdapter extends AF3RequestAdapter<Task> implements ITaskAdapter {

	/** Map of possible Wcets defined by the deployment target. */
	private Map<IExecutionUnitAdapter, Double> wcetMap = new HashMap<>();

	/**
	 * Holds the generic WCET value of the task that is used in case no WCET is defined for a
	 * queried execution unit.
	 */
	private Double genericWCET;

	/** Map of possible energy consumption defined by the deployment target. */
	private Map<IExecutionUnitAdapter, Double> energyMap;

	/** Minimum number of replica (1 == no replication, just the component itself). */
	private int minReplication;

	/** Maximum number of replica (1 == no replication, just the component itself). */
	private int maxReplication;

	/** Period annotation */
	private Double period;

	/** Marks an {@link ITaskAdapter} as strongly causal (snychronous). */
	private boolean stronglyCausal;

	/** SIL annotation */
	private SafetyIntegrityLevel sil;

	/** Reference to the corresponding elements of the diverse implementations. */
	private Collection<ITaskAdapter> replacementComponents;

	/**
	 * "Trace" to the {@link ITaskAdapter} which {@code this} {@link ITaskAdapter} replaced (if
	 * applicable).
	 */
	private ITaskAdapter replacedComponent;

	/**
	 * Creates an {@link ITaskAdapter} for a {@link Task}.
	 * 
	 * @throws InvalidModelException
	 *             if a required annotation cannot be found.
	 */
	public AF3TaskAdapter(Task af3Task, Collection<IComponentAdapter> components,
			Set<IExecutionUnitAdapter> execUnits) throws InvalidModelException {
		super(af3Task);

		// TODO(AD): Remove this HACK to retrieve the copied TimingSpecification. Instead, there
		// shall be a way to get the Timing Constraints by means of ExplorationConstraints.
		FileProject fp = getFileProject(af3Task);
		DSE dse = getFirstChildWithType(fp, DSE.class);

		// Timing
		TimingSpecification ts = getFirstChildWithType(dse, TimingSpecification.class);
		BigDecimal af3Period = TimingUtils.getPeriod(ts, af3Task, TaskStartEvent.class);
		if(af3Period == null || af3Period.doubleValue() <= 0.0) {
			throw new InvalidModelException(af3Task,
					"A non-specified period value and negative periods are not allowed.");
		}
		period = af3Period.doubleValue();
		Wcet wcet = pickFirstInstanceOf(Wcet.class, af3Task.getSpecifications());
		if(wcet.getValue() != null) {
			genericWCET = wcet.getValue().doubleValue();
		}
		for(IExecutionUnitAdapter execUnitAdp : execUnits) {
			BigDecimal exUWcet = af3Task.getWcet(execUnitAdp.getObject(ExecutionUnit.class));
			if(exUWcet != null && exUWcet.doubleValue() > 0.0) {
				wcetMap.put(execUnitAdp, exUWcet.doubleValue());
			}
		}
		// Ignore the absence of WCET information since it is only relevant for scheduling.

		// Safety & Reliability
		ReplicationBounds replicationSpec = getAnnotationChecked(af3Task, ReplicationBounds.class);
		minReplication = replicationSpec.getMinReplication();
		maxReplication = replicationSpec.getMaxReplication();

		// Properties inherited from components.
		SafetyIntegrityLevel tempSIL = null;
		for(IComponentAdapter comp : components) {
			// TODO: Move all this logic to the IComponentAdapter.
			Component af3Component = comp.getObject(Component.class);
			if(af3Component.isStronglyCausal()) {
				stronglyCausal = true;
			}

			tempSIL = getAnnotationChecked(af3Component, SafetyIntegrityLevel.class);
			if(sil == null) {
				sil = tempSIL;
			} else if(sil.getValue() != tempSIL.getValue()) {
				throw new InvalidModelException(af3Component,
						"Only components with the same SIL may be mapped to a Task.");
			}
		}

		replacementComponents = new HashSet<ITaskAdapter>();
	}

	/** {@inheritDoc} */
	@Override
	public Double getWcet(IExecutionUnitAdapter executionUnit) {
		Double wcet = wcetMap.get(executionUnit);
		return (wcet != null) ? wcet : genericWCET;
	}

	/** Returns the */
	private static long getDataSizeOf(Collection<ISignalAdapter> signals) {
		return signals.stream().map(s -> s.getSignalSize()).reduce((long)0, (a, b) -> a + b);
	}

	/** {@inheritDoc} */
	@Override
	public Double getEnergyConsumption(IExecutionUnitAdapter executionUnit) {
		return energyMap.get(executionUnit);
	}

	/** {@inheritDoc} */
	@Override
	public Enumerator getRequiredSafetyIntegrityLevel() {
		return sil.getValue();
	}

	/** {@inheritDoc} */
	@Override
	public Double getPeriod() {
		return period;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isStronglyCausal() {
		return stronglyCausal;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isInstantiationRequired() {
		return !replacementComponents.isEmpty();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isInstantiation() {
		return replacedComponent != null;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<ITaskAdapter> getReplacementComponents() {
		// TODO(#3264)
		// return replacementComponents;
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public ITaskAdapter getReplacedComponent() {
		// TODO(#3264)
		// return replacedComponent;
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void setReplacedComponent(ITaskAdapter abstractComponent) {
		replacedComponent = abstractComponent;
	}

	/** {@inheritDoc} */
	@Override
	public void updateInstatiationData() {
		replacedComponent = null;
	}

	/** {@inheritDoc} */
	@Override
	public int getMinReplication() {
		return minReplication;
	}

	/** {@inheritDoc} */
	@Override
	public int getMaxReplication() {
		return maxReplication;
	}

	/** {@inheritDoc} */
	@Override
	public AF3TaskAdapter clone() {
		try {
			return (AF3TaskAdapter)super.clone();
		} catch(CloneNotSupportedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Component getAssignedComponent() {
		return null;
	}
}
