/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ILogicalArchitectureAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DirectedWeightedMultigraph;

/**
 * Creates a Directed acyclic task graph of all the {@link ITaskAdapter}s.
 * <li>Each
 * {@link ISignalAdapter} is examined:</li>
 * <li>If there exists no path between a target of the
 * {@link ISignalAdapter} and it source, add an edge to the {@link DefaultTaskGraph}.</li>
 * <li>
 * Otherwise, a loop is detected. Then, the detected loop is cut either by not adding the edge (if
 * the source vertex is strongly causal) or by removing an edge between the vertices (i.e.
 * {@link ITaskAdapter}s) that are closest to the current {@link ISignalAdapter} by traversing the
 * found path backwards and adding the current edge.</li>
 * <p>
 * <b>NOTE:</b> This method expects <b>correct</b> {@link ILogicalArchitectureAdapter}s as input,
 * i.e. at least one {@link ITaskAdapter} of each loop must be 'strongly causal'.
 * 
 * @author diewald
 */
public class AcyclicTaskGraph extends DefaultTaskGraph<ITaskAdapter, AcyclicTaskGraph> {

	/** Constructor. */
	public AcyclicTaskGraph() {
		super();
	}

	/**
	 * Constructor which creates an acyclic {@link ITaskGraph} from a given possibly cyclic
	 * {@link ITaskGraph}.
	 */
	public <T extends ITaskGraph<?>> AcyclicTaskGraph(T taskGraph) {
		this();
		hyperPeriod = taskGraph.getHyperPeriod();
		createAcyclicGraph(taskGraph);
	}

	/**
	 * Creates an acyclic {@link ITaskGraph} from a given general {@link ITaskGraph} based on a
	 * calculation of the minimum spanning tree of the transitions between tasks, i.e.
	 * {@link SignalAdapterWeightedEdge}s.
	 */
	private <T extends ITaskGraph<?>> void createAcyclicGraph(T taskGraph) {
		DirectedWeightedMultigraph<ITaskAdapter, SignalAdapterWeightedEdge> tg =
				(DirectedWeightedMultigraph<ITaskAdapter, SignalAdapterWeightedEdge>)taskGraph
						.getGraph();
		Graphs.addAllVertices(graph, tg.vertexSet());

		for(SignalAdapterWeightedEdge edge : tg.edgeSet()) {
			ITaskAdapter senderTask = tg.getEdgeSource(edge);
			if(!senderTask.isStronglyCausal()) {
				ITaskAdapter receiverTask = tg.getEdgeTarget(edge);
				graph.addEdge(senderTask, receiverTask,
						new SignalAdapterWeightedEdge(edge.getSignalAdapter()));
			}
		}
	}
}
