/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import static java.util.Collections.unmodifiableSet;
import static java.util.stream.Collectors.toSet;
import static org.fortiss.af3.exploration.alg.util.GraphUtils.getAllPredescesorsOf;
import static org.fortiss.af3.exploration.alg.util.GraphUtils.getAllSuccessorsOf;
import static org.fortiss.af3.exploration.alg.util.GraphUtils.replaceVertex;
import static org.fortiss.af3.schedule.utils.MathUtils.gcdDoubleCollection;
import static org.jgrapht.Graphs.predecessorListOf;
import static org.jgrapht.Graphs.successorListOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.fortiss.af3.exploration.alg.display.JGraphTVisualizer;
import org.fortiss.af3.exploration.alg.dse.RandomExploration;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.jgrapht.Graph;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DirectedWeightedMultigraph;
import org.jgrapht.traverse.DepthFirstIterator;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * <p>
 * The TaskGraph represents the data dependencies of a single connected component between the
 * deployable components (= tasks) of a systems (Tasks = vertices, channels of logical architecture
 * = edges). Therefore, a system typically consists of multiple {@link DefaultTaskGraph}s.
 * </p>
 * The TaskGraph also contains
 * <ul>
 * <li>Execution times of tasks.</li>
 * <li>The fault-detectors used by the tasks</li>
 * <li>Task execution times (currently assuming homogeneous cores)</li>
 * </ul>
 * 
 * @author diewald
 * @author huang
 */

public abstract class DefaultTaskGraph<T extends ITaskAdapter, G extends DefaultTaskGraph<T, G>>
		implements ITaskGraph<T>, Cloneable {
	/** Underlying graph implementation */
	// TODO: Add a weight w.r.t. transmission time, when deployed
	protected DirectedWeightedMultigraph<T, SignalAdapterWeightedEdge> graph;

	// TODO: Currently, all tasks in a {@link TaskGraph} must have the same period
	/** The period time of this {@link DefaultTaskGraph}. */
	protected Double hyperPeriod;

	// TODO: Fault detector library
	/** Fault detector library (map: task -> set of fault detectors) */
	protected Map<ITaskAdapter, SortedSet<FaultDetector>> taskFaultDetectorsMap;

	/** Random generator */
	private RandomExploration gen = RandomExploration.getInstance();

	/**
	 * Associates "abstract" {@link ITaskAdapter}s with its instances contained in
	 * this {@link ITaskGraph}.
	 */
	private Multimap<ITaskAdapter, ITaskAdapter> abstractToInstanceTaskMap;

	/** TEMPORAL PRECISION. */
	public static int TEMP_PRECISION_EXP = 9;

	/** Creates an empty {@link DefaultTaskGraph}. */
	public DefaultTaskGraph() {
		graph = new DirectedWeightedMultigraph<>(SignalAdapterWeightedEdge.class);

		taskFaultDetectorsMap = new HashMap<ITaskAdapter, SortedSet<FaultDetector>>();

		abstractToInstanceTaskMap = HashMultimap.create();
	}

	/** Registers a deployable component (=task) with the {@link DefaultTaskGraph}. */
	@Override
	public boolean addTask(T task) {
		// TODO(#3264)
		// if(task.isInstantiation()) {
		// abstractToInstanceTaskMap.put(task.getReplacedComponent(), task);
		// }
		// // FIXME: Hack, remove the map.
		// abstractToInstanceTaskMap.put(task, task);
		return graph.addVertex(task);
	}

	/** {@inheritDoc} */
	@Override
	public boolean addTask(T task, T refTask) {
		// FIXME: Hack, remove the map.
		abstractToInstanceTaskMap.put(refTask, task);
		return graph.addVertex(task);
	}

	/** Registers a deployable component (=task) with the {@link DefaultTaskGraph}. */
	@Override
	public boolean addAllTask(Collection<T> tasks) {
		return Graphs.addAllVertices(graph, tasks);
	}

	/** {@inheritDoc} */
	@Override
	public ITaskAdapter replaceTask(T task, T replacement) {
		replaceVertex(graph, task, replacement);
		abstractToInstanceTaskMap.removeAll(task);
		return task;
	}

	/** {@inheritDoc} */
	@Override
	public void removeTask(ITaskAdapter task) {
		graph.removeVertex((T)task);

		// Remove from the cache.
		abstractToInstanceTaskMap.removeAll(task);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public boolean addSignal(ISignalAdapter signalAdapter) {
		for(ITaskAdapter recvTask : signalAdapter.getTargets()) {
			if(!graph.addEdge((T)signalAdapter.getSource(), (T)recvTask,
					new SignalAdapterWeightedEdge(signalAdapter))) {
				return false;
			}
		}
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean addAllSignals(Collection<ISignalAdapter> signalAdapters) {
		for(ISignalAdapter signal : signalAdapters) {
			if(!addSignal(signal)) {
				return false;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public Set<T> getTasks() {
		return graph.vertexSet();
	}

	/** {@inheritDoc} */
	@Override
	public Set<T> getSinkTasks(T sourceTask) {
		return unmodifiableSet(new HashSet<>(successorListOf(graph, sourceTask)));
	}

	/** {@inheritDoc} */
	@Override
	public Set<T> getTransientSinkTasks(T sourceTask) {
		return getAllSuccessorsOf(graph, sourceTask);
	}

	/** {@inheritDoc} */
	@Override
	public Set<T> getSourceTasks(T sinkTask) {
		return unmodifiableSet(new HashSet<>(predecessorListOf(graph, sinkTask)));
	}

	/** {@inheritDoc} */
	@Override
	public Set<T> getTransientSourceTasks(T sinkTask) {
		return getAllPredescesorsOf(graph, sinkTask);
	}

	/** {@inheritDoc} */
	@Override
	public Set<SignalAdapterWeightedEdge> getSignals() {
		return graph.edgeSet();
	}

	/** {@inheritDoc} */
	@Override
	public Set<ISignalAdapter> getSinkSignals(T sourceTask) {
		return graph.outgoingEdgesOf(sourceTask).stream().map(s -> s.getSignalAdapter())
				.collect(toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Set<ISignalAdapter> getSourceSignals(T sinkTask) {
		return graph.incomingEdgesOf(sinkTask).stream().map(s -> s.getSignalAdapter())
				.collect(toSet());
	}

	// TODO: Get rid of this method!
	/** {@inheritDoc} */
	@Override
	public DirectedWeightedMultigraph<T, SignalAdapterWeightedEdge> getGraph() {
		return graph;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public DirectedWeightedMultigraph<ITaskAdapter, SignalAdapterWeightedEdge> getGenericGraph() {
		return (DirectedWeightedMultigraph<ITaskAdapter, SignalAdapterWeightedEdge>)graph;
	}

	/** {@inheritDoc} */
	@Override
	public Double getHyperPeriod() {
		return hyperPeriod;
	}

	/** {@inheritDoc} */
	@Override
	public void updateHyperPeriod() {
		Collection<Double> periodList = new ArrayList<>();
		getTasks().forEach(t -> periodList.add(t.getPeriod()));
		hyperPeriod = gcdDoubleCollection(periodList, TEMP_PRECISION_EXP);
	}

	/** Returns the fault detector with a given index / ID for a given task. */
	@Override
	public FaultDetector getFaultDetector(ITaskAdapter task, int id) {
		if(taskFaultDetectorsMap.containsKey(task)) {
			Object[] faultDetectors = taskFaultDetectorsMap.get(task).toArray();
			if(id < faultDetectors.length) {
				return (FaultDetector)faultDetectors[id];
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void addFaultDetector(ITaskAdapter task, FaultDetector faultDetector) {
		if(!taskFaultDetectorsMap.containsKey(task)) {
			SortedSet<FaultDetector> faultDetectors = new TreeSet<FaultDetector>();
			taskFaultDetectorsMap.put(task, faultDetectors);
		}
		taskFaultDetectorsMap.get(task).add(faultDetector);
	}

	/** {@inheritDoc} */
	@Override
	public int getRandomFaultDetectorId(ITaskAdapter task) {
		if(taskFaultDetectorsMap.get(task) == null)
			return 0;
		return gen.nextInt(taskFaultDetectorsMap.get(task).size());
	}

	/** {@inheritDoc} */
	@Override
	public void display() {
		@SuppressWarnings("unchecked") JGraphTVisualizer<ITaskAdapter, SignalAdapterWeightedEdge> taskViz =
				new JGraphTVisualizer<ITaskAdapter, SignalAdapterWeightedEdge>(
						(Graph<ITaskAdapter, SignalAdapterWeightedEdge>)this.getGraph());
		taskViz.displayGraph("Task graph");
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String rval = "";

		Iterator<T> iter = new DepthFirstIterator<T, SignalAdapterWeightedEdge>(graph);
		while(iter.hasNext()) {
			T task = iter.next();
			rval = rval.concat("Task " + task.getName() + ": ");
			for(SignalAdapterWeightedEdge edge : graph.outgoingEdgesOf(task)) {
				rval = rval.concat("->" + graph.getEdgeTarget(edge).getName());
			}
			rval = rval.concat("\n");
		}

		return rval;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public G copy() {
		G clonedTaskGraph;
		try {
			clonedTaskGraph = (G)super.clone();
		} catch(CloneNotSupportedException e) {
			throw new AssertionError();
		}
		clonedTaskGraph.graph = new DirectedWeightedMultigraph<T, SignalAdapterWeightedEdge>(
				SignalAdapterWeightedEdge.class);
		Graphs.addGraph(clonedTaskGraph.graph, graph);
		clonedTaskGraph.taskFaultDetectorsMap =
				new HashMap<ITaskAdapter, SortedSet<FaultDetector>>(taskFaultDetectorsMap);

		return clonedTaskGraph;
	}

	/** {@inheritDoc} */
	@Override
	public boolean containsTask(ITaskAdapter task) {
		if(task.isInstantiationRequired()) {
			// FIXME: Hack to get the DSE working again. To fix it, replace the calls that need the
			// instantiated component from an abstaract component, fix at the coller side.
			return abstractToInstanceTaskMap.containsKey(task) ? true
					: graph.containsVertex((T)task);
		}
		return graph.containsVertex((T)task);
	}

	/** {@inheritDoc} */
	@Override
	public boolean containsAnyTaskOf(Collection<ITaskAdapter> taskSet) {
		for(ITaskAdapter task : taskSet) {
			if(containsTask(task)) {
				return true;
			}
		}
		return false;
	}
}
