/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature;

import static java.util.stream.Collectors.toSet;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.module.CreatorModule;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.alg.service.module.IExplorationTransformationModule;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.operators.OperatorModule;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.MutableClassToInstanceMap;
import com.google.inject.Module;
import com.google.inject.TypeLiteral;

/**
 * Implementation of the {@link IExplorationFeatureService}.
 * 
 * @author diewald
 */
public abstract class ExplorationFeature implements IExplorationFeature {

	/** Name of the {@link IExplorationFeature}. */
	private String name;

	/** Set referencing all registered candidate solution types. */
	private Set<TypeLiteral<? extends Phenotype>> phenotypes = new HashSet<>();

	/** Map associating Operators with the {@link Genotype}s they modify. */
	private Map<TypeLiteral<? extends Genotype>, CreatorModule<?>> creatorGenoMap = new HashMap<>();

	/** Map associating {@link OperatorModule}s with the {@link Genotype}s they modify. */
	private Multimap<TypeLiteral<? extends Genotype>, OperatorModule<?>> operatorGenoMap =
			HashMultimap.create();

	/** References the set of transformers associated with this {@link IExplorationFeature}. */
	private Map<TypeLiteral<SuperSet<? extends IModelElement>>, IExplorationTransformationModule<? extends SuperSet<? extends IModelElement>>> outputTransformers =
			new HashMap<>();

	/**
	 * References all registered exploration modules. Differing between {@link Creator}s,
	 * {@link Decoder}s etc. is currently not needed.
	 */
	private ClassToInstanceMap<Module> registeredExpModules = MutableClassToInstanceMap.create();

	/**
	 * Internal list of {@link Module}s from other {@link IExplorationFeature}s overridden by this
	 * one.
	 */
	private Collection<Class<? extends Module>> moduleOverrides = new ArrayList<>();

	/** Set of {@link InputModule}s provided by this {@link IExplorationFeature}. */
	private Map<TypeLiteral<?>, InputModule<?>> inputModules = new HashMap<>();

	/** Constructor. */
	public ExplorationFeature(String name) {
		this.name = name;
	}

	/** Adds the given {@link InputModule}s to this {@link IExplorationFeature}. */
	public void addInputModule(InputModule<?> inModule) {
		inputModules.put(inModule.getProvidedType(), inModule);
	}

	/**
	 * Registers an optimization variable ({@code genotype}) with the DSE, along with its
	 * initializer ({@code creator}) and modificator ({@code operator}).
	 * 
	 * @param genotype
	 *            optimization variable type.
	 * @param creator
	 *            initializes the opt. variable.
	 * @param operators
	 *            modifications applied during the process iterations.
	 */
	public <G extends Genotype> void addVariable(TypeLiteral<G> genotype, CreatorModule<G> creator,
			Collection<OperatorModule<?>> operators) {
		// Input check: Allow only fully specified variable registrations.
		if(genotype == null) {
			throw new RuntimeException(getClass() + ": The passed genotype is null!");
		} else if(creator == null) {
			throw new RuntimeException(getClass() + ": The passed creator is null!");
		} else if(operators == null || operators.isEmpty()) {
			throw new RuntimeException(getClass() + ": The passed operator set is empty or null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<CreatorModule<G>> creatorClass =
				(Class<CreatorModule<G>>)creator.getClass();
		if(registeredExpModules.getInstance(creatorClass) != null) {
			throw new RuntimeException(getClass() + ": The module " + creatorClass.getSimpleName() +
					" has already been registered!");
		}
		creatorGenoMap.put(genotype, creator);
		registeredExpModules.putInstance(creatorClass, creator);

		for(OperatorModule<?> operator : operators) {
			@SuppressWarnings("unchecked") Class<OperatorModule<?>> operatorClass =
					(Class<OperatorModule<?>>)operator.getClass();
			if(registeredExpModules.getInstance(operatorClass) != null) {
				throw new RuntimeException(getClass() + ": The module " +
						operatorClass.getSimpleName() + " has already been registered!");
			}
			operatorGenoMap.put(genotype, operator);
			registeredExpModules.putInstance(operatorClass, operator);
		}
	}

	/**
	 * Registers a problem definition with the DSE process. The {@code phenotype} is equivalent to
	 * a candidate solution that is produced by the given {@code decoder}.
	 * 
	 * @param phenotype
	 *            candidate solution type.
	 * @param decoder
	 *            transforms variables (and/or other inputs) to candidate solutions.
	 */
	public <P extends Phenotype> void addProblem(TypeLiteral<P> phenotype,
			DecoderModule<P> decoder) {
		// Input check: Allow only fully specified variable registrations.
		if(phenotype == null) {
			throw new RuntimeException(getClass() + ": The passed phenotype is null!");
		} else if(decoder == null) {
			throw new RuntimeException(getClass() + ": The passed decoder is null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<DecoderModule<P>> decoderClass =
				(Class<DecoderModule<P>>)decoder.getClass();
		if(registeredExpModules.getInstance(decoderClass) != null) {
			throw new RuntimeException(getClass() + ": The module " + decoderClass.getSimpleName() +
					" has already been registered!");
		}

		phenotypes.add(phenotype);
		registeredExpModules.putInstance(decoderClass, decoder);
	}

	/**
	 * Adds the given {@link Module} to the list of overridden {@link Module}s from other
	 * {@link IExplorationFeature}s. This information is used when aggregating the list of active
	 * Module injected into the problem.
	 */
	public void addModuleOverride(Class<? extends Module> overriddenModule) {
		moduleOverrides.add(overriddenModule);
	}

	/**
	 * Registers the given {@code evaluator} to quantifies the quality of a solution by some metric.
	 * 
	 * @param evaluator
	 *            applies some metric to a candidate solution.
	 */
	public void addEvaluator(EvaluatorModule<?, ?, ?> evaluator) {
		// Input check: Allow only fully specified variable registrations.
		if(evaluator == null) {
			throw new RuntimeException(getClass() + ": The passed evaluator is null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<EvaluatorModule<?, ?, ?>> evaluatorClass =
				(Class<EvaluatorModule<?, ?, ?>>)evaluator.getClass();
		if(registeredExpModules.getInstance(evaluatorClass) != null) {
			throw new RuntimeException(getClass() + ": The module " +
					evaluatorClass.getSimpleName() + " has already been registered!");
		}

		registeredExpModules.putInstance(evaluatorClass, evaluator);
	}

	/**
	 * Adds the given {@link IExplorationTransformationModule} to the transformers of this
	 * {@link ExplorationFeature}
	 */
	@SuppressWarnings("unchecked")
	public void addTransformator(
			IExplorationTransformationModule<? extends SuperSet<? extends IModelElement>> tModule) {
		// Input check: Allow only "correct" transformers.
		if(tModule == null) {
			throw new RuntimeException(getClass() + ": The passed transformation module is null!");
		}
		outputTransformers.put(
				(TypeLiteral<SuperSet<? extends IModelElement>>)tModule.getProvidedType(), tModule);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Set<TypeLiteral<? extends Genotype>> getGenotypes() {
		return creatorGenoMap.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public Set<CreatorModule<?>> getCreators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof CreatorModule)
				.map(CreatorModule.class::cast).collect(toSet());
	}

	/** {@inheritDoc} */
	@Override
	public CreatorModule<?> getCreatorOf(TypeLiteral<? extends Genotype> genotype) {
		return creatorGenoMap.get(genotype);
	}

	/** {@inheritDoc} */
	@Override
	public Set<OperatorModule<?>> getOperators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof OperatorModule)
				.map(OperatorModule.class::cast).collect(toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Collection<OperatorModule<?>> getOperatorsOf(TypeLiteral<? extends Genotype> genotype) {
		return operatorGenoMap.get(genotype);
	}

	/** {@inheritDoc} */
	@Override
	public Set<DecoderModule<?>> getDecoders() {
		return registeredExpModules.values().stream().filter(m -> m instanceof DecoderModule)
				.map(DecoderModule.class::cast).collect(toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Set<EvaluatorModule<?, ?, ?>> getEvaluators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof EvaluatorModule)
				.map(EvaluatorModule.class::cast).collect(toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends Module>> getModuleOverrides() {
		return moduleOverrides;
	}

	/** {@inheritDoc} */
	@Override
	public Map<TypeLiteral<?>, InputModule<?>> getInputModules() {
		return inputModules;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<TypeLiteral<SuperSet<? extends IModelElement>>> getSolutionTypes() {
		return outputTransformers.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IExplorationTransformationModule<? extends SuperSet<? extends IModelElement>>>
			getTransformationModules() {
		return outputTransformers.values();
	}

	/** {@inheritDoc} */
	@Override
	public String getName() {
		return name;
	}
}
