/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform;

import static java.util.Arrays.asList;

import org.fortiss.af3.exploration.alg.feature.ExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.platform.evaluator.PlatformMappingEvaluatorConstraint;
import org.fortiss.af3.exploration.alg.feature.platform.evaluator.PlatformMappingEvaluatorObjective;
import org.fortiss.af3.exploration.alg.feature.platform.input.PlatformExecUnitInstanceNumberInputModule;
import org.fortiss.af3.exploration.alg.feature.platform.input.PlatformExecUnitParentTemplateMapping;
import org.fortiss.af3.exploration.alg.feature.platform.problem.PlatformCommunicationGraphExecUnitInstanceDecoder;
import org.fortiss.af3.exploration.alg.feature.platform.problem.PlatformExecUnitInstanceIdentityDecoder;
import org.fortiss.af3.exploration.alg.feature.platform.problem.PlatformExecUnitSetDecoder;
import org.fortiss.af3.exploration.alg.feature.platform.transform.PlatformConnectorTransformer;
import org.fortiss.af3.exploration.alg.feature.platform.transform.PlatformExecutionUnitTransformer;
import org.fortiss.af3.exploration.alg.feature.platform.transform.PlatformTransmissionConnectionTransformer;
import org.fortiss.af3.exploration.alg.feature.platform.variable.PlatformExecUnitInstanceCreator;
import org.fortiss.af3.exploration.alg.feature.platform.variable.PlatformExecUnitInstanceEncoding;
import org.fortiss.af3.exploration.alg.feature.platform.variable.PlatformExecUnitInstanceMutateModule;
import org.fortiss.af3.exploration.alg.feature.taskmapping.ExecutionUnitSet;
import org.fortiss.af3.exploration.alg.feature.taskmapping.PlatformExecUnitSetProvider;
import org.fortiss.af3.exploration.alg.module.common.evaluator.mapping.MappingEvaluatorConstraint;
import org.fortiss.af3.exploration.alg.module.common.evaluator.mapping.MappingEvaluatorObjective;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.smt.model.SMTConstraint;
import org.fortiss.af3.exploration.smt.model.SMTObjective;
import org.opt4j.core.DoubleValue;

import com.google.inject.Injector;
import com.google.inject.TypeLiteral;

/**
 * D3SE {@link IExplorationFeature} that explores the number of optimal CPUs/Cores to be used as
 * execution targets of the input task architecture (ref application).
 * 
 * @author diewald
 */
public class PlatformExploration extends ExplorationFeature {

	/** Constructor. */
	public PlatformExploration() {
		super("Target Platform Exploration");

		addInputModule(new PlatformExecUnitParentTemplateMapping());
		// TODO(AD): Remove the module below. It is a hack due to missing information from the
		// system model.
		addInputModule(new PlatformExecUnitInstanceNumberInputModule());

		addVariable(TypeLiteral.get(PlatformExecUnitInstanceEncoding.class),
				new PlatformExecUnitInstanceCreator(),
				asList(new PlatformExecUnitInstanceMutateModule()));

		addProblem(TypeLiteral.get(PlatformCommunicationGraphEncoding.class),
				new PlatformCommunicationGraphExecUnitInstanceDecoder());
		addProblem(TypeLiteral.get(ExecutionUnitSet.class), new PlatformExecUnitSetDecoder());
		addProblem(TypeLiteral.get(PlatformExecUnitInstanceEncoding.class),
				new PlatformExecUnitInstanceIdentityDecoder());
		addModuleOverride(PlatformExecUnitSetProvider.class);

		addEvaluator(new PlatformMappingEvaluatorConstraint(SMTConstraint.class));
		addEvaluator(new PlatformMappingEvaluatorObjective(
				(Class<? extends ExplorationTarget<DoubleValue>>)SMTObjective.class));
		addModuleOverride(MappingEvaluatorConstraint.class);
		addModuleOverride(MappingEvaluatorObjective.class);

		addTransformator(new PlatformExecutionUnitTransformer());
		addTransformator(new PlatformTransmissionConnectionTransformer());
		addTransformator(new PlatformConnectorTransformer());
	}

	/** {@inheritDoc} */
	@Override
	public boolean isApplicableToInputs(Injector dseInjector) {
		return true;
	}
}
