/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform;

import static org.fortiss.af3.exploration.alg.util.GraphUtils.getSubGraph;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.function.Predicate;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.dseml.model.expression.IExpression;
import org.fortiss.af3.exploration.dseml.model.expression.Set;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.exploration.moea.model.dsmlext.DsmlextPackage;
import org.fortiss.af3.exploration.moea.model.dsmlext.FixedPropertyLiteral;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.jgrapht.Graph;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Table;
import com.google.common.collect.Table.Cell;

/**
 * Utility methods for the {@link PlatformExploration} {@link IExplorationFeature}.
 * 
 * @author diewald
 */
public class PlatformExplorationUtils {

	/** Hiding constructor. */
	private PlatformExplorationUtils() {
		// Prevent instantiation.
	}

	/**
	 * Creates a copy of the input {@link IExpression} and updates its corresponding
	 * {@link SuperSet} (copies) to include instances instead of abstract elements. Thereby new
	 * {@link SuperSet}s are created and the references to them get updated in the expression.
	 * 
	 * @param expression
	 *            Source {@link IExpression} to update.
	 * @param setElementUpdates
	 *            {@link Table} that associates templates (rows) with and its contained template
	 *            elements (columns) with their instances.
	 * @param templateMap
	 *            Associates the parent element with its contained templates.
	 * @return the modified {@link IExpression}.
	 */
	public synchronized static IExpression updateTargetResourceSet(IExpression expression,
			Table<IExecutionUnitAdapter, IResourceAdapter, IResourceAdapter> setElementUpdates,
			Map<IExecutionUnitAdapter, IExecutionUnitAdapter> templateMap) {
		IExpression modExpression = copy(expression);

		@SuppressWarnings("unchecked") Collection<Set<IModelElement>> allSets =
				(Collection<Set<IModelElement>>)(Collection<?>)getChildrenWithType(modExpression,
						Set.class);

		Map<Class<? extends IModelElement>, SuperSet<IModelElement>> allRefSets = new HashMap<>();
		for(Set<IModelElement> set : allSets) {
			if(!allRefSets.containsKey(set.getEntryType())) {
				allRefSets.put(set.getEntryType(), set.getSuperSetReference());
			}
		}

		// Create copies of each required superset: They need to be modified.
		Map<Class<? extends IModelElement>, SuperSet<IModelElement>> reqRefSetCopies =
				new HashMap<>();
		Map<SuperSet<IModelElement>, SuperSet<IModelElement>> origToCopyMap = new HashMap<>();
		for(IResourceAdapter abstractElement : setElementUpdates.columnKeySet()) {
			Class<? extends IModelElement> refType =
					abstractElement.getObject(IModelElement.class).getClass();
			Optional<Class<? extends IModelElement>> superSetKey =
					getFirst(allRefSets.keySet(), k -> k.isAssignableFrom(refType));

			if(!superSetKey.isPresent()) {
				continue;
			}

			if(!reqRefSetCopies.containsKey(refType)) {
				SuperSet<IModelElement> superSet = allRefSets.get(superSetKey.get());
				SuperSet<IModelElement> copiedSuperSet = copy(superSet);
				reqRefSetCopies.put(superSetKey.get(), copiedSuperSet);
				origToCopyMap.put(superSet, copiedSuperSet);
			}
		}

		// Replace abstract entries with their instances.
		Multimap<SuperSet<IModelElement>, IModelElement> ssToTemplateMap = HashMultimap.create();
		Multimap<IModelElement, IModelElement> templateToInstancesMap = HashMultimap.create();
		for(Cell<IExecutionUnitAdapter, IResourceAdapter, IResourceAdapter> templateInstCell : setElementUpdates
				.cellSet()) {
			IResourceAdapter templateElementAdp = templateInstCell.getColumnKey();
			Class<?> refType = templateElementAdp.getObject(Object.class).getClass();
			Optional<Class<?>> superSetKey =
					getFirst(reqRefSetCopies.keySet(), k -> k.isAssignableFrom(refType));

			// Skip instantiation if no required set refers to the resource type.
			if(!superSetKey.isPresent()) {
				continue;
			}

			// NOTE: We do not need to test if the SuperSet contains the template: SuperSets define
			// all elements of a type from which the IResourceAdapters are derived.
			SuperSet<IModelElement> superSet = reqRefSetCopies.get(superSetKey.get());
			IModelElement templateElement = templateElementAdp.getObject(IModelElement.class);
			// Remove the abstract entry such that only instances will be placed in the SuperSet.
			superSet.getEntries().remove(templateElement);
			ssToTemplateMap.put(superSet, templateElement);

			// Create a new AF3 element. This will no longer be needed when the Adapters inherit
			// from AF3 model elements.
			IResourceAdapter castedInstElement = templateInstCell.getValue();
			IModelElement instCopy = copy(castedInstElement.getObject(IModelElement.class));
			castedInstElement.setObject(instCopy, IModelElement.class);
			superSet.getEntries().add(instCopy);
			templateToInstancesMap.put(templateElement, instCopy);
		}

		// Update all set references to the newly created SuperSets.
		Collection<Set<IModelElement>> updatedSets = new HashSet<>();
		for(Set<IModelElement> set : allSets) {
			SuperSet<IModelElement> ssRef = set.getSuperSetReference();
			if(origToCopyMap.keySet().contains(ssRef)) {
				SuperSet<IModelElement> updatedSS = origToCopyMap.get(ssRef);
				set.setSuperSetReference(updatedSS);
				for(IModelElement templateElement : ssToTemplateMap.get(updatedSS)) {
					if(set.getEntries().contains(templateElement)) {
						set.getEntries().remove(templateElement);
						set.getEntries().addAll(templateToInstancesMap.get(templateElement));
						updatedSets.add(set);
					}
				}
			}
		}

		// Update the FixedPropertyLiteral to include the new elements in their element --> value
		// map.
		@SuppressWarnings("unchecked") Collection<FixedPropertyLiteral<IModelElement>> fPropLits =
				(Collection<FixedPropertyLiteral<IModelElement>>)(Collection<?>)getChildrenWithType(
						modExpression, FixedPropertyLiteral.class);
		updateFixedPropLits(fPropLits, ssToTemplateMap, templateToInstancesMap, updatedSets);

		return modExpression;
	}

	/**
	 * Updates the {@link FixedPropertyLiteral}s such that they reference the created instances
	 * instead of the template elements.
	 */
	private static void updateFixedPropLits(
			Collection<FixedPropertyLiteral<IModelElement>> fPropLits,
			Multimap<SuperSet<IModelElement>, IModelElement> ssToTemplateMap,
			Multimap<IModelElement, IModelElement> templateToInstancesMap,
			Collection<Set<IModelElement>> updatedSets) {
		for(FixedPropertyLiteral<IModelElement> propLit : fPropLits) {
			Set<IModelElement> propSet = propLit.getSetReference();
			if(updatedSets.contains(propSet)) {
				SuperSet<IModelElement> updatedSS = propSet.getSuperSetReference();

				Map<IModelElement, Object> updatedMap = new HashMap<>();
				for(IModelElement templateElement : ssToTemplateMap.get(updatedSS)) {
					Object value = propLit.getValue(templateElement);
					Collection<IModelElement> instances =
							templateToInstancesMap.get(templateElement);
					instances.forEach(i -> updatedMap.put(i, value));
				}

				propLit.eSet(DsmlextPackage.Literals.FIXED_PROPERTY_LITERAL__MAP, updatedMap);
			}
		}
	}

	/**
	 * Returns a sub {@link Graph} that only contains instantiated {@link IResourceAdapter}s of the
	 * same platform layer and the edges between these elements.
	 * 
	 * @param pcgEnc
	 *            communication graph of the platform.
	 * @param parentExecUnitAdp
	 *            parent {@link IExecutionUnitAdapter} of the instantiated elements.
	 * @param instExecUnitAdp
	 *            instantiated {@link IExecutionUnitAdapter}.
	 * @return see above.
	 */
	public static DefaultDirectedGraph<IResourceAdapter, DefaultEdge> getElementInstanceSubGraph(
			PlatformCommunicationGraphEncoding pcgEnc, IExecutionUnitAdapter parentExecUnitAdp,
			IExecutionUnitAdapter instExecUnitAdp) {
		DefaultDirectedGraph<IResourceAdapter, DefaultEdge> subGraph =
				new DefaultDirectedGraph<>(DefaultEdge.class);
		Predicate<IResourceAdapter> irrelevantResource = v -> {
			boolean isTransmissionUnit = v instanceof TransmissionUnit;
			boolean isSameLevel = v.getObject(EObject.class)
					.eContainer() == (parentExecUnitAdp.getObject(EObject.class));
			return isTransmissionUnit && !isSameLevel;
		};
		getSubGraph(pcgEnc.getActualGraph(), instExecUnitAdp, irrelevantResource, subGraph);

		return subGraph;
	}
}
