/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform.problem;

import static java.util.stream.Collectors.toSet;
import static org.fortiss.af3.exploration.alg.util.GraphUtils.getAllEdgesOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Genotyped;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformExecUnitTemplateEncoding;
import org.fortiss.af3.exploration.alg.feature.platform.variable.PlatformExecUnitInstanceEncoding;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.InputArtifact;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.opt4j.core.Genotype;

import com.google.inject.Provides;

/**
 * Takes a {@link PlatformCommunicationGraphEncoding} genotype and provides it as its
 * {@link Phenotype} representation.
 * 
 * @author diewald
 */
public class PlatformCommunicationGraphExecUnitInstanceDecoder
		extends DecoderModule<PlatformCommunicationGraphEncoding> {

	/** See {@link PlatformCommunicationGraphExecUnitInstanceDecoder}. */
	@Decodes
	@Provides
	public PlatformCommunicationGraphEncoding decode(
			@InputArtifact PlatformCommunicationGraphEncoding pcgEnc,
			@Genotyped PlatformExecUnitInstanceEncoding peuiEnc,
			PlatformExecUnitTemplateEncoding execUnitTemplateEnc) {
		PlatformCommunicationGraphEncoding decodedPCGEnc =
				new PlatformCommunicationGraphEncoding(pcgEnc);
		DefaultDirectedGraph<IResourceAdapter, DefaultEdge> pGraph =
				decodedPCGEnc.getActualGraph();

		for(IExecutionUnitAdapter execUnitContainer : peuiEnc.getAbstractContainerExecUnits()) {
			IExecutionUnitAdapter template =
					execUnitTemplateEnc.getTemplateWithin(execUnitContainer);

			List<IExecutionUnitAdapter> children =
					new ArrayList<>(peuiEnc.getChildInstances(execUnitContainer));
			for(IExecutionUnitAdapter childInst : children) {
				addSubGraph(decodedPCGEnc, pGraph, execUnitTemplateEnc, template, childInst,
						peuiEnc);
			}

			Collection<IResourceAdapter> resources =
					execUnitTemplateEnc.getContainedResources(template);
			pGraph.removeAllVertices(resources);
		}

		// DEBUG HELPER
		// decodedPCGEnc.display();

		return decodedPCGEnc;
	}

	/**
	 * Adds a sub graph to the internal graph of the {@link PlatformCommunicationGraphEncoding} such
	 * that the desired number child {@link IExecutionUnitAdapter}s is reached.
	 */
	private void addSubGraph(PlatformCommunicationGraphEncoding decodedPCGEnc,
			DefaultDirectedGraph<IResourceAdapter, DefaultEdge> pGraph,
			PlatformExecUnitTemplateEncoding execUnitTemplateEnc, IExecutionUnitAdapter template,
			IExecutionUnitAdapter templateInstance, PlatformExecUnitInstanceEncoding peuiEnc) {
		Collection<IResourceAdapter> resources =
				execUnitTemplateEnc.getContainedResources(template);
		DefaultDirectedGraph<IResourceAdapter, DefaultEdge> subGraph =
				decodedPCGEnc.getSubTreeOf(resources);

		// JGraphTVisualizer<IResourceAdapter, DefaultEdge> taskViz =
		// new JGraphTVisualizer<IResourceAdapter, DefaultEdge>(subGraph);
		// taskViz.displayGraph("Platform communication graph");

		Set<IResourceAdapter> transmissionResources = new HashSet<>();
		for(IResourceAdapter resource : resources) {
			Collection<ITransmissionUnitAdapter> outerTUnits =
					Graphs.neighborSetOf(pGraph, resource).stream()
							.filter(v -> v instanceof ITransmissionUnitAdapter)
							.filter(v -> !resources.contains(v))
							.map(ITransmissionUnitAdapter.class::cast).collect(toSet());
			transmissionResources.addAll(outerTUnits);
		}

		// Copy the subgraph, add it to the platform graph, and connect it to the same
		// transmission resources as the template.
		Map<IResourceAdapter, IResourceAdapter> origToCopiedResources =
				peuiEnc.getContainedInstancesOf(templateInstance);
		// NOTE: Here, we are exchanging the template vertex in the subgraph with the instance from
		// the given encoding.
		// TODO(AD): Candidate to convert to a util method.
		Graphs.addAllVertices(pGraph, origToCopiedResources.values());
		for(DefaultEdge edge : getAllEdgesOf(subGraph)) {
			IResourceAdapter oldSource = pGraph.getEdgeSource(edge);
			IResourceAdapter oldTarget = pGraph.getEdgeTarget(edge);
			IResourceAdapter newSource = origToCopiedResources.get(oldSource);
			IResourceAdapter newTarget = origToCopiedResources.get(oldTarget);
			pGraph.addEdge(newSource, newTarget);
		}
		Graphs.addAllEdges(pGraph, subGraph, getAllEdgesOf(subGraph));

		// Connect the inserted subgraph to the rest of the platform.
		// IResourceAdapter constructedInstance = origToCopiedResources.get(template);
		for(IResourceAdapter tResource : transmissionResources) {
			for(Entry<IResourceAdapter, IResourceAdapter> resEntry : origToCopiedResources
					.entrySet()) {
				if(pGraph.containsEdge(resEntry.getKey(), tResource)) {
					pGraph.addEdge(resEntry.getValue(), tResource);
				}
				if(pGraph.containsEdge(tResource, resEntry.getKey())) {
					pGraph.addEdge(tResource, resEntry.getValue());
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public PlatformCommunicationGraphEncoding decode(Genotype arg0) {
		// Required by the framework...
		return null;
	}
}
