/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform.variable;

import static org.apache.commons.collections4.list.UnmodifiableList.unmodifiableList;

import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.stream.IntStream;

import org.apache.commons.collections4.list.UnmodifiableList;
import org.fortiss.af3.exploration.alg.dse.RandomExploration;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.PhenotypeBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformExecUnitTemplateEncoding;
import org.fortiss.af3.exploration.alg.module.common.operator.copy.Copyable;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.opt4j.core.Genotype;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.HashBasedTable;
import com.google.common.collect.ImmutableListMultimap;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableTable;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.Table;

/**
 * Encoding that defines how many child {@link IExecutionUnitAdapter}s shall be instantiated for a
 * parent {@link IExecutionUnitAdapter}.
 * 
 * @author diewald
 */
public class PlatformExecUnitInstanceEncoding extends PhenotypeBase
		implements Genotype, Copyable<PlatformExecUnitInstanceEncoding> {

	/** {@link Random} number generator of the D3SE. */
	private RandomExploration rand = RandomExploration.getInstance();

	/**
	 * Relates parent {@link IExecutionUnitAdapter}s with the instance template
	 * {@link IExecutionUnitAdapter}s inside them.
	 */
	private PlatformExecUnitTemplateEncoding templateExecUnitEnc;

	/**
	 * Assignments of container {@link IExecutionUnitAdapter}s to the number of child
	 * {@link IExecutionUnitAdapter} instances.
	 */
	private ListMultimap<IExecutionUnitAdapter, IExecutionUnitAdapter> subElementInstances =
			ArrayListMultimap.create();

	/**
	 * Assignments of container {@link IExecutionUnitAdapter}s to the number of child
	 * {@link IExecutionUnitAdapter} instances.
	 */
	private Table<IExecutionUnitAdapter, IResourceAdapter, IResourceAdapter> subElementContainedInstances =
			HashBasedTable.create();

	/**
	 * References the numbers of instances of child {@link IExecutionUnitAdapter}s that are
	 * available for a given parent {@link ExecutionUnit}.
	 */
	private ListMultimap<IExecutionUnitAdapter, Integer> allowedInstances =
			ArrayListMultimap.create();

	/** Constructor. */
	public PlatformExecUnitInstanceEncoding(
			ListMultimap<IExecutionUnitAdapter, Integer> allowedInstances,
			PlatformExecUnitTemplateEncoding templateExecUnitEnc) {
		this.allowedInstances.putAll(allowedInstances);
		this.templateExecUnitEnc = templateExecUnitEnc;
	}

	/** Copy Constructor. */
	public PlatformExecUnitInstanceEncoding(PlatformExecUnitInstanceEncoding encoding) {
		this.subElementInstances.putAll(encoding.subElementInstances);
		this.allowedInstances.putAll(encoding.allowedInstances);
		this.subElementContainedInstances.putAll(encoding.subElementContainedInstances);
		// The templateExecUnitEnc is just an input: no copying needed here.
		this.templateExecUnitEnc = encoding.templateExecUnitEnc;
	}

	/**
	 * Returns the set of abstract {@link ExecutionUnit}s whose contained {@link ExecutionUnit}
	 * number is varied.
	 */
	public Set<IExecutionUnitAdapter> getAbstractContainerExecUnits() {
		return allowedInstances.keySet();
	}

	/**
	 * Returns a list of Integers that defines the number of child {@link ExecutionUnit}s that the
	 * container {@link ExecutionUnit} may have.
	 */
	public List<Integer> getAllowedChildInstances(IExecutionUnitAdapter execUnitContainer) {
		return allowedInstances.get(execUnitContainer);
	}

	/**
	 * Returns the selected number of child {@link ExecutionUnit}s assigned to the given
	 * {@link ExecutionUnit} container.
	 */
	public UnmodifiableList<IExecutionUnitAdapter>
			getChildInstances(IExecutionUnitAdapter execUnitContainer) {
		return (UnmodifiableList<IExecutionUnitAdapter>)unmodifiableList(
				subElementInstances.get(execUnitContainer));
	}

	/**
	 * Adjusts the internal list of child instances to match the given number of desired instances.
	 */
	public void setChildInstanceNumber(IExecutionUnitAdapter execUnitContainer,
			Integer instanceNum) {
		List<IExecutionUnitAdapter> childInstances = subElementInstances.get(execUnitContainer);
		if(instanceNum > childInstances.size()) {
			IntStream.range(childInstances.size(), instanceNum)
					.forEach(i -> addInstance(execUnitContainer));
		} else if(instanceNum < childInstances.size()) {
			IntStream.range(instanceNum, childInstances.size())
					.forEach(i -> removeRandomInstance(execUnitContainer));
		}
	}

	/** Removes a random child instance of the given parent from the internal encoding. */
	private void removeRandomInstance(IExecutionUnitAdapter execUnitContainer) {
		List<IExecutionUnitAdapter> instances = subElementInstances.get(execUnitContainer);
		int selIdx = rand.nextInt(instances.size());
		IExecutionUnitAdapter rmElement = instances.remove(selIdx);
		Map<IResourceAdapter, IResourceAdapter> origResources =
				subElementContainedInstances.row(rmElement);
		origResources.clear();
	}

	/** Adds a child instance of the given parent to the internal encoding. */
	private void addInstance(IExecutionUnitAdapter execUnitContainer) {
		IExecutionUnitAdapter template =
				templateExecUnitEnc.getTemplateWithin(execUnitContainer);
		IExecutionUnitAdapter instance = (IExecutionUnitAdapter)template.clone();
		subElementInstances.put(execUnitContainer, instance);

		for(IResourceAdapter containedRes : templateExecUnitEnc
				.getContainedResources(template)) {
			subElementContainedInstances.put(instance, containedRes, containedRes.clone());
		}
	}

	/**
	 * Assigns the number of child {@link ExecutionUnit} instances to the given container
	 * {@link ExecutionUnit}.
	 */
	public void setChildInstances(IExecutionUnitAdapter execUnitContainer,
			List<IExecutionUnitAdapter> instances) {
		subElementInstances.putAll(execUnitContainer, instances);
	}

	/**
	 * Returns the {@link IExecutionUnitAdapter} that serves as a template to create the instances.
	 */
	public IExecutionUnitAdapter getTemplateOf(IExecutionUnitAdapter parentExecUnit) {
		return templateExecUnitEnc.getTemplateWithin(parentExecUnit);
	}

	/** Returns a view on the internal abstract to instances mapping. */
	public ImmutableMap<IExecutionUnitAdapter, IExecutionUnitAdapter> getTemplates() {
		return templateExecUnitEnc.getTemplates();
	}

	/** Returns a view on the internal abstract to instances mapping. */
	public ImmutableListMultimap<IExecutionUnitAdapter, IExecutionUnitAdapter>
			getInstanceMap() {
		return ImmutableListMultimap.copyOf(subElementInstances);
	}

	/** Returns a view on the internal abstract to instances table. */
	public ImmutableTable<IExecutionUnitAdapter, IResourceAdapter, IResourceAdapter>
			getContainedInstanceTable() {
		return ImmutableTable.copyOf(subElementContainedInstances);
	}

	/** Returns a view on the internal abstract to instances mapping. */
	public Map<IResourceAdapter, IResourceAdapter>
			getContainedInstancesOf(IExecutionUnitAdapter instantiatedExecUnit) {
		return subElementContainedInstances.row(instantiatedExecUnit);
	}

	/** {@inheritDoc} */
	@Override
	public PlatformExecUnitInstanceEncoding copy() {
		return new PlatformExecUnitInstanceEncoding(this);
	}

	/** {@inheritDoc} */
	@Override
	public <G extends Genotype> G newInstance() {
		// Not used
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public int size() {
		return subElementInstances.keySet().size();
	}
}
