/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling;

import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toSet;
import static org.eclipse.core.runtime.Assert.isNotNull;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.DSE_DEBUG_CONVERGENCE_SUBSOLVER;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.copyExplorationSpecification;
import static org.fortiss.af3.exploration.smt.modeltransformation.BasicDeploScheduleConstraint.generateDeploScheduleConstraints;
import static org.fortiss.af3.exploration.smt.modeltransformation.BasicScheduleConstraint.generatePureScheduleConstraints;
import static org.fortiss.af3.exploration.smt.modeltransformation.ConstraintDefinitionUtils.createDurationLiteral;
import static org.fortiss.af3.exploration.smt.modeltransformation.ConstraintDefinitionUtils.createStartTimeLiteral;
import static org.fortiss.af3.exploration.smt.modeltransformation.ConstraintDefinitionUtils.generateForAll;
import static org.fortiss.af3.exploration.smt.util.AllocationToSchedule.generateSchedule;
import static org.fortiss.af3.exploration.smt.util.Z3ModelElementFactory.createSMTConstraint;
import static org.fortiss.af3.exploration.smt.util.Z3ModelElementFactory.createSMTObjective;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createArithmeticLiteral;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createBooleanLiteral;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createLessEqual;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createMinimize;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createPlus;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSet;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSum;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSuperSet;
import static org.fortiss.af3.exploration.util.ExplorationModelElementFactory.createExplorationSolution;
import static org.fortiss.tooling.common.util.LambdaUtils.filterByType;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Queue;

import org.apache.commons.lang3.tuple.Pair;
import org.eclipse.core.runtime.Assert;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemAccessTask;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemoryUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.scheduling.problem.SharedMemAccessLatencyEncoding;
import org.fortiss.af3.exploration.alg.feature.scheduling.transform.MessageToRouteTransformation;
import org.fortiss.af3.exploration.alg.feature.taskmapping.TaskToExUTransformation;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.Message;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.MessageEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskExUMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskExUMappingEntry;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.dseml.model.arithmetic.ModelElementPropertyLiteral;
import org.fortiss.af3.exploration.dseml.model.arithmetic.Plus;
import org.fortiss.af3.exploration.dseml.model.arithmetic.Sum;
import org.fortiss.af3.exploration.dseml.model.booleanp.BooleanLiteral;
import org.fortiss.af3.exploration.dseml.model.booleanp.ForAll;
import org.fortiss.af3.exploration.dseml.model.booleanp.comparison.LessEqual;
import org.fortiss.af3.exploration.dseml.model.expression.Set;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.exploration.dseml.model.function.Minimize;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.af3.exploration.model.project.InitialInputDefinitionStep;
import org.fortiss.af3.exploration.model.project.ModelSnapshot;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SolutionState;
import org.fortiss.af3.exploration.smt.model.SMTConstraint;
import org.fortiss.af3.exploration.smt.model.SMTObjective;
import org.fortiss.af3.exploration.smt.modeltransformation.DSMLtoSMTTransformator;
import org.fortiss.af3.exploration.smt.modeltransformation.ScheduleRun;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.Route;
import org.fortiss.af3.platform.model.Segment;
import org.fortiss.af3.schedule.model.PeriodicTimeTrigger;
import org.fortiss.af3.schedule.model.ResourceAllocation;
import org.fortiss.af3.schedule.model.SystemSchedule;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.TaskOutputPort;
import org.fortiss.af3.task.model.allocation.ComponentToTaskAllocationEntry;
import org.fortiss.af3.task.model.allocation.SignalToRouteAllocationEntry;
import org.fortiss.af3.task.model.allocation.TaskToExecutionUnitAllocationEntry;
import org.fortiss.af3.task.model.allocation.TaskWcetEntry;
import org.fortiss.af3.task.model.allocation.TaskWcetTable;
import org.fortiss.af3.timing.model.TimingSpecification;
import org.fortiss.af3.timing.model.annotation.Wcet;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.common.util.LambdaUtils;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.opt4j.core.Genotype;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBasedTable;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.Table;
import com.google.inject.Provides;
import com.microsoft.z3.Context;

/**
 * Generates a {@link StrictTTSchedule} that is constructed by calling the AF3 scheduler based on
 * Z3.
 * 
 * @author diewald
 */
// TODO(AD): Add the following parameters to the schedule: <IExecutionUnitAdapter,
// TaskExUMappingEntry, TaskExUMappingEncoding>
// This implies updating the registration with the feature.
public class SMTStrictTTDecoder extends DecoderModule<StrictTTSchedule> {

	/** Timeout for the SMT backend to find a solution. */
	private final int smtTimeout = 5000; // in milliseconds

	/**
	 * Relative portion of the upper latency bound used as a stopping criterion for the SMT calls.
	 */
	private final BigDecimal relBoundDistance = BigDecimal.valueOf(0.0001);

	/** Step size of the latency bound search. */
	private final BigDecimal STEP_SIZE = BigDecimal.valueOf(2.0);

	/**
	 * Transforms {@link TaskExUMappingEncoding} to a {@link SuperSet} of
	 * {@link TaskToExecutionUnitAllocationEntry}s.
	 */
	private TaskToExUTransformation allocTransformator = new TaskToExUTransformation();

	/** The SMT(Z3) scheduler. */
	private ScheduleRun smtSchedRun = new ScheduleRun();

	/** See {@link SMTStrictTTDecoder}. */
	// TODO(AD): Split up
	@Provides
	@Decodes
	public StrictTTSchedule decode(TaskExUMappingEncoding tmEnc, MessageEncoding msgEnc,
			TaskGraphEncoding tgEnc, PlatformCommunicationGraphEncoding pcgEnc, TaskMapper tMapper,
			SharedMemAccessLatencyEncoding shMemEnc, ExplorationSpecification expSpec) {

		// TODO: Externalize the validation such that the encoding is not required to be passed to
		// the constructor (is there for validation purposes)
		BiMap<EObject, EObject> copyRefMap = HashBiMap.create();
		ExplorationSpecification copiedSpec = copyExplorationSpecification(expSpec, copyRefMap);
		SuperSetMap superSets = copiedSpec.getSearchSpace();

		SuperSet<TaskToExecutionUnitAllocationEntry> taskToExecUnitSet =
				allocTransformator.transform(tmEnc);
		superSets.put(TaskToExecutionUnitAllocationEntry.class, taskToExecUnitSet);

		MessageToRouteTransformation msgToRouteTransformer = new MessageToRouteTransformation();
		SuperSet<SignalToRouteAllocationEntry> sigToRoute =
				msgToRouteTransformer.transform(msgEnc, tMapper);
		Collection<Route> routes =
				sigToRoute.getEntries().stream().map(e -> e.getRoute()).collect(toSet());
		SuperSet<Route> routeSS = createSuperSet(routes, "Routes", Route.class);
		superSets.put(Route.class, routeSS);

		// TODO: Do not query the transmissionUnits from the input set but the platform graph.
		SuperSet<SignalToRouteAllocationEntry> sigToRouteOut =
				createSuperSet(emptyList(), "SignalToRoutes", SignalToRouteAllocationEntry.class);
		Task randomTask = getFirst(tmEnc.getRequesters()).get().getObject(Task.class);
		TimingSpecification tSpec = getTimingSpec(randomTask);
		SystemSchedule sysSched = generateSchedule(superSets, sigToRouteOut, tSpec);

		// superSets.put(Route.class, routeSS);
		SuperSet<ResourceAllocation> rsAllocSet =
				createSuperSet("ResourceAllocations", ResourceAllocation.class);
		rsAllocSet.getEntries().addAll(getChildrenWithType(sysSched, ResourceAllocation.class));

		TaskWcetTable taskWcets = getFirst(superSets.get(TaskWcetTable.class).getEntries()).get();
		taskWcets = copy(taskWcets);
		// adjustMemoryReadDurations(taskWcets, tmEnc, tgEnc, tMapper, pcgEnc);
		SuperSet<TaskWcetTable> taskWcetSet =
				createSuperSet(asList(taskWcets), "Task WCETs", TaskWcetTable.class);
		adjustMemoryReadDurations(taskWcets, tmEnc, shMemEnc);

		// TODO (AD): Here, we directly see the need for internal transformation... We are relying
		// on non-changed input sets.
		List<SMTConstraint> basicScheduleDeploConstraints =
				generateDeploScheduleConstraints(rsAllocSet, superSets.get(Task.class),
						superSets.get(Signal.class), superSets.get(ExecutionUnit.class), routeSS,
						superSets.get(ComponentToTaskAllocationEntry.class), taskWcetSet);
		List<SMTConstraint> basicScheduleConstraints = generatePureScheduleConstraints(rsAllocSet,
				superSets.get(Task.class), superSets.get(ExecutionUnit.class),
				superSets.get(ComponentToTaskAllocationEntry.class),
				superSets.get(TaskToExecutionUnitAllocationEntry.class), sigToRouteOut);

		// We have to add the scheduling constraints at the beginning of the target list.
		List<ExplorationTarget<?>> cachedTargets = copiedSpec.getTargets();
		copiedSpec.getTargets().clear();
		copiedSpec.getTargets().addAll(basicScheduleDeploConstraints);
		copiedSpec.getTargets().addAll(basicScheduleConstraints);

		SuperSetMap updSuperSets = copiedSpec.getSearchSpace();
		updSuperSets.put(ResourceAllocation.class, rsAllocSet);
		updSuperSets.put(SignalToRouteAllocationEntry.class, sigToRouteOut);

		Long schedStartingTime = System.currentTimeMillis();
		Pair<ExplorationSolution, BigDecimal> expSol =
				seekApproxOptimalSchedule(copiedSpec, rsAllocSet);
		Long solveDuration = System.currentTimeMillis() - schedStartingTime;
		if(expSol.getLeft() == null) {
			StrictTTSchedule<IExecutionUnitAdapter, TaskExUMappingEntry, TaskExUMappingEncoding> schedule =
					new StrictTTSchedule<>(tmEnc, BigDecimal.ZERO, solveDuration);
			schedule.setInfeasible();
			return schedule;
		}

		return constructInternalSchedule(tmEnc, msgEnc, pcgEnc, tMapper, expSol.getLeft(),
				expSol.getRight(), solveDuration);
	}

	/**
	 * Adjusts the duration of reader tasks based on their memory access latency. It is represented
	 * by the bandwidth of involved {@link ICommunicationResourceAdapter} to access an
	 * {@link IMemoryUnitAdapter}.
	 * 
	 * @param wcetTable
	 *            Specifies all WCETs of tasks. Used for the task duration in the SMT scheduling.
	 * @param tmEnc
	 *            {@link ITaskAdapter} mapping to extract the compute tasks.
	 * @param shMemEnc
	 *            Encodes the memory access latencies of acquisition and restitution tasks acc. to
	 *            the AER model.
	 */
	private void adjustMemoryReadDurations(TaskWcetTable wcetTable, TaskExUMappingEncoding tmEnc,
			SharedMemAccessLatencyEncoding shMemEnc) {
		for(ITaskAdapter task : filterByType(tmEnc.getRequesters(), IMemAccessTask.class)) {
			BigDecimal latency = shMemEnc.getLatencyOf(task);
			if(latency.doubleValue() < 1.0) {
				latency = BigDecimal.ONE;
			}
			latency = BigDecimal.valueOf(latency.longValue());

			IResourceAdapter resource = getFirst(tmEnc.getResources(task)).get();

			Wcet wcet = wcetTable.getParameter(task.getObject(Task.class),
					resource.getObject(ExecutionUnit.class), TaskWcetEntry.class, Wcet.class);
			wcet.setValue(latency);
		}
	}

	/**
	 * Executes the SMT scheduler to find a solution with minimum latency. Therefore, a latency
	 * bound is encoded as a constraint and the result is checked for feasibility. In a loop, this
	 * bound is modified (1/2 step size) until the step size reached {@value #relBoundDistance} of
	 * the upper latency bound.
	 * 
	 * @param expSpec
	 *            {@link ExplorationSpecification} with {@link ResourceAllocation} skeleton.
	 * @param rsAllocSet
	 *            {@link ResourceAllocation} skeleton.
	 * @return Feasible {@link ExplorationSolution} and the lower bound on the schedule length.
	 */
	private Pair<ExplorationSolution, BigDecimal> seekApproxOptimalSchedule(
			ExplorationSpecification expSpec, SuperSet<ResourceAllocation> rsAllocSet) {
		ExplorationSolution latestValidSolution = null;
		ExplorationSolution expSol = null;
		BigDecimal minLengthBound = BigDecimal.ZERO;

		BigDecimal upperBound = getUpperBoundScheduleLength(rsAllocSet);
		BigDecimal relUpperBound = upperBound.multiply(relBoundDistance);

		BigDecimal diffBound = upperBound;
		BigDecimal activeBound = upperBound;

		Map<String, String> parameters = new HashMap<>();
		parameters.put("timeout", smtTimeout + "");
		Context context = new Context(parameters);
		DSMLtoSMTTransformator transformation = new DSMLtoSMTTransformator(context);

		if(DSE_DEBUG_CONVERGENCE_SUBSOLVER) {
			System.out.println("Lower Bound, Active bound");
			System.out.println("==========================");
		}

		SolutionState solState = SolutionState.UNKNOWN;
		do {
			activeBound = BigDecimal.valueOf(Math.ceil(activeBound.doubleValue()));

			SMTConstraint constraint = createScheduleLengthConstraint(rsAllocSet, activeBound);
			expSpec.getTargets().add(constraint);
			expSol = createExplorationSolution();
			expSol.setExplorationSpec(expSpec);

			transformation.transform(expSpec);

			try {
				smtSchedRun.solveNonOptimized(context, parameters, transformation, expSol, null);
			} catch(Exception e) {
				// If the Z3 solver fails due to an internal bug or machine-specific behavior, we
				// use an best effort approach and just return an already found solution.
				e.printStackTrace();
				return Pair.of(latestValidSolution, minLengthBound);
			}

			expSpec.getTargets().remove(constraint);

			solState = expSol.getSolutionState();
			diffBound = diffBound.divide(STEP_SIZE);

			if(DSE_DEBUG_CONVERGENCE_SUBSOLVER) {
				System.out.println(
						minLengthBound + ", " + activeBound.doubleValue() + ", " + solState);
			}
			switch(solState) {
				case UNSAT:
					minLengthBound = activeBound;
					activeBound = activeBound.add(diffBound);
					break;
				case OPTIMIZED:
				case OPTIMAL:
					activeBound = activeBound.subtract(diffBound);
					latestValidSolution = expSol;
					break;
				case UNKNOWN:
					// no solution within timeout --> drop the candidate
					return Pair.of(latestValidSolution, minLengthBound);
			}
		} while(diffBound.doubleValue() > relUpperBound.doubleValue());

		return Pair.of(latestValidSolution, minLengthBound);
	}

	/**
	 * Constructs the internal schedule representation from the returned {@link SuperSet} of
	 * {@link ResourceAllocation}s from the SMT solver.
	 */
	private StrictTTSchedule constructInternalSchedule(TaskExUMappingEncoding tmEnc,
			MessageEncoding msgEnc, PlatformCommunicationGraphEncoding pcgEnc, TaskMapper tMapper,
			ExplorationSolution schedSolution, BigDecimal minLength, Long solveDuration) {
		StrictTTSchedule<IExecutionUnitAdapter, TaskExUMappingEntry, TaskExUMappingEncoding> schedule =
				new StrictTTSchedule<>(tmEnc, minLength, solveDuration);

		// TODO(AD): Check solution state and document the "get(0)".
		SuperSet<ResourceAllocation> resAllocSol =
				schedSolution.getSolutions().get(0).getSolutionModel(ResourceAllocation.class);
		// TODO(AD): Check solution state and document the "get(0)".
		SuperSet<SignalToRouteAllocationEntry> sigAllocSol = schedSolution.getSolutions().get(0)
				.getSolutionModel(SignalToRouteAllocationEntry.class);

		// TODO(AD): Extract & centralize
		Map<IModelElement, IResourceAdapter> resourceMap = new HashMap<>();
		for(IResourceAdapter currResource : pcgEnc.getActualGraph().vertexSet()) {
			resourceMap.put(currResource.getObject(IModelElement.class), currResource);
		}

		Map<Segment, Message> segMsgMap = generateSegmentMsgMap(msgEnc, sigAllocSol);

		for(ResourceAllocation resAlloc : resAllocSol.getEntries()) {
			IModelElement schedME = resAlloc.getSchedulableEntity().getModelElement();
			// TODO(AD): Add handling of output ports / signals.
			IRequestAdapter requester = null;
			IResourceAdapter resource = null;
			if(schedME instanceof Task) {
				requester = tMapper.getAdapter((Task)schedME);
				// TODO(AD): Only 1:1 mappings are supported as of now.
				resource = getFirst(tmEnc.getResources((ITaskAdapter)requester)).get();
			} else if(schedME instanceof Segment) {
				// FIXME(AD): HACK! This assumes unicast routes.
				Route route = ((Segment)schedME).getRoute();
				ExecutionUnit senderRes = route.getSenderExecutionUnit();
				Collection<ExecutionUnit> receiverRes = route.getReceiverExecutionUnits();
				if(getFirst(receiverRes).get() == senderRes) {
					continue;
				}
				IModelElement segmentResource = ((Segment)schedME).getEntity();
				resource = resourceMap.get(segmentResource);
				Assert.isNotNull(resource);
			}

			PeriodicTimeTrigger trigger = (PeriodicTimeTrigger)resAlloc.getTrigger();
			Double start = trigger.getStartTime().doubleValue();
			Double end = 5.0 + start.doubleValue();
			if(requester instanceof ITaskAdapter) {
				double duration = resAlloc.getDuration().doubleValue();
				end = duration + start.doubleValue();

				// TODO: We only support 1:1 mappings in this place.
				TaskExUMappingEntry mappingEntry =
						getFirst(tmEnc.getMappingEntriesOf((ITaskAdapter)requester)).get();
				schedule.createTimeSlot(mappingEntry, resource, emptyList(), start, end, 0);
			} else {
				Message assocMessage = segMsgMap.get(schedME);
				isNotNull(assocMessage);
				TimeSlot communicationSlot = new TimeSlot(assocMessage,
						(ICommunicationResourceAdapter)resource, emptyList(), start, end);
				schedule.registerTimeSlot(communicationSlot, null, resource);
			}
		}

		return schedule;
	}

	/**
	 * Generates a mapping of {@link Segment}s to the {@link Message}s that were used to construct
	 * their parent {@link Route}s.
	 */
	private Map<Segment, Message> generateSegmentMsgMap(MessageEncoding msgEnc,
			SuperSet<SignalToRouteAllocationEntry> sigRouteAllocs) {
		Map<Segment, Message> segmentMsgMap = new HashMap<>();

		Table<IModelElement, Collection<IModelElement>, Message> senderMsgMap =
				HashBasedTable.create();
		for(Message msg : msgEnc.getAllMessages()) {
			IModelElement senderPort = msg.getSignalAdapter().getObject(IModelElement.class);
			Collection<IModelElement> recvTasks =
					msg.getReceiverMappingEntries().stream().map(e -> e.getSource())
							.map(req -> req.getObject(IModelElement.class)).collect(toList());
			senderMsgMap.put(senderPort, recvTasks, msg);
		}

		for(SignalToRouteAllocationEntry sigRouteEntry : sigRouteAllocs.getEntries()) {
			Signal signal = LambdaUtils.getFirst(sigRouteEntry.getSignals()).get();
			TaskOutputPort sourceTaskPort = signal.getSourceTaskPort();
			// TODO(AD): Here, we are assuming unicast messages.
			Collection<Task> recvTasks = asList(signal.getTargetTaskPort().getTask());
			Message msg = senderMsgMap.get(sourceTaskPort, recvTasks);

			Route route = sigRouteEntry.getRoute();
			Queue<Segment> segments = new LinkedList<>();
			segments.add(route.getRootSegment());
			while(!segments.isEmpty()) {
				Segment curSegment = segments.poll();
				segmentMsgMap.put(curSegment, msg);
				segments.addAll(curSegment.getNext());
			}
		}

		return segmentMsgMap;
	}

	/**
	 * Returns the upper bound of a time-triggered schedule for the given {@link ResourceAllocation}
	 * {@link SuperSet} (= sum of all {@link Task} WCETs).
	 */
	private static BigDecimal getUpperBoundScheduleLength(SuperSet<ResourceAllocation> rsAllocSet) {
		BigDecimal upperBound = BigDecimal.ZERO;

		List<BigDecimal> allWcets = new ArrayList<>();
		for(ResourceAllocation ra : rsAllocSet.getEntries()) {
			IModelElement modelElement = ra.getSchedulableEntity().getModelElement();
			if(!(modelElement instanceof Task)) {
				continue;
			}
			Task task = (Task)modelElement;
			allWcets.add(task.getWcet((ExecutionUnit)ra.getResource()));
		}

		allWcets.sort(Collections.reverseOrder());
		for(BigDecimal wcet : allWcets) {
			upperBound = upperBound.add(wcet);
		}

		return upperBound;
	}

	/**
	 * Creates an {@link SMTConstraint} that enforces that the end time of all
	 * {@link ResourceAllocation}s is below a given {@code upperBound}.
	 */
	private static SMTConstraint createScheduleLengthConstraint(
			SuperSet<ResourceAllocation> resAllocSS, BigDecimal upperBound) {
		org.fortiss.af3.exploration.dseml.model.expression.Set<ResourceAllocation> set =
				createSet(resAllocSS, "AllResources_" + resAllocSS.getName() + "_Set",
						ResourceAllocation.class);
		set.getEntries().addAll(resAllocSS.getEntries());

		ModelElementPropertyLiteral startSet = createStartTimeLiteral(set);
		ModelElementPropertyLiteral durationSet = createDurationLiteral(set);
		Plus endTime = createPlus(startSet, durationSet);

		LessEqual lessEqual = createLessEqual(endTime, createArithmeticLiteral(upperBound));
		ForAll forAll = generateForAll(set, lessEqual, true);

		return createSMTConstraint("Latency: " + upperBound, forAll, true);
	}

	/**
	 * Creates an objective that minimizes the sum of ending times of tasks, for the given
	 * {@link SystemSchedule}.
	 */
	// TODO(AD): Extend for the per-resource case.
	public static SMTObjective
			createMinSumEndTimesObjective(SuperSet<ResourceAllocation> resAllocSS) {
		Set<ResourceAllocation> set =
				createSet(resAllocSS, "aResourceAllocationsSet", ResourceAllocation.class);
		set.getEntries().addAll(resAllocSS.getEntries());

		ModelElementPropertyLiteral startLit = createStartTimeLiteral(set);
		ModelElementPropertyLiteral durationLit = createDurationLiteral(set);
		Plus endTime = createPlus(startLit, durationLit);
		BooleanLiteral trivialCondition = createBooleanLiteral(true);
		Sum sum = createSum(set.getCastedSet(IModelElement.class), trivialCondition, endTime);

		Minimize min = createMinimize(sum);
		return createSMTObjective("Minimize Latency", min);
	}

	/**
	 * Helper to get the {@link TimingSpecification} that is required to construct the SMT
	 * scheduling constraints.
	 */
	private static TimingSpecification getTimingSpec(IHierarchicElement contained) {
		InitialInputDefinitionStep inputStep =
				getParentElement(contained, InitialInputDefinitionStep.class, false);
		Collection<ModelSnapshot> snaps = inputStep.getModelSnapshots();
		Optional<IProjectRootElement> tSpec =
				snaps.stream().filter(s -> s.getRootElement() instanceof TimingSpecification)
						.map(ModelSnapshot::getRootElement).findAny();
		if(tSpec.isPresent()) {
			return (TimingSpecification)tSpec.get();
		}
		throw new RuntimeException("A Timing Specification must be given to the DSE to perform" +
				"the SMT-based scheduling");
	}

	/** {@inheritDoc} */
	@Override
	public StrictTTSchedule<IExecutionUnitAdapter, TaskExUMappingEntry, TaskExUMappingEncoding>
			decode(Genotype genotype) {
		// Required by the framework...
		return null;
	}
}
