/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.problem;

import static org.fortiss.tooling.common.util.LambdaUtils.filterByType;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.PhenotypeBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemAccessTask;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemAccessTask.AER_MEM_TYPE;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.feature.scheduling.TimeSlot;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskExUMappingEncoding;
import org.fortiss.af3.schedule.model.ResourceAllocation;

/**
 * Transforms the internal {@link TimeSlot} representation into {@link ResourceAllocation}s of AF3.
 * 
 * @author diewald
 */
public class SharedMemAccessLatencyEncoding extends PhenotypeBase {

	/** Defines how many bits can be fetched at a time. Used to adjust the temporal measurements. */
	private final static double CACHE_LINE_IN_BITS = 512;

	/** Remembers the interference latencies calculated at construction time. */
	private Map<ITaskAdapter, BigDecimal> latencyMap = new HashMap<>();

	/** Constructor. */
	public SharedMemAccessLatencyEncoding(TaskExUMappingEncoding tmEnc, TaskGraphEncoding tgEnc) {
		for(IMemAccessTask task : filterByType(tmEnc.getRequesters(), IMemAccessTask.class)) {
			TaskGraph tg = tgEnc.getTaskGraphOf(task);

			// Since read/write tasks can be currently mapped to accelerators, we need to get the
			// execution target to be aware of the destination of data prefetching. Here, we know
			// that every execution task has precisely one assoc. reader and write task.
			ITaskAdapter executionTask = null;
			if(task.getAERType() == AER_MEM_TYPE.ACQUISITION) {
				executionTask = getFirst(tg.getSinkTasks(task)).get();
			} else {
				executionTask = getFirst(tg.getSourceTasks(task)).get();
			}

			// Currently, we only support 1:1 mappings.
			IExecutionUnitAdapter execUnit = getFirst(tmEnc.getResources(executionTask)).get();

			BigDecimal taskLatency =
					getLatencyWithInterference(task, tg, execUnit, tmEnc.getResources());
			latencyMap.put(task, taskLatency);
		}
	}

	/** Copy constructor. */
	public SharedMemAccessLatencyEncoding(SharedMemAccessLatencyEncoding encoding) {
		this.latencyMap.putAll(encoding.latencyMap);
	}

	/** Returns the memory access latency of the given task to retrieve its input data. */
	public BigDecimal getLatencyOf(ITaskAdapter task) {
		return latencyMap.get(task);
	}

	/**
	 * Returns the latency of a transmission of the data carried by the given
	 * {@link ISignalAdapter}s over the given {@link ICommunicationResourceAdapter}.
	 * Hereby, the communication from the sender {@link IExecutionUnitAdapter} suffers from
	 * interference from the given set of {@link IExecutionUnitAdapter}s.
	 */
	private BigDecimal getLatencyWithInterference(IMemAccessTask task, TaskGraph tg,
			IExecutionUnitAdapter readerUnit,
			Collection<IExecutionUnitAdapter> interferingExecUnits) {
		Collection<ISignalAdapter> signals;
		if(task.getAERType() == AER_MEM_TYPE.ACQUISITION) {
			signals = tg.getSourceSignals(task);
		} else {
			signals = tg.getSinkSignals(task);
		}
		long dataSize = getDataSizeOf(signals);

		// We do not consider the transmission time as of now
		// Double baseTransmissionTime = commResources.stream()
		// .map(cR -> cR.getTransmissionDuration(dataSize)).reduce(0.0, (a, b) -> a + b);

		Collection<IExecutionUnitAdapter> allOtherExecUnits = new ArrayList<>(interferingExecUnits);
		allOtherExecUnits.remove(readerUnit);
		double interferenceLatency = allOtherExecUnits.stream()
				.map(e -> e.getInterferenceLatency(readerUnit)).reduce(0.0, (a, b) -> a + b);
		interferenceLatency *= (dataSize / CACHE_LINE_IN_BITS);

		double overallLatency = interferenceLatency;

		return BigDecimal.valueOf(overallLatency);
	}

	/** Returns the */
	private long getDataSizeOf(Collection<ISignalAdapter> signals) {
		return signals.stream().map(s -> s.getSignalSize()).reduce((long)0, (a, b) -> a + b);
	}
}
