/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.problem.graph;

import static org.fortiss.af3.exploration.alg.feature.scheduling.problem.graph.ScheduledTask.createScheduledTask;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.AcyclicTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.SignalAdapterWeightedEdge;

/**
 * An {@link ITaskGraph} that captures the dependencies of the schedulable instances of
 * {@link ITaskAdapter}s. It is an unrolled graph that contains as many instances as required to
 * fill a hyper period of the overall system schedule.
 * 
 * @author diewald
 */
public class ScheduleTaskGraph extends DefaultTaskGraph<IScheduleTask, ScheduleTaskGraph> {

	/** Constructor. */
	public ScheduleTaskGraph(AcyclicTaskGraph atg) {
		super();
		constructBasicScheduleGraph(atg);
	}

	/** Creates the underlying graph from the given acyclic task graph. */
	private void constructBasicScheduleGraph(AcyclicTaskGraph atg) {
		Map<ITaskAdapter, IScheduleTask> taskAssocs = new HashMap<>();
		for(ITaskAdapter task : atg.getTasks()) {
			IScheduleTask schedTask = createScheduledTask(task);
			graph.addVertex(schedTask);
			taskAssocs.put(task, schedTask);
		}

		for(SignalAdapterWeightedEdge edge : atg.getGraph().edgeSet()) {
			ITaskAdapter sender = atg.getGraph().getEdgeSource(edge);
			ITaskAdapter target = atg.getGraph().getEdgeTarget(edge);
			IScheduleTask schedSender = taskAssocs.get(sender);
			IScheduleTask schedTarget = taskAssocs.get(target);
			// TODO(AD): Do we need to construct a new edge here?
			graph.addEdge(schedSender, schedTarget,
					new SignalAdapterWeightedEdge(edge.getSignalAdapter()));
		}
	}

	/**
	 * Transforms the underlying graph into an unrolled on such that a full system schedule spanning
	 * over the systems hyperperiod is represented.
	 */
	private void transformToMultiRate(AcyclicTaskGraph atg) {
		// TODO(AD): Here, we need to add the handling of period constraints and the unrolling of
		// the task graph.
	}

	/**
	 * Returns the set of {@link IScheduleTask}s that do not have any dependencies on the prior
	 * execution of some other {@link IScheduleTask}s.
	 */
	public Set<IScheduleTask> getInitialTasks() {
		Set<IScheduleTask> initialTasks = new HashSet<>();
		for(IScheduleTask task : graph.vertexSet()) {
			if(graph.incomingEdgesOf(task).isEmpty()) {
				initialTasks.add(task);
			}
		}
		return initialTasks;
	}
}
