/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.problem.graph;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Arrays;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;

/**
 * Implementation of an {@link IScheduleElement} using the delegate pattern.
 * 
 * @author diewald
 */
public class ScheduledTask implements InvocationHandler, IScheduleElement {

	/** Actual {@link ITaskAdapter} behind this schedule wrapper. */
	protected ITaskAdapter originTask;

	/** Hiding constructor. */
	private ScheduledTask(ITaskAdapter originTask) {
		this.originTask = originTask;
	}

	/**
	 * Constructs a {@link ScheduledTask} instance.
	 * 
	 * @param originTask
	 *            delegate {@link ITaskAdapter} to be scheduled.
	 * @return a {@link ScheduledTask} instance of the given {@link ITaskAdapter}.
	 */
	public static IScheduleTask createScheduledTask(ITaskAdapter originTask) {
		return (IScheduleTask)Proxy.newProxyInstance(ITaskAdapter.class.getClassLoader(),
				new Class[] {IScheduleTask.class, ITaskAdapter.class},
				new ScheduledTask(originTask));
	}

	/** {@inheritDoc} */
	// TODO(AD): Extract this method to a parent class for composing objects. Esp. the
	// getWrappedObject call.
	@Override
	public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
		Class<?>[] argTypes = null;
		if(args != null) {
			Arrays.stream(args).map(a -> a.getClass()).toArray(size -> new Class[size]);
		}

		if(method.getName().equals("getWrappedObject")) {
			return originTask;
		}

		// Object-local (unique) methods.
		try {
			Method delegateMethod = getClass().getMethod(method.getName(), argTypes);
			return delegateMethod.invoke(this, args);
		} catch(NoSuchMethodException e) {
			// Not an error... continue...
			ignoreException();
		}

		try {
			Method delegateMethod = ITaskAdapter.class.getMethod(method.getName(), argTypes);
			return delegateMethod.invoke(originTask, args);
		} catch(NoSuchMethodException e) {
			// Not an error... continue...
			ignoreException();
		}

		return method.invoke(this, args);
	}

	/** {@inheritDoc} */
	@Override
	public Double getDeadline() {
		// TODO(AD): We must consider E2E deadlines here!
		Double periodDeadline = originTask.getPeriod() * (getInstanceCount() + 1);
		return periodDeadline;
	}

	/** {@inheritDoc} */
	@Override
	public Double getOffset() {
		// TODO(AD): This needs to be passed through the input layer first.
		return 0.0;
	}

	/** {@inheritDoc} */
	@Override
	public Double getReadyTime() {
		return getInstanceCount() * originTask.getPeriod() + getOffset();
	}

	/** {@inheritDoc} */
	@Override
	public Integer getInstanceCount() {
		// Must be adjusted for the multirate case.
		return 0;
	}

	/** Just a dummy method to avoid unneeded warnings if an empty catch clause is desired. */
	private void ignoreException() {
		// Dummy.
	}
}
