/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.scheduler;

import static java.util.stream.Collectors.toList;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.math.BigDecimal;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.core.runtime.Assert;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.feature.scheduling.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.feature.scheduling.TimeSlot;
import org.fortiss.af3.exploration.alg.feature.scheduling.problem.SharedMemAccessLatencyEncoding;
import org.fortiss.af3.exploration.alg.feature.scheduling.problem.graph.IScheduleElement;
import org.fortiss.af3.exploration.alg.feature.scheduling.problem.graph.IScheduleTask;
import org.fortiss.af3.exploration.alg.feature.scheduling.problem.graph.ScheduleTaskGraph;
import org.fortiss.af3.exploration.alg.feature.scheduling.problem.graph.ScheduleTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.common.encoding.IMappingEncoding;

/**
 * Base class for EDF task schedulers.
 * 
 * @author huang
 * @author diewald
 */
public abstract class EDFSchedulerBase<T extends IExecutionUnitAdapter, E extends TaskMappingEntry<ITaskAdapter, T>, M extends IMappingEncoding<ITaskAdapter, T, E, ?>>
		extends SchedulerBase {

	/** Remembers which {@link IScheduleTask} was scheduled into which {@link TimeSlot}. */
	private Map<IScheduleTask, TimeSlot> taskToSlot = new HashMap<>();

	/**
	 * Collection of tasks that have already been added to the {@link StrictTTSchedule} in the
	 * scheduling loop.
	 */
	private Collection<IScheduleTask> scheduledTasks = new HashSet<>();

	/** Compares two {@link IScheduleElement}s by their deadlines. */
	private class DeadlineComparator implements Comparator<IScheduleElement> {

		/** {@inheritDoc} */
		@Override
		public int compare(IScheduleElement obj1, IScheduleElement obj2) {
			// edf
			if(obj1.getDeadline() < obj2.getDeadline()) {
				return -1;
			} else if(obj1.getDeadline() < obj2.getDeadline()) {
				return 1;
			} else {
				return 0;
			}
		}
	}

	/**
	 * Creates and returns a release queue for the given {@link ScheduleTaskGraphEncoding}. A
	 * release queue contains those {@link IScheduleTask}s which remain to be scheduled. It is
	 * initially constructed from those schedule elements that do not have predecessors.
	 */
	private SortedSet<IScheduleTask> createReadyQueue(ScheduleTaskGraphEncoding stgEnc) {
		TreeSet<IScheduleTask> releaseQueue = new TreeSet<IScheduleTask>(new DeadlineComparator());

		for(ScheduleTaskGraph stg : stgEnc.getGraphs()) {
			releaseQueue.addAll(stg.getInitialTasks());
		}

		return releaseQueue;
	}

	/**
	 * Simple EDF scheduler. It works as follows:
	 * <ul>
	 * <li>the task graph with shortest deadline goes first</li>
	 * <li>the tasks within the task graph are scheduled according to topological order</li>
	 * </ul>
	 */
	public StrictTTSchedule<T, E, M> schedule(ScheduleTaskGraphEncoding stgEnc, M taskMapping,
			SharedMemAccessLatencyEncoding shMemEnc) {
		StrictTTSchedule<T, E, M> schedule =
				new StrictTTSchedule<T, E, M>(taskMapping, BigDecimal.ZERO, (long)0);

		SortedSet<IScheduleTask> readyQueue = createReadyQueue(stgEnc);

		// maintain the current time on each resource
		Map<IExecutionUnitAdapter, Double> currentTimes =
				new HashMap<IExecutionUnitAdapter, Double>();
		taskMapping.getResources().forEach(r -> currentTimes.put(r, Double.valueOf(0.0)));

		// the main loop
		while(!readyQueue.isEmpty()) {
			// Pick first task in the ready queue...
			IScheduleTask scheduleTask = readyQueue.first();
			// ...and schedule it
			scheduleObject(scheduleTask, stgEnc, taskMapping, shMemEnc, schedule, currentTimes);

			// Check if any successor becomes ready
			ScheduleTaskGraph stg = stgEnc.getGraphOf(scheduleTask);
			for(IScheduleTask succTask : stg.getSinkTasks(scheduleTask)) {
				Collection<IScheduleTask> predTasksOfSuccessor = stg.getSourceTasks(succTask);
				if(scheduledTasks.containsAll(predTasksOfSuccessor)) {
					readyQueue.add(succTask);
				}
			}

			// Book keeping: Remove the scheduled tasks from the ready list, mark it scheduled.
			readyQueue.remove(scheduleTask);
			scheduledTasks.add(scheduleTask);
		}

		if(!schedule.validateSchedule(true)) {
			throw new RuntimeException("Schedule is not valid...");
		}

		return schedule;
	}

	/** Inserts a single {@link IScheduleTask} into the {@link StrictTTSchedule} */
	private Double scheduleObject(IScheduleTask scheduleTask, ScheduleTaskGraphEncoding stgEnc,
			M encoding, SharedMemAccessLatencyEncoding shMemEncoding,
			StrictTTSchedule<T, E, M> schedule, Map<IExecutionUnitAdapter, Double> currentTimeMap) {

		ScheduleTaskGraph stg = stgEnc.getGraphOf(scheduleTask);

		// Only support 1:1 mappings for now.
		E mappingEntry =
				getFirst(encoding.getMappingEntriesOf(scheduleTask.getWrappedObject())).get();
		IExecutionUnitAdapter execUnit = mappingEntry.getTarget();
		// IExecutionUnitAdapter hardwareTarget =
		// (IExecutionUnitAdapter)getFirst(partEnc.getAllocatedResourcesFor(partition))
		// .get();

		// All task-time related calculations.
		Double startTime = currentTimeMap.get(execUnit);
		Double readyTime = scheduleTask.getReadyTime();
		// The start time should be no less than the ready time
		startTime = (startTime > readyTime) ? startTime : readyTime;
		// Add execution time
		Double execTime = getDuration(
				(IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?>)encoding,
				shMemEncoding, scheduleTask, mappingEntry);
		Assert.isTrue(execTime > 0.0);
		Double finishTime = startTime + execTime;

		// Build time slot
		Collection<TimeSlot> predSlots = stg.getSourceTasks(scheduleTask).stream()
				.map(t -> taskToSlot.get(t)).collect(toList());
		TimeSlot timeSlot = schedule.createTimeSlot(mappingEntry, execUnit, predSlots, startTime,
				finishTime, scheduleTask.getInstanceCount());

		// Register the generated TimeSlot at its predecessors.
		predSlots.forEach(s -> s.addSuccessor(timeSlot));

		// Book keeping
		currentTimeMap.put(execUnit, finishTime);

		return finishTime;
	}

	/** Returns the execution time for the given {@link ITaskAdapter}. */
	protected abstract Double getDuration(
			IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?> encoding,
			SharedMemAccessLatencyEncoding shMemEnc, IScheduleTask schedTask, E mappingEntry);
}
