/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.transform;

import static java.util.stream.Collectors.toList;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSuperSet;
import static org.fortiss.af3.platform.utils.PlatformModelElementFactory.createRoute;
import static org.fortiss.af3.platform.utils.PlatformModelElementFactory.createSegment;
import static org.fortiss.af3.task.util.TaskModelElementFactory.createSignalToRouteAllocationEntry;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.feature.scheduling.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.Message;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.MessageEncoding;
import org.fortiss.af3.exploration.alg.service.module.base.transformation.ExplorationTransformationModuleBase;
import org.fortiss.af3.exploration.alg.service.module.base.transformation.Transforms;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.platform.model.Route;
import org.fortiss.af3.platform.model.Segment;
import org.fortiss.af3.schedule.model.ResourceAllocation;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.TaskOutputPort;
import org.fortiss.af3.task.model.allocation.SignalToRouteAllocationEntry;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.traverse.TopologicalOrderIterator;

import com.google.inject.Provides;

/**
 * Transforms a {@link StrictTTSchedule} into the set of corresponding {@link ResourceAllocation}s.
 * 
 * @author diewald
 */
public class MessageToRouteTransformation
		extends ExplorationTransformationModuleBase<SuperSet<SignalToRouteAllocationEntry>> {

	/**
	 * See {@link MessageToRouteTransformation}.
	 * 
	 * @param msgEnc
	 *            encoding of the messages sent between the {@link ITaskAdapter}s.
	 * @param taskMapper
	 *            maps {@link ITaskAdapter}s to their source elements of the input system model.
	 * @return SuperSet of AF3 {@link ResourceAllocation}s.
	 */
	@Transforms
	@Provides
	public <N extends EObject, S extends EObject> SuperSet<SignalToRouteAllocationEntry>
			transform(MessageEncoding msgEnc, TaskMapper taskMapper) {
		SuperSet<SignalToRouteAllocationEntry> sigToRouteAllocs =
				createSuperSet("SigToRoute", SignalToRouteAllocationEntry.class);

		Map<Message, Route> msgToRouteMap = new HashMap<>();
		for(Message msg : msgEnc.getAllMessages()) {
			Route route = createRoute();
			msgToRouteMap.put(msg, route);

			DefaultDirectedGraph<IResourceAdapter, DefaultEdge> graph = msg.getRoute();
			TopologicalOrderIterator<IResourceAdapter, DefaultEdge> topoIter =
					new TopologicalOrderIterator<>(graph);

			Map<IResourceAdapter, Segment> resourceSegmentMap = new HashMap<>();
			while(topoIter.hasNext()) {
				IResourceAdapter resAdp = topoIter.next();
				Segment segment = createSegment(resAdp.getObject(IModelElement.class));
				resourceSegmentMap.put(resAdp, segment);

				// Tree structure: There is only one predecessor.
				Optional<DefaultEdge> incoming = getFirst(graph.incomingEdgesOf(resAdp));
				if(route.getRootSegment() == null) {
					route.setRootSegment(segment);
				} else {
					// A predecessor must always be present here.
					IResourceAdapter predecessor = graph.getEdgeSource(incoming.get());
					Segment prevSegment = resourceSegmentMap.get(predecessor);
					prevSegment.getNext().add(segment);
				}
			}
		}

		for(Entry<Message, Route> entry : msgToRouteMap.entrySet()) {
			Message msg = entry.getKey();
			ISignalAdapter sigAdp = msg.getSignalAdapter();
			TaskOutputPort sendingPort = sigAdp.getObject(TaskOutputPort.class);
			SignalToRouteAllocationEntry sigAlloc = null;

			Collection<IRequestAdapter> recvTasks = msg.getReceiverMappingEntries().stream()
					.map(e -> e.getSource()).collect(toList());

			for(Signal signal : sendingPort.getOutGoingSignals()) {
				Task recvAF3Task = (Task)signal.getTargetTaskPort().getOwner();
				ITaskAdapter task = taskMapper.getAdapter(recvAF3Task);
				if(!recvTasks.contains(task)) {
					// Do not consider signals that are exchanged on the same resources.
					continue;
				}

				if(sigAlloc == null) {
					sigAlloc = createSignalToRouteAllocationEntry(signal, entry.getValue());
					sigToRouteAllocs.getEntries().add(sigAlloc);
				} else {
					sigAlloc.getSignals().add(signal);
				}
			}
		}

		return sigToRouteAllocs;
	}
}
