/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.transform;

import static java.math.BigDecimal.valueOf;
import static org.eclipse.core.runtime.Assert.isNotNull;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSuperSet;
import static org.fortiss.af3.schedule.utils.AF3ScheduleModelElementFactory.createPeriodicTimeTrigger;
import static org.fortiss.af3.schedule.utils.AF3ScheduleModelElementFactory.createResourceAllocation;
import static org.fortiss.af3.schedule.utils.AF3ScheduleModelElementFactory.createSchedulableEntity;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.feature.scheduling.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.feature.scheduling.TimeSlot;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.Message;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.service.module.base.transformation.ExplorationTransformationModuleBase;
import org.fortiss.af3.exploration.alg.service.module.base.transformation.Transforms;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.platform.model.Route;
import org.fortiss.af3.platform.model.Segment;
import org.fortiss.af3.schedule.model.PeriodicTimeTrigger;
import org.fortiss.af3.schedule.model.ResourceAllocation;
import org.fortiss.af3.schedule.model.SchedulableEntity;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.TaskOutputPort;
import org.fortiss.af3.task.model.allocation.SignalToRouteAllocationEntry;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.model.INamedElement;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;
import com.google.inject.Provides;

/**
 * Transforms a {@link StrictTTSchedule} into the set of corresponding {@link ResourceAllocation}s.
 * 
 * @author diewald
 */
public class ResourceAllocationTransformation
		extends ExplorationTransformationModuleBase<SuperSet<ResourceAllocation>> {

	/**
	 * See {@link ResourceAllocationTransformation}.
	 * 
	 * @param schedule
	 *            strictly time-triggered schedule of resource allocations.
	 * @param sigToRoute
	 *            allocations of {@link Signal}s to the {@link Route}s they traverse to their
	 *            receiver.
	 * @return SuperSet of AF3 {@link ResourceAllocation}s.
	 */
	@SuppressWarnings("unchecked")
	@Transforms
	@Provides
	public <N extends EObject, S extends EObject> SuperSet<ResourceAllocation> transform(
			@SuppressWarnings("rawtypes") StrictTTSchedule schedule,
			SuperSet<SignalToRouteAllocationEntry> sigToRoute) {
		SuperSet<ResourceAllocation> resAllocSet =
				createSuperSet("ResAllocs", ResourceAllocation.class);

		Table<Message, IResourceAdapter, Collection<Signal>> segmentLUT = HashBasedTable.create();

		// TODO(AD): This casting is actually not needed. It seems that the compiler is going mad
		// here due to the raw StricTTSchedule parameter.
		for(TimeSlot timeSlot : (Collection<TimeSlot>)(Collection<?>)schedule.getTimeSlots()) {
			double period = 0.0;
			// Element creation & association.
			IModelElement schedElement = null;
			if(timeSlot.getDeployedRequester() instanceof TaskMappingEntry) {
				ITaskAdapter schedTask = timeSlot.getDeployedRequester().getSource();
				schedElement = schedTask.getObject(IModelElement.class);
				period = schedTask.getPeriod();
			} else {
				// TODO(AD): The deployed requester is actually a Segment. Here, the
				// TDMABusScheduler must be updated to operate on segments and the segment must be a
				// IRequestAdapter.
				// For now we are "spreading" the multicast messages based on TaskOutputPorts to
				// Signals in the following code.
				schedElement = extractScheduledSegment(sigToRoute, segmentLUT, timeSlot);
			}

			String name = schedElement.toString();
			if(schedElement instanceof INamedElement) {
				name = ((INamedElement)schedElement).getName();
			}

			ResourceAllocation resAlloc = createResourceAllocation(name);
			SchedulableEntity schedEntity = createSchedulableEntity(name);
			schedEntity.setModelElement(schedElement);
			resAlloc.setSchedulableEntity(schedEntity);

			// Timing information.
			PeriodicTimeTrigger trigger =
					createPeriodicTimeTrigger(valueOf(period), valueOf(timeSlot.getStartTime()));
			resAlloc.setTrigger(trigger);
			resAlloc.setDuration(valueOf(timeSlot.getDuration()));

			resAllocSet.getEntries().add(resAlloc);
		}

		return resAllocSet;
	}

	/**
	 * Returns the {@link Segment} that corresponds to the scheduled {@link TimeSlot} of a
	 * communication.
	 */
	private IModelElement extractScheduledSegment(SuperSet<SignalToRouteAllocationEntry> sigToRoute,
			Table<Message, IResourceAdapter, Collection<Signal>> segmentLUT, TimeSlot timeSlot) {
		Message schedMsg = timeSlot.getDeployedMessage();
		IResourceAdapter resource = timeSlot.getResource();
		TaskOutputPort msgPort = schedMsg.getSignalAdapter().getObject(TaskOutputPort.class);
		IModelElement targetReq = getFirst(schedMsg.getReceiverMappingEntries()).get().getSource()
				.getObject(IModelElement.class);

		Collection<Signal> handleSignals = segmentLUT.get(schedMsg, resource);
		handleSignals = handleSignals != null ? handleSignals : new HashSet<>();

		Signal schedSignal = null;
		for(Signal msgSignal : msgPort.getOutGoingSignals()) {
			if(targetReq == msgSignal.getTarget().getOwner()) {
				schedSignal = msgSignal;
				handleSignals.add(schedSignal);
			}
		}
		isNotNull(schedSignal);

		final Signal finSchedSignal = schedSignal;
		Route route = sigToRoute.getEntries().stream()
				.filter(sR -> sR.getSignals().contains(finSchedSignal)).findAny().get().getRoute();

		Segment schedSegment = route.getAllSegments().stream()
				.filter(s -> s.getEntity() == resource.getObject(IModelElement.class)).findAny()
				.get();

		// Bookkeeping
		segmentLUT.put(schedMsg, resource, handleSignals);

		return schedSegment;
	}
}
