/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.taskmapping.problem;

import java.util.ArrayList;
import java.util.Collection;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IModelElementAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.SignalAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.feature.safetyarch.variable.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.feature.safetyarch.variable.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.feature.taskinstantiation.problem.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskinstantiation.problem.InstantiatedTaskGraphEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskinstantiation.variable.TaskInstanceEncoding;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.common.encoding.IMappingEntry;
import org.opt4j.core.problem.Decoder;

/**
 * {@link Decoder} that takes an {@link TaskInstanceEncoding} as input and generates an
 * {@link InstantiatedTaskGraphEncoding} from it.
 * 
 * @author diewald
 */
public abstract class TaskMappingToGraphDecoderBase<G extends ITaskGraph<ITaskAdapter>, I extends InstantiatedTaskGraphEncodingBase<G>, T extends IExecutionUnitAdapter, E extends IMappingEntry<ITaskAdapter, T>>
		extends DecoderModule<I> {

	/**
	 * Replaces vertices in the underlying graph that are abstract and thus require instantiation
	 * acc. to the input {@link TaskInstanceEncoding}.
	 */
	protected void instantiateTaskGraphs(I encoding, TaskInstanceEncoding tInstEnc) {
		for(ITaskGraph<ITaskAdapter> atg : encoding.getTaskGraphs()) {
			for(ITaskAdapter task : atg.getTasks()) {
				if(tInstEnc.getAbstractElements().contains(task)) {
					ITaskAdapter replacement = tInstEnc.getInstanceOf(task);
					atg.replaceTask(task, replacement);
				}
			}
		}
	}

	/**
	 * Removes signals from the {@link InstantiatedTaskGraphEncoding} that were introduced by
	 * replications but which connect independent safety channels. This is not desired, so remove
	 * all edges (signals) in the task graph that induce connections between the safety channels.
	 */
	// TODO: The removal of signals between safety channels should not be needed here. Instead,
	// those signals should be already removed in the given AbstractTaskGraphEncoding.
	private void removeInterSafetyChannelSignals(SafetyFunctionArchEncoding sfEnc,
			InstantiatedTaskGraphEncoding itgEnc) throws DecodingException {
		for(SafetyFunctionArchEntry sfEntry : sfEnc.getAllEntries()) {
			@SuppressWarnings("unchecked") Collection<ITaskAdapter> sfCompAdps =
					(Collection<ITaskAdapter>)(Collection<?>)(sfEntry.getSafetyFunctionAdapter()
							.getRealizingComponents());
			assert (sfCompAdps.size() > 0);
			ITaskAdapter sfCompAdp = sfCompAdps.stream().findAny().get();

			// Get the "containing" task graph. We must check all replacement components (instances)
			// since we don't know which one was instantiated (or it would be more costly to find
			// out).
			ITaskGraph<ITaskAdapter> sfTG = null;
			for(ITaskAdapter curSFCompAdp : sfCompAdp.getReplacementComponents()) {
				sfTG = itgEnc.getTaskGraphOf(curSFCompAdp);
				if(sfTG != null) {
					break;
				}
			}
			if(sfTG == null) {
				throw new DecodingException(sfEnc, this, "Could not retrieve the TaskGraph" +
						" of the Component Adapter " + sfCompAdp.getName() + ".");
			}

			Collection<ITaskAdapter> connectedSFUnits = new ArrayList<>();
			Collection<ITaskAdapter> connectedDiagUnits = new ArrayList<>();
			Collection<ITaskAdapter> connectedOutUnit = new ArrayList<>();
			for(int channelIdx = 0; channelIdx < sfEntry.getNumChannels(); channelIdx++) {
				@SuppressWarnings("unchecked") Collection<ITaskAdapter> sfUnits =
						(Collection<ITaskAdapter>)(Collection<?>)sfEntry.getSafetyFunctionAdapter()
								.getRealizingComponents();
				@SuppressWarnings("unchecked") Collection<ITaskAdapter> diagUnits =
						(Collection<ITaskAdapter>)(Collection<?>)sfEntry.getSafetyFunctionAdapter()
								.getDiagnosticComponents();
				// TODO: knowledge about the modeL: output unit not instantiated:
				ITaskAdapter outputUnit = sfEntry.getSafetyFunctionAdapter().getOutputComponent();
				// TODO: only for a single diag unit.
				Collection<ITaskAdapter> realizingSFUnits = itgEnc.getTasks().stream()
						.filter(d -> sfUnits.contains(((ITaskAdapter)d).getReplacedComponent()))
						.collect(Collectors.toList());
				Collection<ITaskAdapter> realizingDiagUnits = itgEnc.getTasks().stream()
						.filter(d -> diagUnits.contains(((ITaskAdapter)d).getReplacedComponent()))
						.collect(Collectors.toList());
				Collection<ITaskAdapter> outComps = itgEnc
						.getTasks().stream().filter(d -> ((IModelElementAdapter)d)
								.getObject(Object.class) == outputUnit.getObject(Object.class))
						.collect(Collectors.toList());

				// Pick an sf component instance that is assigned to the current channel.
				ITaskAdapter currentChannelSFUnit = null;
				for(ITaskAdapter curSFUnit : realizingSFUnits) {
					if(!connectedSFUnits.contains(curSFUnit)) {
						// Remember that the diagnostic unit is associated with one channel --> no
						// elimination of the signal.
						connectedSFUnits.add(curSFUnit);
						currentChannelSFUnit = curSFUnit;
						break;
					}
				}
				// Pick an diagnostic component instance that is assigned to the current channel.
				ITaskAdapter currentChannelDiagUnit = null;
				for(ITaskAdapter curDiagUnit : realizingDiagUnits) {
					if(!connectedDiagUnits.contains(curDiagUnit)) {
						// Remember that the diagnostic unit is associated with one channel --> no
						// elimination of the signal.
						connectedDiagUnits.add(curDiagUnit);
						currentChannelDiagUnit = curDiagUnit;
						break;
					}
				}
				// Pick an io component instance that is assigned to the current channel.
				ITaskAdapter currentChannelOutputUnit = null;
				for(ITaskAdapter curOutUnit : outComps) {
					if(!connectedOutUnit.contains(curOutUnit)) {
						// Remember that the diagnostic unit is associated with one channel --> no
						// elimination of the signal.
						connectedDiagUnits.add(curOutUnit);
						currentChannelOutputUnit = curOutUnit;
						break;
					}
				}

				updateTaskGraph(sfTG, realizingSFUnits, realizingDiagUnits, outComps,
						currentChannelSFUnit, currentChannelDiagUnit, currentChannelOutputUnit);
			}
		}
	}

	/**
	 * Updates the given {@link DefaultTaskGraph} to match the number of channels.
	 * 
	 * @param sfTG
	 *            {@link ITaskGraph} containing the safety function instances.
	 * @param realizingSFUnits
	 *            {@link ITaskAdapter}s realizing the safety function's protection
	 *            element.
	 * @param realizingDiagUnits
	 *            {@link ITaskAdapter}s realizing the safety function's diagnostic
	 *            element.
	 * @param outComps
	 *            {@link ITaskAdapter}s realizing the safety function's output
	 *            aggregation element, e.g., a voter.
	 * @param currentChannelSFUnit
	 *            {@link ITaskAdapter} of the currently processed protection element.
	 * @param currentChannelDiagUnit
	 *            {@link ITaskAdapter} of the currently processed diagnostic element.
	 * @param currentChannelOutputUnit
	 *            {@link ITaskAdapter} of the currently processed output aggregation
	 *            element.
	 */
	private void updateTaskGraph(ITaskGraph<ITaskAdapter> sfTG,
			Collection<ITaskAdapter> realizingSFUnits, Collection<ITaskAdapter> realizingDiagUnits,
			Collection<ITaskAdapter> outComps, ITaskAdapter currentChannelSFUnit,
			ITaskAdapter currentChannelDiagUnit, ITaskAdapter currentChannelOutputUnit) {
		for(ITaskAdapter curSFUnit : realizingSFUnits) {
			if(currentChannelSFUnit != curSFUnit) {
				// Eliminate all other signals to different channels.
				SignalAdapterWeightedEdge interChannelSignal =
						sfTG.getGraph().getEdge(currentChannelDiagUnit, curSFUnit);
				sfTG.getGraph().removeEdge(interChannelSignal);
				interChannelSignal = sfTG.getGraph().getEdge(currentChannelOutputUnit, curSFUnit);
				sfTG.getGraph().removeEdge(interChannelSignal);
			}
		}
		for(ITaskAdapter curDiagUnit : realizingDiagUnits) {
			if(currentChannelDiagUnit != curDiagUnit) {
				// Eliminate all other signals to different channels.
				SignalAdapterWeightedEdge interChannelSignal =
						sfTG.getGraph().getEdge(currentChannelSFUnit, curDiagUnit);
				sfTG.getGraph().removeEdge(interChannelSignal);
				interChannelSignal = sfTG.getGraph().getEdge(currentChannelOutputUnit, curDiagUnit);
				sfTG.getGraph().removeEdge(interChannelSignal);
			}
		}
		for(ITaskAdapter curOutUnit : outComps) {
			if(currentChannelOutputUnit != curOutUnit) {
				// Eliminate all other signals to different channels.
				SignalAdapterWeightedEdge interChannelSignal =
						sfTG.getGraph().getEdge(currentChannelSFUnit, curOutUnit);
				sfTG.getGraph().removeEdge(interChannelSignal);
				interChannelSignal = sfTG.getGraph().getEdge(currentChannelDiagUnit, curOutUnit);
				sfTG.getGraph().removeEdge(interChannelSignal);
			}
		}
	}
}
