/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.encoding;

import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toSet;
import static org.eclipse.core.runtime.Assert.isTrue;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.DSE_VALIDATE_ENCODING;
import static org.fortiss.tooling.common.util.LambdaUtils.filterByType;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.RandomExploration;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.PhenotypeBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.moea.model.constraints.AssignmentConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.CouplingConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.HashBasedTable;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Table;
import com.google.common.collect.Table.Cell;

/**
 * Base class for m:n allocation encodings/tables.
 * 
 * @author diewald
 */
public abstract class MappingEncodingBase<S extends IRequestAdapter, T extends IResourceAdapter, E extends IMappingEntry<S, T>, M extends IMappingEncoding<S, T, E, M>>
		extends PhenotypeBase implements IMappingEncoding<S, T, E, M> {

	/** References the {@link Random} number generator singleton of the DSE. */
	protected RandomExploration rand = RandomExploration.getInstance();

	/**
	 * Allocation table.
	 * <li>key1: The Requester (e.g., task)</li>
	 * <li>key2: The target resource (e.g., cores)</li>
	 * <li>value: Mapping entries. Multiple entry in the list represent redundant mapping</li>
	 */
	protected Table<S, T, List<E>> allocationTable = HashBasedTable.create();

	/** Map of disallowed {@link ITaskAdapter} -> {@link IExecutionUnitAdapter} allocations. */
	protected ListMultimap<S, T> validAllocations = ArrayListMultimap.create();

	/** References the set of {@link IRequestAdapter}s to be allocated. */
	protected Set<S> requesters = new HashSet<>();

	/** References the set of {@link IResourceAdapter}s serving as allocation targets. */
	protected Set<T> resources = new HashSet<>();

	/** Constructor. */
	// TODO(AD): Check, if we can "attach" a setConstraint method, and if it would be useful at all.
	protected MappingEncodingBase(Set<S> requesters, Set<T> resources,
			Collection<InternalConstraint<?>> constraints) {
		updateValidAllocations(requesters, resources, constraints);
	}

	/** Copy constructor. */
	protected MappingEncodingBase(MappingEncodingBase<S, T, E, M> encoding) {
		Map<E, E> copyAssocMap = new HashMap<>();
		updateValidAllocations(encoding.getRequesters(), encoding.getResources(),
				encoding.constraints);
		this.allocationTable = copyAllocationMap(encoding.allocationTable, copyAssocMap);
		this.constraints =
				encoding.constraints.stream().map(c -> EcoreUtils.copy(c)).collect(toList());
	}

	/**
	 * Creates a copy of the internal allocation map of {@code this}
	 * {@link TaskMappingEncodingBase}.
	 */
	@SuppressWarnings("unchecked")
	protected Table<S, T, List<E>> copyAllocationMap(Table<S, T, List<E>> origTable,
			Map<E, E> copyAssocMap) {
		Table<S, T, List<E>> copiedTable = HashBasedTable.create();
		for(Cell<S, T, List<E>> tmEntries : origTable.cellSet()) {
			List<E> copiedEntries = new LinkedList<>();
			for(E entry : tmEntries.getValue()) {
				E mappingEntry = (E)entry.copy();
				copiedEntries.add(mappingEntry);
				copyAssocMap.put(entry, mappingEntry);
			}
			copiedTable.put(tmEntries.getRowKey(), tmEntries.getColumnKey(), copiedEntries);
		}
		return copiedTable;
	}

	/**
	 * Updates the map of valid {@link IRequestAdapter} to {@link IResourceAdapter} allocations.
	 * Hereby, allocation constraints such as "allocate only to" or "do not allocate to" are
	 * considered.
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void updateValidAllocations(Set<S> requesters, Set<T> resources,
			Collection<InternalConstraint<?>> constraints) {
		this.requesters.clear();
		this.resources.clear();
		this.requesters.addAll(requesters);
		this.resources.addAll(resources);
		this.constraints = new ArrayList<>(constraints);

		validAllocations.clear();

		Multimap<S, T> constrainedAllocs = HashMultimap.create();
		for(AssignmentConstraint ac : filterByType(constraints, AssignmentConstraint.class)) {
			// Unchecked --> The framework ensures that we get the correct types here.
			// TODO(AD): Place a check in the transformator to check for overlapping constraints.
			ac.getSources()
					.forEach(s -> constrainedAllocs.putAll((S)s, (Collection<T>)ac.getTargets()));
		}

		for(S requester : requesters) {
			if(constrainedAllocs.containsKey(requester)) {
				validAllocations.putAll(requester, constrainedAllocs.get(requester));
			} else {
				validAllocations.putAll(requester, resources);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<T> getValidTargetResources(S source) {
		CouplingConstraint cConstr = getConstraint(CouplingConstraint.class, source);
		if(cConstr != null) {
			Optional<S> existingSrc =
					getFirst(cConstr.getSetOf(source), s -> allocationTable.containsRow(s));
			if(existingSrc.isPresent()) {
				Set<T> tgtSet = getMappingEntriesOf(existingSrc.get()).stream()
						.map(e -> e.getTarget()).collect(toSet());
				return ImmutableList.copyOf(tgtSet);
			}
		}
		return validAllocations.get(source);
	}

	/**
	 * Adds the given entry to the internal allocation table.
	 * 
	 * @param entry
	 *            {@link IMappingEntry} to be added to the table.
	 */
	protected void addAllocation(E entry) {
		List<E> mappingEntries = allocationTable.get(entry.getSource(), entry.getTarget());
		if(mappingEntries == null) {
			mappingEntries = new LinkedList<>();
			allocationTable.put(entry.getSource(), entry.getTarget(), mappingEntries);
		}
		mappingEntries.add(entry);
	}

	/**
	 * Constructs an {@IMappingEntry} for the given requester. The deployment targets can be given
	 * as a list, otherwise the deployment targets are taken from the platform.
	 */
	protected E createEntry(S requester, List<T> deploymentTargets) {
		// Select from remaining allowed deployment targets if no set of desired deployment targets
		// is given
		if(deploymentTargets.isEmpty()) {
			deploymentTargets = new ArrayList<T>();
			deploymentTargets.addAll(getValidTargetResources(requester));
		}

		// TODO(AD): This one has to be considered in MappingEncoding: getValidTargets.
		// deploymentTargets.removeAll(getExcludedDeploymentTargets(deployableComponent));
		if(deploymentTargets.isEmpty()) {
			throw createNoTargetsException(requester);
		}
		T target = deploymentTargets.get(rand.nextInt(deploymentTargets.size()));
		return createEntry(requester, target);
	}

	/**
	 * Creates an (encoding specific) mapping entry the describes that a given
	 * {@link IResourceAdapter} is mapped to a deployment target.
	 */
	protected abstract E createEntry(S requester, T target);

	/** {@inheritDoc} */
	@Override
	public E createEntry(S requester) {
		E newEntry = createEntry(requester, emptyList());
		addAllocation(newEntry);
		return newEntry;
	}

	/** {@inheritDoc} */
	@Override
	public E createEntry(S requester, E existingEntry) {
		if(allocationTable.get(requester, existingEntry.getTarget()) != existingEntry) {
			throw new IllegalArgumentException("The mapping entry " + existingEntry +
					" is not part of the encoding " + this + ".");
		}
		E newEntry = createEntry(requester, existingEntry.getTarget());
		addAllocation(newEntry);
		return newEntry;
	}

	/** Removes the given {@code entry} from the allocation table. */
	@Override
	public void removeEntry(E entry) {
		List<E> mappingEntries = allocationTable.get(entry.getSource(), entry.getTarget());
		mappingEntries.remove(entry);
	}

	/** {@inheritDoc} */
	@Override
	public void resetResourceOf(E entry, T resource) {
		CouplingConstraint cConstr = getConstraint(CouplingConstraint.class, entry.getSource());
		if(DSE_VALIDATE_ENCODING && cConstr == null) {
			isTrue(getValidTargetResources(entry.getSource()).contains(resource));
		}
		if(cConstr == null) {
			resetEntry(entry, resource);
		} else {
			// Migrate all entries belonging to the same CouplingConstraint.
			Collection<S> setRequesters = cConstr.getSetOf(entry.getSource());
			for(S requester : setRequesters) {
				getMappingEntriesOf(requester).forEach(e -> resetEntry(e, resource));
			}
		}
	}

	/**
	 * Allocates the given {@code entry} to the given {@code resource}. Therby, the
	 * {@link IMappingEntry} and the {@link #allocationTable} are updated.
	 */
	protected void resetEntry(E entry, T resource) {
		removeEntry(entry);
		entry.resetTarget(resource);
		addAllocation(entry);
	}

	/** {@inheritDoc} */
	@Override
	public List<E> getMappingEntriesOf(S source) {
		return allocationTable.row(source).values().parallelStream().flatMap(c -> c.stream())
				.collect(toList());
	}

	/** {@inheritDoc} */
	@Override
	public List<E> getMappingEntriesOf(T target) {
		return allocationTable.column(target).values().parallelStream().flatMap(c -> c.stream())
				.collect(toList());
	}

	/** {@inheritDoc} */
	@Override
	public Collection<E> getMappingEntries() {
		return allocationTable.values().parallelStream().flatMap(c -> c.stream()).collect(toList());
	}

	/** {@inheritDoc} */
	@Override
	public List<S> getRequesters(T resource) {
		return allocationTable.column(resource).keySet().stream().collect(toList());
	}

	/** {@inheritDoc} */
	@Override
	public List<T> getResources(S requester) {
		return allocationTable.row(requester).keySet().stream().collect(toList());
	}

	/** {@inheritDoc} */
	@Override
	public ImmutableSet<S> getRequesters() {
		return ImmutableSet.copyOf(requesters);
	}

	/** {@inheritDoc} */
	@Override
	public ImmutableSet<T> getResources() {
		return ImmutableSet.copyOf(resources);
	}

	/** {@inheritDoc} */
	@Override
	public int size() {
		return allocationTable.size();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isValidAllocations() {
		for(E entry : getMappingEntries()) {
			List<T> validTargets = getValidTargetResources(entry.getSource());
			if(!validTargets.contains(entry.getTarget())) {
				return false;
			}
		}
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String str = "";
		for(S requester : getRequesters()) {
			str += "Mapping: " + requester.getName() + " -> " + "{";
			if(allocationTable.row(requester) != null) {
				for(E tgt : getMappingEntriesOf(requester)) {
					str += tgt.getTarget().getName() + ", ";
				}
			}
			str += "}\n";
		}
		return str;
	}

	/** Creates an exception indicating an over-constrained problem. */
	private ExplorationRuntimeException createNoTargetsException(S requester) {
		return new ExplorationRuntimeException("Could not create an allocation entry for the" +
				" requester " + requester.getName() + " due to missing resources for its" +
				" allocation. This typically implies an over-constrained problem.");
	}
}
