/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.evaluator.mapping;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Genotyped;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Evaluate;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Validates;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.feature.partition.variable.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.partition.variable.TaskPartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.module.common.encoding.IMappingEncoding;
import org.fortiss.af3.exploration.alg.module.common.evaluator.BooleanValue;
import org.fortiss.af3.exploration.dseml.model.expression.IExpression;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.smt.model.SMTConstraint;
import org.opt4j.core.Value;

import com.google.inject.Inject;
import com.google.inject.Provides;

/**
 * Evaluator for {@link IExpression}s of {@link ExplorationConstraint}s. The actual evaluation
 * function resides in the evaluate method.
 * 
 * @author diewald
 */
public class MappingEvaluatorConstraint extends
		EvaluatorModule<Boolean, ExplorationConstraint<Boolean>, TaskMappingEncodingBase<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?>> {

	/** Prototype constructor. */
	public MappingEvaluatorConstraint(
			Class<? extends ExplorationConstraint<Boolean>> explorationTargetType) {
		super(explorationTargetType);
	}

	/**
	 * Constructor for execution: Defines the systemModelAdapter abstraction and the
	 * {@link ExplorationTarget} to evaluate.
	 */
	public MappingEvaluatorConstraint(SMTConstraint explorationTarget) {
		super(explorationTarget);
	}

	/** Container for optional dependencies. */
	protected static class OptionalVirtMapping {
		/** Optional virtual mapping (Partitions) */
		@Inject(optional = true)
		@Genotyped
		TaskPartitionMappingEncoding virtMapping;
	}

	/**
	 * Evaluates the {@link IExpression} given in the {@link ExplorationConstraint} subject to the
	 * given {@link TaskPartitionMappingEncoding} and {@link PartitionMappingEncoding} and returns
	 * the result. Therefore, this evaluation calls the {@link MappingParserConstraint} and provides
	 * a boolean whether the expression is satisfied.
	 */
	@Provides
	@Evaluate(MappingEvaluatorConstraint.class)
	public BooleanValue evaluate(
			IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?> mapping,
			OptionalVirtMapping virtMappingContainer, TaskMapper taskMapper) {
		@SuppressWarnings("unchecked") IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?> virtMapping =
				(IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?>)(IMappingEncoding<?, ?, ?, ?>)virtMappingContainer.virtMapping;
		MappingParserConstraint parser = new MappingParserConstraint(mapping, virtMapping,
				explorationTarget.getExpression(), taskMapper);
		try {
			return new BooleanValue(parser.parse());
		} catch(Exception e) {
			throw new ExplorationRuntimeException(
					"Failed to parse the expression " + explorationTarget.getName() + ".", e);
		}
	}

	/** Validates the {@link ExplorationConstraint} versus the penalty value. */
	@Provides
	@Validates
	public Boolean validate(Value<?> value) {
		Double evalValue = value.getDouble();
		// evalValue is a penalty, i.e. the constraint is satisfied as soon as the penalty is zero.
		if(evalValue == 0.0) {
			return false;
		}
		return true;
	}
}
