/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.input;

import static java.util.stream.Collectors.toSet;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemAccessTask.AER_MEM_TYPE;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemoryUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3MemAccessTask;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3TaskAdapter;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.module.InputProvider;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.allocation.ComponentToTaskAllocationEntry;

import com.google.inject.Provides;
import com.google.inject.Singleton;

/**
 * Module that provides the {@link ITaskAdapter}s extracted from the source task architecture.
 * 
 * @author diewald
 */
public class DeployableTaskModule extends InputModule<Set<ITaskAdapter>> {

	/** Provides the {@link ITaskAdapter}s of the task architecture. */
	@Provides
	@InputProvider
	@Singleton
	public Set<ITaskAdapter> getTasks(SuperSet<Task> af3Tasks,
			SuperSet<ComponentToTaskAllocationEntry> componentTaskAllocs,
			Set<IComponentAdapter> components, Set<IExecutionUnitAdapter> execUnits,
			Set<IMemoryUnitAdapter> memUnits) {
		Set<ITaskAdapter> tasks = new HashSet<>();
		for(Task currentTask : af3Tasks.getEntries()) {
			// @CodeFormatterOff
			Set<Component> assocAF3Components = componentTaskAllocs.getEntries().stream()
					.filter(e -> e.getTargetElement() == currentTask) // Get only relevant entries
					.map(e -> e.getSourceElements()) // Trace back to mapped components
					.flatMap(List::stream) // Unfold the list (aggregate)
					.map(Component.class::cast) // Fixup the generic IModelElement return type
					.collect(toSet());
			// @CodeFormatterOn
			Collection<IComponentAdapter> assocComponents = components.stream()
					.filter(c -> assocAF3Components.contains(c.getObject(Component.class)))
					.collect(toSet());

			if(currentTask.getName().contains("READ")) {
				tasks.add(new AF3MemAccessTask(currentTask, assocComponents, execUnits, memUnits,
						AER_MEM_TYPE.ACQUISITION));
			} else if(currentTask.getName().contains("WRITE")) {
				tasks.add(new AF3MemAccessTask(currentTask, assocComponents, execUnits, memUnits,
						AER_MEM_TYPE.RESTITUTION));
			} else {
				tasks.add(new AF3TaskAdapter(currentTask, assocComponents, execUnits));
			}
		}
		return tasks;
	}
}
