/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.input;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3ResourceConnectionAdapter;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.module.InputProvider;
import org.fortiss.af3.exploration.alg.util.AF3Utils;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.GatewayUnit;
import org.fortiss.af3.platform.model.MemoryUnit;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.inject.Module;
import com.google.inject.Provides;
import com.google.inject.Singleton;

/**
 * {@link Module} providing the set of extracted {@link IResourceConnectionAdapter}s of the target
 * platform.
 * 
 * @author diewald
 */
public class ResourceConnectionModule extends InputModule<Set<IResourceConnectionAdapter>> {

	/**
	 * Provides the {@link IResourceConnectionAdapter}s connecting {@link IResourceAdapter}s of the
	 * target platform.
	 */
	@InputProvider
	@Provides
	@Singleton
	public Set<IResourceConnectionAdapter> getResourceConnections(
			Set<IResourceAdapter> allResources, SuperSet<ExecutionUnit> af3ExecUnits,
			SuperSet<MemoryUnit> af3MemUnits) {

		Set<IResourceConnectionAdapter> resourceConnections = new HashSet<>();

		// Initial traversal list of IConnectors that are leaves of the platform graph, i.e.
		// communication endpoints. Currently, those are the target execution units and memory
		// units.
		LinkedList<IConnector> deployableConnectors = new LinkedList<IConnector>();
		af3ExecUnits.getEntries().forEach(e -> deployableConnectors.addAll(e.getConnectors()));
		af3MemUnits.getEntries().forEach(m -> deployableConnectors.addAll(m.getConnectors()));

		// TODO: rewrite the code below. Handling of hull elements and virtual targets should be
		// considered in a single util method. Based on the collected target connectors, the
		// resourceconnectionadapters should be generated.
		Collection<IConnector> visitedConnectors = new HashSet<IConnector>();
		while(!deployableConnectors.isEmpty()) {
			IConnector currentConnector = deployableConnectors.poll();
			visitedConnectors.add(currentConnector);
			// NOTE: Here, source element does not imply "source of the connection", but
			// "source of the traversal".
			IHierarchicElement sourceElement = currentConnector.getOwner();
			if(sourceElement instanceof ExecutionUnit || sourceElement instanceof GatewayUnit ||
					sourceElement instanceof TransmissionUnit ||
					sourceElement instanceof MemoryUnit) {
				// Path walk to determine the *real* connection targets, i.e. ignore hull elements
				// like Tiles whose contained Cores (or partitions) were selected as deployment
				// targets.
				Collection<IConnector> targetConnectors = AF3Utils
						.getEffectiveConnectionTargets(currentConnector, af3ExecUnits.getEntries());

				// Create an IResourceConnectionAdapter between the sourceElement of the traversal
				// and the *real* targets.
				AF3ResourceConnectionAdapter connAdp =
						createResourceConnection(allResources, currentConnector, targetConnectors);
				if(connAdp != null) {
					resourceConnections.add(connAdp);
				}
				visitedConnectors.addAll(targetConnectors);

				// Collect the IConnectors attached to the target elements of the created
				// IResourceConnectionAdapter for further traversing the PlatformArchitecture.
				//
				// TODO Connections between IModelElements of different PlatformArchitectures which
				// used to be realized via ResourceLink specifications were automatically considered
				// (ResourceLink has been removed in favor of af3.allocation).
				Collection<IHierarchicElement> targetElements = new HashSet<IHierarchicElement>();
				for(IConnector currentTargetConnector : targetConnectors) {
					targetElements.add(currentTargetConnector.getOwner());
				}
				AF3Utils.determineNextConnectionTargets(deployableConnectors, visitedConnectors,
						targetElements, af3ExecUnits.getEntries());
			}
		}

		return resourceConnections;
	}

	/**
	 * Creates a {@link IResourceConnectionAdapter} between the given source and the target
	 * elements. This method takes {@link IHierarchicElement}s as inputs and identifies the matching
	 * {@link IResourceConnectionAdapter}s of the given elements.
	 */
	private AF3ResourceConnectionAdapter createResourceConnection(
			Set<IResourceAdapter> allResources, IConnector sourceConnector,
			Collection<IConnector> targetConnectors) {
		IResourceAdapter sourceResource = null;
		Collection<IResourceAdapter> targetResources = new HashSet<IResourceAdapter>();
		Map<IResourceAdapter, ConnectionType> targetConnType =
				new HashMap<IResourceAdapter, ConnectionType>();
		Collection<IHierarchicElement> targetElements = new HashSet<IHierarchicElement>();

		// Capture information about the target elements and the IConnectors that are the actual AF3
		// targets of this connection. It is used to determine the connection direction for each
		// target.
		Multimap<IHierarchicElement, IConnector> targetConnsAtElement = HashMultimap.create();
		for(IConnector targetConnector : targetConnectors) {
			targetElements.add(targetConnector.getOwner());
			targetConnsAtElement.put(targetConnector.getOwner(), targetConnector);
		}

		// Identify the corresponding IResourceAdapters of the collection of target
		// IHierarchicalElements and determine the direction of their connection.
		for(IResourceAdapter resource : allResources) {
			if(resource.getObject(IModelElement.class) == sourceConnector.getOwner()) {
				sourceResource = resource;
			} else if(targetElements.contains(resource.getObject(IModelElement.class))) {
				targetResources.add(resource);
				ConnectionType type = ConnectionType.BIDIRECTIONAL;
				boolean firstRoundFlag = true;
				for(IConnector targetConnector : targetConnsAtElement
						.get((IHierarchicElement)resource.getObject(IModelElement.class))) {
					if(firstRoundFlag) {
						firstRoundFlag = false;
						type = AF3Utils.getConnectionTypeDirection(sourceConnector,
								targetConnector);
					} else if(type != AF3Utils.getConnectionTypeDirection(sourceConnector,
							targetConnector)) {
						type = ConnectionType.BIDIRECTIONAL;
					}
				}
				targetConnType.put(resource, type);
			}
		}
		assert (sourceResource != null);

		return new AF3ResourceConnectionAdapter(sourceResource, targetConnType);
	}
}
