/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.input;

import static org.fortiss.af3.schedule.utils.MathUtils.lcmDoubleCollection;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.SignalAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.module.InputProvider;
import org.jgrapht.Graphs;
import org.jgrapht.alg.connectivity.ConnectivityInspector;

import com.google.inject.Provides;
import com.google.inject.Singleton;

/**
 * Constructs a {@link TaskGraphEncoding} from an input task architecture that contains the set of
 * independent {@link TaskGraph}s.
 * 
 * @author diewald
 */
public class TaskGraphInputModule extends InputModule<TaskGraphEncoding> {

	/**
	 * Constant defining the allowed tolerance for the calculation of the greatest common divisor
	 * used in the calculation of the effective periods.
	 */
	protected final static int GCD_PRECISION = 3;

	/** See {@link TaskGraphInputModule}. */
	@InputProvider
	@Provides
	@Singleton
	public TaskGraphEncoding getTaskGraphEnc(Set<ITaskAdapter> tasks, Set<ISignalAdapter> signals) {
		DefaultTaskGraph<ITaskAdapter, ?> monolithicTaskGraph = new TaskGraph();
		monolithicTaskGraph.addAllTask(tasks);
		monolithicTaskGraph.addAllSignals(signals);

		Set<TaskGraph> independentTaskGraphs = new HashSet<>();
		// Compute vertex sets of maximum connected components of temporary task graph
		ConnectivityInspector<ITaskAdapter, SignalAdapterWeightedEdge> ci =
				new ConnectivityInspector<ITaskAdapter, SignalAdapterWeightedEdge>(
						monolithicTaskGraph.getGenericGraph());
		List<Set<ITaskAdapter>> connectedSets = ci.connectedSets();
		// Reconstruct TaskGraphs from vertex sets of connected components
		for(Set<ITaskAdapter> connectedSet : connectedSets) {
			TaskGraph taskGraph = new TaskGraph();
			Graphs.addAllVertices(taskGraph.getGraph(), connectedSet);
			for(ISignalAdapter signal : signals) {
				if(connectedSet.contains(signal.getSource())) {
					taskGraph.addSignal(signal);
				}
			}
			// Check if tasks within current sub TaskGraph share the same common period
			taskGraph.updateHyperPeriod();
			independentTaskGraphs.add(taskGraph);
		}

		Double hyperPeriod = determineHyperPeriod(independentTaskGraphs);

		return new TaskGraphEncoding(independentTaskGraphs, hyperPeriod);
	}

	/**
	 * Calculates and returns the hyperperiod of the set of all given {@link ITaskGraph}s by
	 * calculating the least common multiple from the hyperperiods of the {@link ITaskGraph}s.
	 */
	private Double determineHyperPeriod(Collection<TaskGraph> allTaskGraphs) {
		Set<Double> taskgraphPeriods = new TreeSet<Double>();
		for(ITaskGraph<?> currentGraph : allTaskGraphs) {
			Double period = currentGraph.getHyperPeriod();
			if(period != null && period > 0) {
				taskgraphPeriods.add(period);
			}
		}

		if(taskgraphPeriods.isEmpty()) {
			return null;
		}
		return lcmDoubleCollection(taskgraphPeriods, GCD_PRECISION);
	}
}
