/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.operator.crossover.mapping;

import static java.lang.Math.max;
import static java.lang.Math.min;
import static org.eclipse.core.runtime.Assert.isTrue;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.DSE_VALIDATE_ENCODING;

import java.util.ArrayList;
import java.util.List;

import org.fortiss.af3.exploration.alg.dse.RandomExploration;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.opt4j.operators.crossover.Crossover;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;

/**
 * Base class for crossover operations on the allocations of {@link ITaskAdapter}s to
 * {@link IExecutionUnitAdapter}s. It requires two parent {@link TaskMappingEncodingBase}s whose
 * allocations are swapped.
 * 
 * @author huang
 * @author diewald
 */
public abstract class CrossoverAllocationBase<T extends IExecutionUnitAdapter, E extends TaskMappingEntry<ITaskAdapter, T>, M extends TaskMappingEncodingBase<ITaskAdapter, T, E, ?>>
		implements Crossover<M> {

	/** Random generator */
	protected RandomExploration rand = RandomExploration.getInstance();

	/**
	 * Performs the actual crossover (swapping) operation between the offspring encodings {#link
	 * newEncoding1} and {#link newEncoding2} for the allocations/mapping of the given
	 * {@link ITaskAdapter}.
	 */
	protected void crossoverAllocation(ITaskAdapter deployableComponent, M encoding1,
			M encoding2) {
		List<E> newAssignedResources1 =
				new ArrayList<>(encoding1.getMappingEntriesOf(deployableComponent));
		List<E> newAssignedResources2 =
				new ArrayList<>(encoding2.getMappingEntriesOf(deployableComponent));

		// Only swap as many entries as the smaller list has -> avoid inconsistencies due to wrong
		// replication number.
		BiMap<E, E> swapMap = HashBiMap.create();
		int size1 = newAssignedResources1.size();
		int size2 = newAssignedResources2.size();
		int sizeDiff = max(size1, size2) - min(size1, size2);
		int offset = rand.nextInt(sizeDiff + 1);
		for(int i = 0; i < min(size1, size2); i++) {
			E elem1 = size1 > size2 ? newAssignedResources1.get(i + offset)
					: newAssignedResources1.get(i);
			E elem2 = size2 > size1 ? newAssignedResources2.get(i + offset)
					: newAssignedResources2.get(i);
			swapMap.put(elem1, elem2);
			if(size1 > size2) {
				newAssignedResources1.set(i + offset, elem2);
				newAssignedResources2.set(i, elem1);
			} else {
				newAssignedResources1.set(i, elem2);
				newAssignedResources2.set(i + offset, elem1);
			}
		}

		encoding1.setAssignedResources(deployableComponent, newAssignedResources1);
		encoding2.setAssignedResources(deployableComponent, newAssignedResources2);

		if(DSE_VALIDATE_ENCODING) {
			isTrue(encoding1.isValidAllocations());
			isTrue(encoding2.isValidAllocations());
		}
	}
}
