/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.operator.mutate.mapping;

import static java.lang.Math.max;
import static org.eclipse.core.runtime.Assert.isTrue;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.DSE_VALIDATE_ENCODING;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.feature.partition.variable.Partition;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;

import com.google.inject.Inject;

/**
 * Mutate operator manipulates the allocation of tasks to resources by replacing an existing
 * allocation with a new random allocation.
 * 
 * @author huang
 */
// FIXME(AD): Use the generic types here!!!
public class MutateAllocation<S extends TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, T extends TaskMappingEncodingBase<ITaskAdapter, IExecutionUnitAdapter, S, ?>>
		extends MutateTasksBase<S, T> {
	/** Constructs a new allocation mutation operator for a given exploration run */
	@Inject
	public MutateAllocation(Set<ITaskAdapter> tasks) {
		super(tasks);
	}

	/** {@inheritDoc} */
	@Override
	public void mutate(T encoding, double rate) {
		for(ITaskAdapter task : encoding.getRequesters()) {
			if(rand.nextDouble() < rate) {
				// Perform the re-allocation
				try {
					reAllocateMappingEntry(encoding, task);
				} catch(ExplorationException e) {
					// Opt4J does not allow to throw Exceptions here.
					throw new RuntimeException(e);
				}
			}
		}

		if(DSE_VALIDATE_ENCODING) {
			isTrue(encoding.isValidAllocations());
			isTrue(checkReplicationCount(encoding));
		}
	}

	/** Checks the replication count and prints a message in case of errors. */
	private boolean checkReplicationCount(T encoding) {
		for(ITaskAdapter task : encoding.getRequesters()) {
			int allocResNum = encoding.getResources(task).size();
			if(allocResNum != 1) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Performs the actual re-allocation of the given {@link ITaskAdapter}.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	private void reAllocateMappingEntry(T encoding, ITaskAdapter task) throws ExplorationException {
		// The assigned resources must be an indexed list, such that randomly determined allocations
		// can be directly removed using a random number generator.
		List<S> assignedResources = new ArrayList<>(encoding.getMappingEntriesOf(task));

		// Reassign a random allocation...
		// max(...) is required as nextInt does not support a bound of 0.
		int selIdx = max(0, rand.nextInt(assignedResources.size()) - 1);
		S selectedEntry = assignedResources.get(selIdx);

		// This respects {@link ExcludeDeploymentConstraint}s.
		@SuppressWarnings("unchecked") List<Partition> targets =
				(List<Partition>)(List<?>)encoding.getValidTargetResources(task);
		if(DSE_VALIDATE_ENCODING) {
			isTrue(targets.size() > 0);
		}

		if(targets.size() > 1) {
			int selExecU = max(0, rand.nextInt(targets.size()) - 1);
			IExecutionUnitAdapter tgtResource = targets.get(selExecU);
			encoding.resetResourceOf(selectedEntry, tgtResource);
		}
	}
}
