/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWith;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Predicate;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExpSpecTransformationService;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecTransformationModule;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;

import com.google.common.collect.BiMap;
import com.google.inject.Injector;

/**
 * Provides the {@link IExpSpecTransformationService}.
 * 
 * @author diewald
 */
public class ExpSpecTransformationService implements IExpSpecTransformationService {

	/** Set of {@link IExpSpecTransformationModule}s registered in this module */
	protected Collection<IExpSpecTransformationModule<? extends EObject, ? extends EObject>> modules =
			new ArrayList<>();

	/** Holds the instance of the evaluation service. */
	private volatile static ExpSpecTransformationService INSTANCE;

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExpSpecTransformationService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExpSpecTransformationService();
		}
		return INSTANCE;
	}

	/** Constructor. */
	private ExpSpecTransformationService() {
		// Prevent direct instantiation.
	}

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public <I extends EObject, O extends EObject> List<ExplorationTarget<?>> transform(
			Injector mainInjector, List<ExplorationTarget<?>> expTargets,
			List<ExplorationTarget<?>> origExpTargets, BiMap<EObject, EObject> copyRefMap)
			throws ExplorationException {
		for(IExpSpecTransformationModule<? extends EObject, ? extends EObject> module : modules) {
			Map<EObject, EObject> transformedMap = new HashMap<>();

			Collection<I> inputElements = getApplicableElements(module, expTargets);
			for(I inputElement : inputElements) {
				@SuppressWarnings("unchecked") IExpSpecTransformationModule<I, O> castedModule =
						(IExpSpecTransformationModule<I, O>)module;
				O out = castedModule.transform(mainInjector, inputElement);
				transformedMap.put(inputElement, out);
			}

			if(module.outputIsReplacement()) {
				// NOTE: If a transformation module takes an ExplorationTarget as input and modifies
				// it, there is nothing to do here. This code ensures proper handling of
				// sub-expression modifications.
				for(Entry<EObject, EObject> tEntry : transformedMap.entrySet()) {
					// We are using the original specification for identifying references. Thus, we
					// must use the original elements here.
					EObject origEntry = copyRefMap.inverse().get(tEntry.getKey());
					Collection<Setting> settings =
							EcoreUtil.UsageCrossReferencer.find(origEntry, origExpTargets);
					for(Setting ref : settings) {
						ref.set(tEntry.getValue());
					}
				}
			} else {
				for(Entry<EObject, EObject> tEntry : transformedMap.entrySet()) {
					EObject value = tEntry.getValue();
					if(value instanceof ExplorationTarget) {
						expTargets.add((ExplorationTarget<?>)tEntry.getValue());
					} else if(value != null) {
						throw new ExplorationException("The module " +
								module.getClass().getSimpleName() + " is incorrectly specified. " +
								"If the output elements are additional, they must be " +
								"ExplorationTargets.");
					}
				}
			}
		}

		return expTargets;
	}

	/**
	 * Returns a set of elements that shall undergo a transformation. They are extracted from the
	 * set of given {@link ExplorationTarget}s.
	 * Returned elements satisfy the input filter {@link Predicate} of the given module, or are
	 * assignable from the input type of the module if no {@link Predicate} is given.
	 */
	@SuppressWarnings("unchecked")
	private <T extends EObject> Collection<T> getApplicableElements(
			IExpSpecTransformationModule<? extends EObject, ? extends EObject> module,
			Collection<ExplorationTarget<?>> targets) {
		Collection<EObject> inputElements = new ArrayList<>();
		Class<T> inputType = (Class<T>)module.getInputType();
		Predicate<? extends EObject> pred = module.getInputFilter();
		Predicate<? extends EObject> finPred =
				pred != null ? pred : e -> inputType.isAssignableFrom(e.getClass());
		// We examine both ExplorationTargets and their children to handle target generation and
		// modification.
		targets.stream().filter((Predicate<? super ExplorationTarget<?>>)finPred)
				.forEach(inputElements::add);
		targets.forEach(t -> inputElements.addAll(getChildrenWith(t, (Predicate<EObject>)finPred)));
		return (Collection<T>)inputElements;
	}

	/** {@inheritDoc} */
	@Override
	public void registerModule(
			IExpSpecTransformationModule<? extends EObject, ? extends EObject> module) {
		modules.add(module);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return "The " + getClass().getSimpleName() + " service allows to define transformation " +
				"transformation rules for " + ExplorationSpecification.class.getSimpleName() +
				"that are executed before the D3SE is actually launched. These rules are encoded " +
				"in " + IExpSpecTransformationModule.class.getSimpleName() + "s that may be " +
				"composed at arbitrary depth levels. These modules are allowed to use " +
				IExpSpecTransformationModule.class.getSimpleName() + "s that process one element " +
				"per execution provided by the parent " +
				IExpSpecTransformationModule.class.getSimpleName() + ".";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Exploration Specification Transformation Service";
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
