/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static org.eclipse.core.runtime.Assert.isTrue;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.common.util.LambdaUtils.isAssignableFromAny;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.impl.EObjectImpl;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.alg.service.module.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.inject.Module;

/**
 * Implementation of the {@link IExplorationEvaluatationService} which handles the instantiation of
 * the Evaluators based on the Exploration target. The evaluators are registered as extensions.
 * 
 * @author diewald
 */
public class ExplorationEvaluationService implements IExplorationEvaluatationService {

	/** Holds the instance of the evaluation service. */
	private volatile static ExplorationEvaluationService INSTANCE;

	/**
	 * Associates the registered {@link Evaluator}s with their required types from the
	 * {@link IExplorationTransformationService}
	 */
	private Multimap<Class<IExplorationTargetEvaluator<?, ?, ?>>, Class<?>> requiredTransformedModels =
			HashMultimap.create();

	/**
	 * Maps {@link ExplorationTarget} types to the {@link EvaluatorModule}s responsible for them.
	 */
	private Multimap<Class<ExplorationTarget<?>>, IEvaluatorModule<?, ?, ?>> evaluatorPrototypeMap =
			HashMultimap.create();

	/** Associates {@link EvaluatorModule}s with the {@link IExplorationFeature}s defining them. */
	private Map<IEvaluatorModule<?, ?, ?>, IExplorationFeature> evalFeatureAssoc = new HashMap<>();

	/** Constructor. */
	private ExplorationEvaluationService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExplorationEvaluationService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationEvaluationService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public void registerEvaluator(IEvaluatorModule<?, ?, ?> evaluatorModule,
			IExplorationFeature feature) {
		// ExplorationTargets are EMF EObjects, so we receive an impl class here. It's interface
		// shall be registered.
		Class<? extends EObject> regClass = evaluatorModule.getExplorationTargetType();
		if(EObjectImpl.class.isAssignableFrom(regClass) && regClass.getInterfaces().length == 0) {
			throw new ExplorationRuntimeException(
					"The evaluator " + evaluatorModule.getClass().getSimpleName() +
							" registers for an implementation class of an ExplorationTarget type." +
							" This is not supported. Instead, use the generated interface" +
							" specifying the ExplorationTarget.");
		}
		evaluatorPrototypeMap.put(
				(Class<ExplorationTarget<?>>)evaluatorModule.getExplorationTargetType(),
				evaluatorModule);
		evalFeatureAssoc.put(evaluatorModule, feature);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Collection<Class<?>>
			getRequiredTransformedModels(Set<IExplorationTargetEvaluator<?, ?, ?>> evaluators) {
		Set<Class<?>> retVal = new HashSet<>();
		evaluators.forEach(e -> retVal.addAll(requiredTransformedModels
				.get((Class<IExplorationTargetEvaluator<?, ?, ?>>)e.getClass())));
		return retVal;
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		Collection<Class<?>> allRequiredInputs = new HashSet<>();

		// Pre-processing code for the Model Transformation Service.
		// for(IEvaluatorModule<?, ?, ?> evaluator : getInstance().getAllHandlers()) {
		// allRequiredInputs.addAll(evaluator.getRequiredInputTypes());
//			// @CodeFormatterOff
//			Collection<Class<?>> curNonPhenotypeInputTypes =
//					filterSet(allRequiredInputs,
//							t -> !(Phenotype.class.isAssignableFrom(t)
//								   || Genotype.class.isAssignableFrom(t)));
//			// @CodeFormatterOn
		//
		// // // TODO: check also if the phenotypes exist.
		// // try {
		// // checkInputConsistency(curNonPhenotypeInputTypes);
		// // } catch(ExplorationServiceException e) {
		// // throw new RuntimeException("Error starting the ExplorationEvaluationService.", e);
		// // }
		//
		// // Here, we assume that all *checked* (see above) remaining models must be provided by
		// // the transformation service.
		// @SuppressWarnings("unchecked") Class<IExplorationTargetEvaluator<?, ?, ?>> evaluatorType
		// =
		// (Class<IExplorationTargetEvaluator<?, ?, ?>>)evaluator.getClass();
		// requiredTransformedModels.putAll(evaluatorType, curNonPhenotypeInputTypes);
		// }
	}

	// /**
	// * Returns whether the given collection of non-Phenotype types can be obtained by the
	// * {@link IExplorationTransformationService}.
	// */
	// private void checkInputConsistency(Collection<Class<?>> requiredNonPhenoInputs)
	// throws ExplorationServiceException {
	// for(Class<?> curInputType : requiredNonPhenoInputs) {
	// if(!getService(IExplorationTransformationService.class).canTransformOutputType(
	// curInputType)) {
	// throw new ExplorationServiceException(IExplorationEvaluatationService.class,
	// "The required input type " + curInputType.getSimpleName() +
	// " is not a Phenotype, nor can it be produced by the" +
	// " IExplorationTransformationService. Please check the required" +
	// " types and/or the registration of the needed ITransformationStrategies.");
	// }
	// }
	// }

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		// Nothing to be done here.
	}

	/** {@inheritDoc} */
	@Override
	public <R, T extends ExplorationTarget<R>> IEvaluatorModule<R, T, ?> getEvaluatorInstance(
			T explorationTarget, Collection<Class<? extends Module>> overriddenModules,
			Collection<IExplorationFeature> enabledFeatures) {
		@SuppressWarnings("unchecked") Class<ExplorationTarget<?>> identifier =
				(Class<ExplorationTarget<?>>)explorationTarget.getClass();
		if(EObjectImpl.class.isAssignableFrom(identifier)) {
			// We have an Impl class, get the interface which is used as the identifier. the
			// registration mechanism ensures to have only one interface here.
			@SuppressWarnings("unchecked") Class<ExplorationTarget<?>>[] interfaces =
					(Class<ExplorationTarget<?>>[])identifier.getInterfaces();
			identifier = interfaces[0];
		}

		@SuppressWarnings("unchecked") Collection<IEvaluatorModule<R, T, ?>> evaluatorPrototypes =
				new ArrayList<>(
						(Collection<IEvaluatorModule<R, T, ?>>)(Collection<?>)evaluatorPrototypeMap
								.get(identifier));
		if(explorationTarget instanceof InternalConstraint && evaluatorPrototypes.size() == 0) {
			return null;
		}

		isTrue(evaluatorPrototypes.size() > 0, "No evaluator prototype is registered for " +
				explorationTarget.getClass().getSimpleName() + ".");

		@SuppressWarnings("unchecked") Collection<IEvaluatorModule<R, T, ?>> rmList =
				evaluatorPrototypes.stream()
						.filter(eP -> isAssignableFromAny(
								(Collection<Class<?>>)(Collection<?>)overriddenModules, eP))
						.collect(Collectors.toList());

		for(IEvaluatorModule<R, T, ?> evalMod : evaluatorPrototypes) {
			IExplorationFeature assocFeature = evalFeatureAssoc.get(evalMod);
			if(!(assocFeature == null || enabledFeatures.contains(assocFeature))) {
				rmList.add(evalMod);
			}
		}
		evaluatorPrototypes.removeAll(rmList);

		isTrue(evaluatorPrototypes.size() > 0, "Error: All evaluators applicable to " +
				explorationTarget.getClass().getSimpleName() + " are overridden.");
		isTrue(evaluatorPrototypes.size() < 2,
				"Error: More than one evaluator is " + "registered for " +
						explorationTarget.getClass().getSimpleName() +
						" and not overridden by the selected features.");

		return getFirst(evaluatorPrototypes).get().newInstance(explorationTarget);
	}

	/** {@inheritDoc} */
	@Override
	public <R, T extends ExplorationTarget<R>> Set<IEvaluatorModule<?, ?, ?>> getEvaluatorInstances(
			Set<? extends ExplorationTarget<?>> explorationTarget,
			Comparator<IEvaluatorModule<?, ?, ?>> prioComparator,
			Collection<Class<? extends Module>> overriddenModules,
			Collection<IExplorationFeature> enabledFeatures) {
		Set<IEvaluatorModule<?, ?, ?>> evalModuleInstances = new TreeSet<>(prioComparator);

		for(ExplorationTarget<?> target : explorationTarget) {
			EvaluatorModule<?, ?, ?> evalModule =
					(EvaluatorModule<?, ?, ?>)getEvaluatorInstance(target, overriddenModules,
							enabledFeatures);
			if(evalModule != null) {
				evalModuleInstances.add(evalModule);
			}
		}

		return evalModuleInstances;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return "This services enables the registration of evaluators (quantifiers) to the DSE.";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Evaluator Service";
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
