/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.getMethodAnnotatedWith;
import static org.fortiss.af3.exploration.alg.util.ExplorationFeatureUtils.getEnabledDecoders;
import static org.fortiss.af3.exploration.alg.util.ExplorationFeatureUtils.getEnabledInputModules;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Logger;

import org.apache.commons.lang3.reflect.MethodUtils;
import org.fortiss.af3.exploration.alg.dse.InputProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Decodes;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.InputArtifact;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.module.InputProvider;
import org.fortiss.af3.exploration.alg.service.IExplorationInputService;
import org.fortiss.af3.exploration.alg.service.IExplorationService;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;

import com.google.inject.AbstractModule;
import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.Module;
import com.google.inject.TypeLiteral;
import com.google.inject.spi.Element;
import com.google.inject.spi.ProviderInstanceBinding;

/**
 * Implementation of the {@link IExplorationService} {@link IExplorationInputService}.
 * 
 * @author diewald
 */
public class ExplorationInputService implements IExplorationInputService {

	/** Holds the registered {@link IExplorationFeature}s. */
	private Map<TypeLiteral<?>, InputModule<?>> registeredInputModules = new HashMap<>();

	/** Holds the instance of the evaluation service. */
	private volatile static ExplorationInputService INSTANCE;

	/** Constructor. */
	private ExplorationInputService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExplorationInputService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationInputService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public void registerExplorationInput(InputModule<?> inputModule) {
		Method inputProvider = getMethodAnnotatedWith(inputModule.getClass(), InputProvider.class);
		TypeLiteral<?> providedType = TypeLiteral.get(inputProvider.getGenericReturnType());
		registeredInputModules.put(providedType, inputModule);
	}

	/** {@inheritDoc} */
	@Override
	public EncodingProviderModule<?> getInputProviderModule(
			Collection<IExplorationFeature> enabledFeatures,
			EncodingProviderModule<?> systemModelProviders) {
		Map<TypeLiteral<?>, InputModule<?>> allEnabledInputModules = new HashMap<>();
		allEnabledInputModules.putAll(registeredInputModules);
		allEnabledInputModules.putAll(getEnabledInputModules(enabledFeatures));

		// Collect all types produced by decoders to identify those that replace input artifacts.
		HashSet<TypeLiteral<?>> decodedTypes = new HashSet<>();
		for(DecoderModule<?> decoder : getEnabledDecoders(enabledFeatures)) {
			List<Method> decodeMethods =
					MethodUtils.getMethodsListWithAnnotation(decoder.getClass(), Decodes.class);
			for(Method annMethod : decodeMethods) {
				Type retType = annMethod.getGenericReturnType();
				decodedTypes.add(TypeLiteral.get(retType));
			}
		}

		// Create a map of inputs that must be marked explicitly to avoid conflicts with artifacts
		// produced by decoders.
		Map<TypeLiteral<?>, Class<? extends Annotation>> overriddenInputs = new HashMap<>();
		for(Entry<TypeLiteral<?>, InputModule<?>> inModule : allEnabledInputModules.entrySet()) {
			if(decodedTypes.contains(inModule.getKey())) {
				overriddenInputs.put(inModule.getKey(), InputArtifact.class);
			}
		}

		Map<TypeLiteral<? extends Object>, Object> inputElements =
				getInputArtifacts(systemModelProviders, allEnabledInputModules);

		InputProviderGenerator inProviderGen =
				new InputProviderGenerator(inputElements.keySet(), overriddenInputs);
		EncodingProviderModule<?> inProviderModule =
				inProviderGen.generateInstanceProviders(inputElements.entrySet());

		return inProviderModule;
	}

	/**
	 * Creates the input artifacts by querying the return type of each {@link InputModule}. The
	 * returned list also contains any transitive dependency such as the {@link Injector}.
	 */
	private Map<TypeLiteral<? extends Object>, Object> getInputArtifacts(
			EncodingProviderModule<?> systemModelProviders,
			Map<TypeLiteral<?>, InputModule<?>> selectedInputModules) {
		Collection<Module> modules = new ArrayList<>();
		modules.add(systemModelProviders);
		modules.add(new InputModuleHolder(selectedInputModules.values()));
		Injector inModInjector = Guice.createInjector(modules);

		// Create each object constructed by InputModules
		for(Entry<TypeLiteral<?>, InputModule<?>> inModule : selectedInputModules.entrySet()) {
			inModInjector.getInstance(Key.get(inModule.getKey()));
		}
		List<Element> constructedElements = inModInjector.getElements();

		// Filter out system model artifacts (SuperSet), and guice internal ones.
		Map<TypeLiteral<? extends Object>, Object> inputElements = new HashMap<>();
		for(Element element : constructedElements) {
			if(element instanceof ProviderInstanceBinding) {
				ProviderInstanceBinding<?> pInstBinding = (ProviderInstanceBinding<?>)element;
				TypeLiteral<?> typeLit = pInstBinding.getKey().getTypeLiteral();
				if(!SuperSet.class.isAssignableFrom(typeLit.getRawType()) &&
						!Injector.class.isAssignableFrom(typeLit.getRawType()) &&
						!Logger.class.isAssignableFrom(typeLit.getRawType())) {
					inputElements.put(typeLit, inModInjector.getInstance(pInstBinding.getKey()));
				}
			}
		}

		return inputElements;
	}

	/** Module defining the bindings of {@link InputModule}s. */
	private class InputModuleHolder extends AbstractModule {

		/** References the active {@link InputModule}s. */
		Collection<InputModule<?>> inputModules;

		/** Constructor. */
		public InputModuleHolder(Collection<InputModule<?>> inputModules) {
			this.inputModules = inputModules;
		}

		/** {@inheritDoc} */
		@Override
		protected void configure() {
			super.configure();
			for(InputModule<?> inModule : inputModules) {
				install(inModule);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
