/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal.transformation;

import static org.fortiss.af3.exploration.util.ExplorationModelElementFactory.createSuperSetMap;

import java.util.ArrayList;
import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.alg.service.module.IExplorationTransformationModule;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;

import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.TypeLiteral;

/**
 * Guice-based singleton implementation of the {@link IExplorationTransformationService}.
 * 
 * @author diewald
 */
public class ExplorationTransformationService implements IExplorationTransformationService {

	/** References all {@link IExplorationTransformationModule}s registered with this service. */
	private Collection<IExplorationTransformationModule<?>> registeredTransformators =
			new ArrayList<>();

	/** Holds the instance of the evaluation service. */
	private volatile static ExplorationTransformationService INSTANCE;

	/** Constructor. */
	private ExplorationTransformationService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExplorationTransformationService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationTransformationService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		// Nothing to do.
	}

	/** {@inheritDoc} */
	@Override
	public <T extends IModelElement> SuperSetMap transform(
			Collection<IExplorationFeature> enabledFeatures,
			EncodingProviderModule<?> inputProvider, ProvidingPhenotype phenotype,
			Collection<TypeLiteral<SuperSet<? extends IModelElement>>> outputSuperSets) {
		Collection<IExplorationTransformationModule<?>> tEnabledModules = new ArrayList<>();
		enabledFeatures.forEach(f -> tEnabledModules.addAll(f.getTransformationModules()));

		Collection<com.google.inject.Module> transformationModules =
				new ArrayList<>(tEnabledModules);
		TransformationInterceptor interceptor = new TransformationInterceptor();
		transformationModules.add(new TransformationInterceptModule(interceptor));
		transformationModules.add(inputProvider);
		transformationModules.add(phenotype.getPhenotypeProviderModule());
		transformationModules.add(phenotype.getGenotypeProviderModule());
		Injector interceptInjector = Guice.createInjector(transformationModules);

		Collection<com.google.inject.Module> tModulesAndInputs = new ArrayList<>();
		tModulesAndInputs.add(new SuperSetCache());
		tModulesAndInputs.add(inputProvider);
		tModulesAndInputs.add(phenotype.getPhenotypeProviderModule());
		tModulesAndInputs.add(phenotype.getGenotypeProviderModule());
		for(IExplorationTransformationModule<?> module : tEnabledModules) {
			tModulesAndInputs.add(interceptInjector.getInstance(module.getClass()));
		}
		Injector transformationInjector = Guice.createInjector(tModulesAndInputs);

		@SuppressWarnings({"unchecked",
				"rawtypes"}) Class<SuperSet<? extends IModelElement>> commonOutType =
						(Class)SuperSet.class;
		interceptor.setInterceptParameters(transformationInjector, null, commonOutType,
				SuperSetCache.class);

		for(TypeLiteral<SuperSet<? extends IModelElement>> outType : outputSuperSets) {
			transformationInjector.getInstance(Key.get(outType));
		}

		SuperSetCache results = transformationInjector.getInstance(SuperSetCache.class);
		SuperSetMap outSuperSetMap = createSuperSetMap();
		for(SuperSet<? extends IModelElement> superSet : results.values()) {
			@SuppressWarnings("unchecked") Class<T> key = (Class<T>)superSet.getEntryType();
			@SuppressWarnings("unchecked") SuperSet<T> castedSet = (SuperSet<T>)superSet;
			outSuperSetMap.put(key, castedSet);
		}
		return outSuperSetMap;
	}

	/** {@inheritDoc} */
	@Override
	public void registerTransformator(IExplorationTransformationModule<?> transformator) {
		registeredTransformators.add(transformator);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
