/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import java.util.Collection;
import java.util.Map;
import java.util.Optional;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.exception.ConstraintGenerationException;
import org.fortiss.af3.exploration.dseml.model.booleanp.allocation.Allocation;
import org.fortiss.af3.exploration.dseml.model.booleanp.allocation.Dislocation;
import org.fortiss.af3.exploration.dseml.model.booleanp.allocation.ILocationExpression;
import org.fortiss.af3.exploration.dseml.model.expression.Set;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationFactory;
import org.fortiss.af3.exploration.model.ExplorationObjective;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.MoeaFactory;
import org.fortiss.af3.exploration.moea.model.adapter.AdapterFactory;
import org.fortiss.af3.exploration.moea.model.adapter.IRequesterSet;
import org.fortiss.af3.exploration.moea.model.allocation.AllocationFactory;
import org.fortiss.af3.exploration.moea.model.allocation.InterferenceLatencyTable;
import org.fortiss.af3.exploration.moea.model.constraints.AssignmentConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.ConstraintsFactory;
import org.fortiss.af3.exploration.moea.model.constraints.CouplingConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.IsolatedCommunicationSet;
import org.fortiss.af3.exploration.moea.model.constraints.PeriodConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.ReplicationConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.SeparationConstraint;
import org.fortiss.af3.exploration.moea.model.dsmlext.DsmlextFactory;
import org.fortiss.af3.exploration.moea.model.dsmlext.DsmlextPackage;
import org.fortiss.af3.exploration.moea.model.dsmlext.FixedPropertyLiteral;
import org.fortiss.af3.exploration.moea.model.objectives.LatencyMinObjective;
import org.fortiss.af3.exploration.moea.model.objectives.ObjectivesFactory;
import org.fortiss.af3.exploration.moea.model.objectives.PartitionMinObjective;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.fortiss.af3.exploration.moea.model.parameters.ExecutionModel;
import org.fortiss.af3.exploration.moea.model.parameters.ParametersFactory;
import org.fortiss.af3.exploration.moea.model.parameters.SystemParameters;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.common.util.LambdaUtils;

import com.google.inject.TypeLiteral;

/**
 * Model element factory for design space exploration.
 * 
 * @author diewald
 */
public class DesignSpaceExplorationModelElementFactory {

	/** Creates an instance of a {@link DseSpecification} with default parameters. */
	public static DseSpecification createDseSpecification(String name) {
		DseSpecification dse = MoeaFactory.eINSTANCE.createDseSpecification();
		dse.setName(name);
		return dse;
	}

	/**
	 * Creates a DseSpecification and adds the {@link ExplorationSpecification} to it using the
	 * given {@code modelContext}.
	 */
	public static DseSpecification
			createDefaultDesignSpaceExploration(ExplorationSpecification expSpec) {
		DseSpecification dseSpec = createDseSpecification("Default DSE model",
				ExecutionModel.TT_FAIL_SILENCE, createDefaultEAOptimizer());
		dseSpec.setTargetSpecification(expSpec);
		return dseSpec;
	}

	/** Creates an instance of a {@link DseSpecification} with the given parameters. */
	public static DseSpecification createDseSpecification(String name, ExecutionModel execModel,
			EAOptimizer eaOptimizer) {
		DseSpecification dse = createDseSpecification(name);
		dse.setParameters(createSystemParameters(execModel));
		dse.setOptimizer(eaOptimizer);
		// A goal specification does not require any default values, so created here.
		dse.setTargetSpecification(ExplorationFactory.eINSTANCE.createExplorationSpecification());
		return dse;
	}

	/** "Wrapper" for creating an instance of default DSE parameters for the model. */
	public static SystemParameters createSystemParameters(ExecutionModel execModel) {
		SystemParameters dseParams = ParametersFactory.eINSTANCE.createSystemParameters();
		dseParams.setExecutionModel(execModel);
		return dseParams;
	}

	/** Creates an default EA optimizer configuration */
	public static EAOptimizer createDefaultEAOptimizer() {
		// Parameter defaults are preset in exploration.ecore.
		EAOptimizer optimizer = ParametersFactory.eINSTANCE.createEAOptimizer();
		return optimizer;
	}

	/** Creates an "empty" {@link AssignmentConstraint} constraint. */
	public static AssignmentConstraint createAssignmentConstraint(ILocationExpression operation,
			TypeLiteral<?> srcType, TypeLiteral<?> tgtType) {
		AssignmentConstraint assignmentConstraint =
				ConstraintsFactory.eINSTANCE.createAssignmentConstraint();
		assignmentConstraint.setName("AssignmentConstraint");
		assignmentConstraint.setOperation(operation);
		if(operation instanceof Allocation) {
			assignmentConstraint.setName(assignmentConstraint.getName() + " (Allocation)");
		} else if(operation instanceof Dislocation) {
			assignmentConstraint.setName(assignmentConstraint.getName() + " (Dislocation)");
		}

		assignmentConstraint.setSourceType(srcType);
		assignmentConstraint.setTargetType(tgtType);

		return assignmentConstraint;
	}

	/**
	 * Creates an {@link IsolatedCommunicationSet} {@link InternalConstraint} valid for the given
	 * set of {@code entries}.
	 */
	public static IsolatedCommunicationSet createIsolatedCommunicationSet(Collection<?> entries) {
		IsolatedCommunicationSet isoCommSet =
				ConstraintsFactory.eINSTANCE.createIsolatedCommunicationSet();
		isoCommSet.setName("Isolation Constraint");
		isoCommSet.getEntries().addAll(entries);
		return isoCommSet;
	}

	/**
	 * Creates an {@link ReplicationConstraint} {@link InternalConstraint} valid for the given
	 * {@link IRequestAdapter}.
	 */
	public static ReplicationConstraint createReplicationConstraint(IRequestAdapter requester,
			int min, int max) {
		ReplicationConstraint replConstr =
				ConstraintsFactory.eINSTANCE.createReplicationConstraint();
		replConstr.setName("Replication Constraint");
		replConstr.setRequester(requester);
		replConstr.setMin(min);
		replConstr.setMax(max);
		return replConstr;
	}

	/**
	 * Creates a {@link CouplingConstraint} {@link InternalConstraint} of the given set of
	 * {@link IRequestAdapter}s.
	 */
	public static <T extends IRequestAdapter> CouplingConstraint createCouplingConstraint(
			Collection<T> coupledElements, TypeLiteral<?> srcType, TypeLiteral<?> tgtType) {
		CouplingConstraint couplingConstr = ConstraintsFactory.eINSTANCE.createCouplingConstraint();
		couplingConstr.setName("CouplingConstraint");
		couplingConstr.setSourceType(srcType);
		couplingConstr.setTargetType(tgtType);

		Optional<T> coupledElement = LambdaUtils.getFirst(coupledElements);
		if(coupledElement.isPresent()) {
			if(srcType.getRawType().isAssignableFrom(coupledElement.get().getClass())) {
				couplingConstr.getSources().addAll(coupledElements);
			} else if(tgtType.getRawType().isAssignableFrom(coupledElement.get().getClass())) {
				couplingConstr.getTargets().addAll(coupledElements);
			}
		}

		return couplingConstr;
	}

	/**
	 * Creates a {@link SeparationConstraint} {@link InternalConstraint} of the given set of
	 * isolated sets of {@link IRequestAdapter}s.
	 */
	public static SeparationConstraint
			createSeparationConstraint(Collection<Collection<IRequestAdapter>> sepSets) {
		SeparationConstraint sepConstr = ConstraintsFactory.eINSTANCE.createSeparationConstraint();
		sepConstr.setName("SeparationConstraint");
		for(Collection<IRequestAdapter> isoSet : sepSets) {
			IRequesterSet reqSet = createIRequesterSet(isoSet);
			sepConstr.getSeparatedSets().add(reqSet);
		}
		return sepConstr;
	}

	/** Returns a {@link IRequesterSet} of the given collection of {@link IRequestAdapter}s. */
	public static IRequesterSet createIRequesterSet(Collection<IRequestAdapter> isoSet) {
		IRequesterSet retSet = AdapterFactory.eINSTANCE.createIRequesterSet();
		retSet.getEntries().addAll(isoSet);
		return retSet;
	}

	/** Creates a {@link FixedPropertyLiteral} and assigns the given value. */
	public static <T> FixedPropertyLiteral<T> createFixedPropertyLiteral(Map<?, T> valueMap,
			Set<IModelElement> referencedSet) {
		FixedPropertyLiteral<T> literal = DsmlextFactory.eINSTANCE.createFixedPropertyLiteral();
		literal.eSet(DsmlextPackage.Literals.FIXED_PROPERTY_LITERAL__MAP, valueMap);
		literal.setSetReference(referencedSet);
		return literal;
	}

	/**
	 * Factory method to create {@link PeriodConstraint}s.
	 * 
	 * @throws ConstraintGenerationException
	 *             if the period value is invalid ({@code null}, or <= 0).
	 */
	public static PeriodConstraint createPeriodConstraint(IRequestAdapter requester, Double period)
			throws ConstraintGenerationException {
		if(requester == null) {
			throw new ConstraintGenerationException(PeriodConstraint.class,
					"Cannot generate a PeriodConstraint for a 'null' Component.");
		}
		if(period == null || period <= 0.0) {
			throw new ConstraintGenerationException(PeriodConstraint.class,
					"Invalid period specified for the Component " + requester.getName() + ".");
		}

		PeriodConstraint periodConstraint = ConstraintsFactory.eINSTANCE.createPeriodConstraint();
		periodConstraint.setName("Period Constraint(" + periodConstraint.getId() + ")");
		periodConstraint.setRequester(requester);
		periodConstraint.setPeriod(period);
		return periodConstraint;
	}

	/**
	 * Creates a single constraint that states a solution may not contain partitions without
	 * assigned tasks.
	 */
	public static ExplorationConstraint<?> createNoEmptyPartitionConstraint() {
		return ConstraintsFactory.eINSTANCE.createNoEmptyPartitionConstraint();
	}

	/**
	 * Creates a single constraint that states a partition may only host tasks with the same SIL.
	 */
	public static ExplorationConstraint<?> createSameSilInPartitionConstraint() {
		return ConstraintsFactory.eINSTANCE.createSameSilInPartitionConstraint();
	}

	/**
	 * Creates an {@link ExplorationObjective} that aims to minimize the number of instantiated
	 * partitions.
	 */
	public static PartitionMinObjective createPartitionMinObjective() {
		PartitionMinObjective partMinObjective =
				ObjectivesFactory.eINSTANCE.createPartitionMinObjective();
		partMinObjective.setName("Partition Count Minimization");
		return partMinObjective;
	}

	/**
	 * Creates an {@link ExplorationObjective} that aims to minimize the finish times of tasks of a
	 * resource by their sums.
	 */
	public static LatencyMinObjective createLatencyMinObjective() {
		LatencyMinObjective latencyMinObjective =
				ObjectivesFactory.eINSTANCE.createLatencyMinObjective();
		latencyMinObjective.setName("Latency Minimization");
		return latencyMinObjective;
	}

	/** Creates an {@link InterferenceLatencyTable}. */
	public static InterferenceLatencyTable createInterferenceLatencyTable() {
		return AllocationFactory.eINSTANCE.createInterferenceLatencyTable();
	}
}
