/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static com.google.common.collect.Sets.newHashSet;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.ui.services.IEvaluationService;
import org.fortiss.af3.exploration.alg.ExplorationAlgActivator;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

import com.google.inject.Injector;
import com.google.inject.Module;
import com.google.inject.TypeLiteral;

/**
 * Utility methods for the interaction with the {@link IExplorationFeatureService}.
 * 
 * @author diewald
 */
public class ExplorationFeatureUtils {

	/**
	 * From a set of enabled {@link IExplorationFeature}s of the given {@link DseSpecification},
	 * get the {@link IModelElement} types that can be produced by the registered Transformators.
	 * 
	 * @param enabledFeatures
	 *            Collection of enabled features of a DSE run. Obtainable via
	 *            {@link #getEnabledApplicableFeatures(DseSpecification, Injector, boolean)}.
	 * @return set of producable outputs.
	 */
	public static Set<TypeLiteral<SuperSet<? extends IModelElement>>>
			getTransformedOutputsByFeatures(Collection<IExplorationFeature> enabledFeatures) {
		Set<TypeLiteral<SuperSet<? extends IModelElement>>> outTypes = new HashSet<>();
		for(IExplorationFeature feature : enabledFeatures) {
			outTypes.addAll(feature.getSolutionTypes());
		}
		return outTypes;
	}

	/**
	 * From a set of enabled {@link IExplorationFeature}s of the given {@link DseSpecification},
	 * this method returns those features that are executable w.r.t. the set of available data from
	 * the input models provided by the {@code dseInjector}.
	 * 
	 * @param dseSpec
	 *            contains the enabled {@link IExplorationFeature}s.
	 * @param dseInjector
	 *            provides the available input models.
	 * @param logDisabled
	 *            if non-applicable shall be logged.
	 * @return collection of enabled and applicable {@link IExplorationFeature}s.
	 */
	public static Collection<IExplorationFeature> getEnabledApplicableFeatures(
			DseSpecification dseSpec, Injector dseInjector, boolean logDisabled) {
		Collection<IExplorationFeature> userEnabledFeatures =
				getService(IExplorationFeatureService.class)
						.getExplorationFeatures(dseSpec.getFeatures());
		Collection<IExplorationFeature> enabledFeatures = new HashSet<>();
		for(IExplorationFeature feature : userEnabledFeatures) {
			if(feature.isApplicableToInputs(dseInjector)) {
				enabledFeatures.add(feature);
			} else {
				// @CodeFormatterOff
				LoggingUtils.warning(ExplorationAlgActivator.getDefault(),
						"The exploration feature " + feature.getName() + " has been disabled" +
						" since the input models do not contain sufficient data to execute this" +
						" feature.");
				// @CodeFormatterOn
			}
		}
		return enabledFeatures;
	}

	/**
	 * Returns the set of {@link DecoderModule}s activated by the given {@link Collection} of
	 * enabled {@link IExplorationFeature}s.
	 */
	public static Set<DecoderModule<?>>
			getEnabledDecoders(Collection<IExplorationFeature> enabledFeatures) {
		Set<DecoderModule<?>> enabledDecoders = newHashSet();
		enabledFeatures.forEach(f -> enabledDecoders.addAll(f.getDecoders()));

		return removeOverrides(enabledDecoders, enabledFeatures);
	}

	/**
	 * Removes those {@link Module}s from the input set that are defined as overrides in the
	 * {@code enabledFeature}s.
	 */
	private static <M extends Module> Set<M> removeOverrides(Set<M> modules,
			Collection<IExplorationFeature> enabledFeatures) {
		Collection<Class<? extends Module>> overrides = new ArrayList<>();
		enabledFeatures.forEach(f -> overrides.addAll(f.getModuleOverrides()));

		Collection<Module> rmList = new ArrayList<>();
		for(Class<? extends Module> override : overrides) {
			modules.stream().filter(d -> d.getClass().isAssignableFrom(override))
					.forEach(e -> rmList.add(e));
		}
		rmList.forEach(d -> modules.remove(d));

		return modules;
	}

	/**
	 * Returns the set of {@link DecoderModule}s activated by the given {@link Collection} of
	 * enabled {@link IExplorationFeature}s.
	 */
	public static Map<TypeLiteral<?>, InputModule<?>>
			getEnabledInputModules(Collection<IExplorationFeature> enabledFeatures) {
		Map<TypeLiteral<?>, InputModule<?>> enabledInputModules = new HashMap<>();
		enabledFeatures.forEach(f -> enabledInputModules.putAll(f.getInputModules()));
		return enabledInputModules;
	}

	/**
	 * Returns the set of enabled Evaluators from the {@link IEvaluationService}, subject to the
	 * given {@link ExplorationSpecification}.
	 * 
	 * @param expSpec
	 *            contains the {@link ExplorationTarget}s to evaluate.
	 * @param enabledFeatures
	 *            collection of enabled features of the current dse run.
	 * @return see above.
	 */
	public static Set<IEvaluatorModule<?, ?, ?>> getEnabledEvaluators(
			ExplorationSpecification expSpec, Collection<IExplorationFeature> enabledFeatures) {
		Collection<Class<? extends Module>> overrides = new ArrayList<>();
		enabledFeatures.forEach(f -> overrides.addAll(f.getModuleOverrides()));

		Set<IEvaluatorModule<?, ?, ?>> enabledEvaluators =
				getService(IExplorationEvaluatationService.class).getEvaluatorInstances(
						new HashSet<>(expSpec.getTargets()), createPriorityComparator(), overrides,
						enabledFeatures);

		return removeOverrides(enabledEvaluators, enabledFeatures);
	}

	/** Comparator that prefers evaluators of {@link ExplorationConstraint}s. */
	private static Comparator<IEvaluatorModule<?, ?, ?>> createPriorityComparator() {
		return new Comparator<IEvaluatorModule<?, ?, ?>>() {
			@Override
			public int compare(IEvaluatorModule<?, ?, ?> o1, IEvaluatorModule<?, ?, ?> o2) {
				if((o1.getExplorationTarget() instanceof ExplorationConstraint<?>) &&
						(o2.getExplorationTarget() instanceof ExplorationConstraint<?>)) {
					return o1.hashCode() - o2.hashCode();
				} else if(o1.getExplorationTarget() instanceof ExplorationConstraint<?>) {
					return 1;
				} else if(o2.getExplorationTarget() instanceof ExplorationConstraint<?>) {
					return -1;
				}
				return o1.hashCode() - o2.hashCode();
			}
		};
	}
}
