/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.jgrapht.Graphs.predecessorListOf;
import static org.jgrapht.Graphs.successorListOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Predicate;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.jgrapht.Graph;
import org.jgrapht.Graphs;
import org.jgrapht.alg.connectivity.ConnectivityInspector;
import org.jgrapht.graph.DirectedWeightedMultigraph;
import org.jgrapht.graph.MaskSubgraph;

/**
 * Utility methods that extend the analysis capabilities for JGrapht graphs.
 * 
 * @author diewald
 */
public class GraphUtils {

	/** Constructor. Prevent initialization. */
	private GraphUtils() {
	}

	/**
	 * Returns the first found {@link ITaskGraph} from the given collection that contains the given
	 * {@link ITaskAdapter}.
	 * 
	 * @param vertex
	 *            {@link ITaskAdapter} to search for.
	 * @param graphCollection
	 *            Collection of {@link ITaskGraph}s to be examined.
	 * @return First {@link ITaskGraph} containing the given vertex.
	 */
	public static <S extends ITaskAdapter, T extends ITaskGraph<S>> T getGraphOf(S vertex,
			Collection<T> graphCollection) {
		for(T graph : graphCollection) {
			if(graph.containsTask(vertex)) {
				return graph;
			}
		}
		// Fallback: get graph via the ref object. Must be in the same graph.
		// FIXME: remove this fallback!!!
		return getGraphByObjOf(vertex, graphCollection);
		// return null;
	}

	/**
	 * Returns the first found {@link ITaskGraph} from the given collection that contains the given
	 * {@link ITaskAdapter}. The {@link ITaskGraph} is identified by comparing the
	 * referenced objects of the contained tasks, which must be in the same graphs by construction.
	 * 
	 * @param vertex
	 *            {@link ITaskAdapter} to search for.
	 * @param graphCollection
	 *            Collection of {@link ITaskGraph}s to be examined.
	 * @return First {@link ITaskGraph} containing the given vertex.
	 */
	public static <S extends ITaskAdapter, T extends ITaskGraph<S>> T getGraphByObjOf(S vertex,
			Collection<T> graphCollection) {
		for(T graph : graphCollection) {
			for(ITaskAdapter comp : graph.getTasks()) {
				if(comp.getObject(Object.class) == vertex.getObject(Object.class)) {
					return graph;
				}
			}
		}
		return null;
	}

	/**
	 * Determines all predecessors of the given vertex, not just the direct ones as provided by the
	 * {@link Graphs} utility class.
	 * 
	 * @param graph
	 *            The graph containing the vertex {#link vertex}.
	 * @param vertex
	 *            The vertex whose predecessors shall be determined.
	 * @return Collection of predecessor vertices.
	 */
	public static <V, E> Set<V> getAllPredescesorsOf(DirectedWeightedMultigraph<V, E> graph,
			V vertex) {
		Set<V> predescessorList = new HashSet<V>();
		ArrayList<V> workingQueue = new ArrayList<V>();
		workingQueue.addAll(Graphs.predecessorListOf(graph, vertex));
		while(!workingQueue.isEmpty()) {
			// get(0): Existence ensured by while clause.
			V currentPredescessor = workingQueue.get(0);
			workingQueue.remove(0);

			if(!predescessorList.contains(currentPredescessor)) {
				predescessorList.add(currentPredescessor);
				workingQueue.addAll(predecessorListOf(graph, currentPredescessor));
			}
		}
		return predescessorList;
	}

	/**
	 * Determines all successors of the given vertex, not just the direct ones as provided by the
	 * {@link Graphs} utility class.
	 * 
	 * @param graph
	 *            The graph containing the vertex {#link vertex}.
	 * @param vertex
	 *            The vertex whose predecessors shall be determined.
	 * @return Collection of predecessor vertices.
	 */
	public static <V, E> Set<V> getAllSuccessorsOf(DirectedWeightedMultigraph<V, E> graph,
			V vertex) {
		Set<V> successorList = new HashSet<V>();
		ArrayList<V> workingQueue = new ArrayList<V>();
		workingQueue.addAll(successorListOf(graph, vertex));
		while(!workingQueue.isEmpty()) {
			// get(0): Existence ensured by while clause.
			V currentSuccescessor = workingQueue.get(0);
			workingQueue.remove(0);

			if(!successorList.contains(currentSuccescessor)) {
				successorList.add(currentSuccescessor);
				workingQueue.addAll(successorListOf(graph, currentSuccescessor));
			}
		}
		return successorList;
	}

	/** Returns all edges of the given {@link Graph}. */
	public static <V, E> Set<E> getAllEdgesOf(Graph<V, E> graph) {
		Set<E> edges = new HashSet<>();
		graph.vertexSet().forEach(v -> edges.addAll(graph.incomingEdgesOf(v)));
		graph.vertexSet().forEach(v -> edges.addAll(graph.outgoingEdgesOf(v)));
		return edges;
	}

	/**
	 * Constructs a sub graph from the given {@code srcGraph} that is specified by including the
	 * given {@code subGraphVertex} and disconnected to the rest of the graph by the given
	 * {@link Predicate}. The resulting subgraph (as its vertices and edges) is added to the given
	 * {@code targetGraph}.
	 * 
	 * @param srcGraph
	 *            {@link Graph} containing the subtree.
	 * @param subGraphVertex
	 *            Vertex included in the subtree.
	 * @param neighborCutPred
	 *            {@link Predicate} to disconnect the subtree from the remaining vertices of the
	 *            graph.
	 * @param targetGraph
	 *            {@link Graph} used as the target of the operation.
	 */
	public static <V, E, G extends Graph<V, E>> void getSubGraph(G srcGraph, V subGraphVertex,
			Predicate<V> neighborCutPred, G targetGraph) {
		MaskSubgraph<V, E> subGraph = new MaskSubgraph<>(srcGraph, neighborCutPred, e -> false);

		ConnectivityInspector<V, E> ci = new ConnectivityInspector<>(subGraph);
		Set<V> subVertices = ci.connectedSetOf(subGraphVertex);

		Graphs.addAllVertices(targetGraph, subVertices);
		Graphs.addAllEdges(targetGraph, subGraph, getAllEdgesOf(subGraph));
	}

	/**
	 * Replaces the given {@code oldVertex} with the given {@code newVertex} in the given
	 * {@link Graph}.
	 * 
	 * @param graph
	 *            {@link Graph} to modify.
	 * @param oldVertex
	 *            vertex to replace.
	 * @param newVertex
	 *            vertex to insert.
	 */
	public static <V, E> void replaceVertex(Graph<V, E> graph, V oldVertex, V newVertex) {
		Collection<E> outgoingEdges = graph.outgoingEdgesOf(oldVertex);
		Collection<E> incomingEdges = graph.incomingEdgesOf(oldVertex);

		graph.addVertex(newVertex);
		outgoingEdges.forEach(out -> graph.addEdge(newVertex, graph.getEdgeTarget(out), out));
		incomingEdges.forEach(in -> graph.addEdge(graph.getEdgeSource(in), newVertex, in));

		graph.removeVertex(oldVertex);
	}
}
